(* CSL with Barriers, v0.1
 * 
 * It is our intention to release the files under a BSD-style license shortly
 * after some badly-needed cleanup work,  
 * 
 * Until then:
 * Most of this file is 
 * (c) 2010 Aquinas Hobor, Cristian Gherghina, and Christopher Chak
 * All Rights Reserved
 *
 * See below for information regarding copyright status of material taken from the
 * Concurrent C Minor Project
 *
 * Permission granted to use these proof scripts only for the purpose of 
 * evaluating "Barriers in Concurrent Separation Logic."
 *)

(* THIS FILE CONTAINS EXCERPTS FROM THE CONCURRENT 
 * C MINOR PROJECT, FILES: coqlib2.v, list_lemmas.v, sepalg_list.v
 *  
 * Such excerpts are bracketed by comments; are  
 * (c) 2007-2010 Andrew Appel, Rob Dockins, Aquinas Hobor;
 * and are made available under the following restrictions: 
 *    PERMISSION GRANTED ONLY FOR USE IN EVALUATING
 *    THE PAPER 'BARRIERS IN CONCURRENT SEPARATION 
 *    LOGIC', BY AQUINAS HOBOR AND CRISTIAN GHERGHINA
 *)

(* 
 * This file defines some custom tactics and general helper lemmas.  Some
 * may be moved into the MSL.
 *)

Require Export msl.

Tactic Notation "disc" := (try discriminate).
Tactic Notation "contr" := (try contradiction).
Tactic Notation  "icase" constr(v) := (destruct v; disc; contr; auto).
Tactic Notation "LEM" constr(P) :=
  (destruct (classic (P))).
Tactic Notation "omegac" := (elimtype False; omega).
Tactic Notation "copy" hyp(H) := (generalize H; intro).

Definition ojoin (A : Type) (b1 b2 b3 : option A) : Prop :=
  match (b1, b2, b3) with
    | (None, None, None) => True
    | (Some x, None, Some y) => x = y
    | (None, Some x, Some y) => x = y
    | _ => False
   end.
Implicit Arguments ojoin.

Lemma saf_ojoin : forall A, sepalgfacts (@ojoin A).
Proof.
  constructor.

(*1*)
  intros.
  destruct x.
  destruct y; try contradiction.
  destruct z; try contradiction.
  destruct z'; try contradiction.
  congruence.
  destruct y.
  destruct z; try contradiction.
  destruct z'; try contradiction.
  congruence.
  destruct z; try contradiction.
  destruct z'; try contradiction.
  trivial.
  
(*2*)
  intros.
  destruct d.
  destruct c.
  unfold ojoin in H0.
  contradiction.
  exists b.
  split.
  unfold ojoin.
  destruct b; trivial.
  destruct e.
  unfold ojoin in H0.
  subst.
  auto.
  unfold ojoin in H0; contradiction.
  destruct a.
  unfold ojoin in H.
  destruct b; contradiction.
  destruct b.
  unfold ojoin in H.
  contradiction.
  exists c.
  split.
  destruct c.
  unfold ojoin.
  auto.
  trivial.
  auto.

(*3*)
  intros.
  destruct b.
  destruct a.
  unfold ojoin in H.
  inversion H.
  destruct c.
  unfold ojoin in H.
  unfold ojoin.
  auto.
  unfold ojoin in H.
  contradiction.
  destruct c.
  destruct a.
  unfold ojoin in H.
  unfold ojoin.
  auto.
  unfold ojoin in *.
  exact H.
  unfold ojoin in *.
  exact H.

(*4*)
  intros.
  destruct b.
  destruct a1.
  destruct a2.
  unfold ojoin in *.
  contradiction.
  unfold ojoin in H.
  contradiction.
  destruct a2.
  unfold ojoin in H0.
  contradiction.
  auto.
  destruct a1.
  destruct a2.
  destruct c.
  unfold ojoin in *.
  simpl in *; congruence.
  unfold ojoin in *.
  contradiction.
  destruct c.
  unfold ojoin in H0.
  contradiction.
  unfold ojoin in H.
  contradiction.
  destruct c.
  unfold ojoin in H; contradiction.
  destruct a2.
  unfold ojoin in H0; contradiction.
  auto.

(* exist identities *)
  intro.
  exists None.
  destruct a.
  unfold ojoin.
  trivial.
  unfold ojoin.
  trivial.
  (* self_join *)
  intros.
  destruct a.
  unfold ojoin in H.
  contradiction.
  unfold ojoin in H.
  induction b.
  contradiction.
  constructor I.
Qed.
Implicit Arguments saf_ojoin.

Definition sa_option (A : Type) : sepalg (option A) :=
  SepAlg (@saf_ojoin A).

Lemma nidentity_pair2 : forall (A B : Type) (saB : sepalg B) (a : A) (b : B),
  ~(@identity (A * B) (sa_prod (sa_equiv A) saB) (a, b)) ->
  ~identity b.
Proof.
  repeat intro.
  apply H.
  rewrite identity_unit_equiv.
  rewrite identity_unit_equiv in H0.
  firstorder.
Qed.
Implicit Arguments nidentity_pair2.

Lemma n2_identity_prod : forall P Q (sa_p : sepalg P) (sa_q : sepalg Q) p q,
  ~identity q -> ~(@identity (P *Q) (sa_prod sa_p sa_q) (p, q)).
Proof.
  repeat intro.
  apply H.
  rewrite identity_unit_equiv in H0.
  rewrite identity_unit_equiv.
  destruct H0.
  trivial.
Qed.
Implicit Arguments n2_identity_prod.

(* New list facts *)

Lemma map_comp : forall A B C (f : B -> C) (g : A -> B),
  map (f oo g) = map f oo map g.
Proof.
  intros.
  extensionality L.
  induction L.
  trivial.
  unfold compose in *.
  simpl.
  rewrite IHL.
  trivial.
Qed.

Lemma map_comp' : forall A B C (f : B -> C) (g : A -> B) L,
  map f (map g L) = map (fun x => f (g x)) L.
  intros.
  change (map f (map g L)) with (((map f) oo (map g)) L).
  rewrite <- map_comp.
  trivial.
Qed.

(* This zip always succeds and just drops extra elements. *)
Fixpoint zip (A : Type) (B : Type) (L1 : list A) (L2 : list B) : list (A * B) :=
  match (L1, L2) with
    | (a :: L1', b :: L2') => (a,b) :: zip A B L1' L2'
    | _ => nil
   end.
Implicit Arguments zip.

Lemma zip_unzip: forall A B (L1 : list A) (L2 : list B),
  length L1 = length L2 ->
  forall i,
    match (nth_error L1 i, nth_error L2 i, nth_error (zip L1 L2) i) with
      | (None, None, None) => True
      | (Some a, Some b, Some (a', b')) => a' = a /\ b' = b
      | _ => False
    end.
Proof.
  induction L1; simpl; auto; intros.
  destruct i; icase L2; simpl; trivial.
  destruct L2; simpl in *.
  omegac.
  destruct i; simpl.
  tauto.
  inversion H.
  spec IHL1 L2 H1.
  apply (IHL1 i).
Qed.

Lemma zip_fst : forall A B (L1 : list A) (L2 : list B),
  length L1 = length L2 ->
  map (@fst _ _) (zip L1 L2) = L1.
Proof.
  induction L1; intros.
  icase L2; auto.
  icase L2.
  simpl in H.
  simpl.
  rewrite IHL1; auto.
Qed.

Lemma zip_snd : forall A B (L1 : list A) (L2 : list B),
  length L1 = length L2 ->
  map (@snd _ _) (zip L1 L2) = L2.
Proof.
  induction L1; intros.
  icase L2; auto.
  icase L2.
  simpl in H.
  simpl.
  rewrite IHL1; auto.
Qed.

Lemma nth_error_bound : forall A L i (x : A),
  nth_error L i = Some x -> 
  i < length L.
Proof.
  induction L.
  intros.
  icase i.
  intros.
  destruct i.
  simpl.
  omega.
  simpl in H.
  simpl.
  spec IHL i x H.
  omega.
Qed.

Lemma can_nth_error : forall A (L : list A) i,
  i < length L ->
  exists x,
    nth_error L i = Some x.
Proof.
  induction L; simpl; intros.
  omegac.
  destruct i.
  exists a.
  trivial.
  simpl.
  apply IHL.
  omega.
Qed.

Lemma nth_error_bound2 : forall A (L : list A) i,
  nth_error L i = None -> 
  i >= length L.
Proof.
  induction L; intros.
  icase i.
  simpl.
  omega.
  icase i.
  simpl in *.
  spec IHL i H.
  omega.
Qed.
Implicit Arguments nth_error_bound2.

(* Utility lemma *)
Lemma eq_nth_error: forall A i (L L' : list A),
  L = L' ->
  nth_error L i = nth_error L' i.
Proof.
  congruence.
Qed.
Implicit Arguments eq_nth_error.

Lemma nth_error_map: forall A B (f : A -> B) L n,
  nth_error (map f L) n = option_map f (nth_error L n).
Proof.
  induction L; intros.
  simpl.
  icase n; trivial.
  destruct n; trivial.
  simpl.
  apply IHL.
Qed.

Lemma In_nth_error: forall A (x : A) L,
  In x L ->
  exists j,
    nth_error L j = Some x.
Proof.
  induction L; intros.
  contr.
  destruct H.
  subst.
  exists 0.
  trivial.
  spec IHL H.
  destruct IHL.
  exists (S x0).
  simpl.
  trivial.
Qed.

Lemma nth_error_In: forall A L j (x : A),
  nth_error L j = Some x ->
  In x L.
Proof.
  induction L; intros.
  icase j.
  destruct j.
  inversion H.
  subst x.
  simpl.
  auto.
  right.
  apply IHL with j.
  apply H.
Qed.

Definition count_optionlist (A : Type) (AOL : list (option A)) : nat :=
  fold_right (fun m n => match m with None => n | Some _ => n + 1 end) 0 AOL.
Implicit Arguments count_optionlist.

Lemma count_optionlist_None : forall A (L : list (option A)),
  count_optionlist (None :: L) = count_optionlist L.
Proof.
  trivial.
Qed.

Lemma count_optionlist_Some : forall A L (a : A),
  count_optionlist ((Some a) :: L) = count_optionlist L + 1.
Proof.
  trivial.
Qed.

Lemma count_optionlist_nth_error: forall A (L: list (option A)),
  count_optionlist L = 0 ->
  forall j x,
  nth_error L j = Some (Some x) ->
  False.
Proof.
  induction L; intros.
  icase j.
  destruct j.
  inversion H0.
  subst.
  simpl in H.
  omega.
  simpl in H0.
  destruct a; simpl in H.
  omega.
  spec IHL H j x H0.
  trivial.
Qed.

Lemma count_optionlist_Slength: forall A (al : list (option A)),
  ~(count_optionlist al = S (length al)).
Proof.
  repeat intro.
  assert (count_optionlist al > length al) by omega.
  clear H.
  induction al.
  simpl in H0.
  omega.
  apply IHal.
  destruct a;
  simpl in H0;
  omega.
Qed.

Definition fullolist (A : Type) (L : list (option A)) : Prop :=
  count_optionlist L = length L.
Implicit Arguments fullolist.

Definition dualolist (A : Type) (L : list (option A)) (L' : list A) : Prop :=
  L = map (@Some A) L'.
Implicit Arguments dualolist.

Lemma fullo_dualo : forall A (L : list (option A)),
  fullolist L ->
  exists L', dualolist L L'.
Proof.
  induction L.
  exists nil.
  compute.
  trivial.
  intro.
  unfold fullolist in H.
  destruct a.
  simpl in H.
  assert (count_optionlist L = length L) by omega.
  spec IHL H0.
  destruct IHL as [L' ?].
  exists (a :: L').
  unfold dualolist.
  simpl.
  rewrite H1.
  trivial.
  simpl in H.
  contradiction (count_optionlist_Slength A L).
Qed.

Lemma dualo_map: forall A B L L' (f : A -> B),
  dualolist L L' ->
  dualolist (map (option_map f) L) (map f L').
Proof.
  unfold dualolist.
  induction L; intros.
  icase L'.
  icase L'.
  inversion H.
  subst a.
  simpl.
  spec IHL L' f H2.
  congruence.
Qed.

Lemma dualo_length : forall A L (L' : list A),
  dualolist L L' ->
  length L = length L'.
Proof.
  induction L; intros.
  icase L'; trivial.
  icase L'.
  inversion H.
  simpl.
  rewrite <- H2.
  rewrite (IHL L'); trivial.
Qed.

Lemma None_fullolist: forall A (L : list (option A)),
  ~ (fullolist (None :: L)).
Proof.
  repeat intro.
  unfold fullolist in H.
  simpl in H.
  contradiction (count_optionlist_Slength _ L).
Qed.  

Lemma Some_fullolist: forall A L (a : A),
  fullolist (Some a :: L) ->
  fullolist L.
Proof.
  intros.
  red in H.
  simpl in H.
  red.
  omega.
Qed.
Implicit Arguments Some_fullolist.

Lemma fullo_no_None: forall A (L : list (option A)) i,
  fullolist L ->
  nth_error L i = Some None ->
  False.
Proof.
  induction L; intros.
  icase i.
  destruct i.
  simpl in *.
  inversion H0.
  subst.
  eapply None_fullolist; eauto.
  destruct a.
  2: eapply None_fullolist; eauto.
  apply Some_fullolist in H.
  spec IHL i H H0.
  trivial.
Qed.
Implicit Arguments fullo_no_None.

Fixpoint n_None (A : Type) (n : nat) : list (option A) :=
  match n with 0 => nil | S n' => None :: (n_None A n') end.

(* The following facts are a complete reimplementation & clean-up of some 
    material in ecm/list_lemmas.v.  They are NOT from CCM Project. *)

Lemma nth_error_eq: forall A (l1 l2 : list A),
  (forall i, nth_error l1 i = nth_error l2 i) ->
  l1 = l2.
Proof.
  induction l1; intros.
  spec H 0.
  simpl in H.
  icase l2.
  trivial.
  destruct l2.
  spec H 0.
  inversion H.
  generalize (H 0); intro.
  simpl in H0.
  inversion H0.
  subst.
  rewrite (IHl1 l2).
  trivial.
  intro.
  spec H (S i).
  apply H.
Qed.

Fixpoint replace (A : Type) (xs : list A) (n : nat) (a : A) : option (list A) :=
  match (n, xs) with
    | (0, x :: xs') => Some (a :: xs')
    | (S n, x :: xs') => match replace A xs' n a with
                                    | Some xs'' => Some (x :: xs'')
                                    | None => None
                                  end
    | (_, nil) => None
   end.
Implicit Arguments replace.

Lemma replace_nth : forall A n (xs : list A) a xs',
  replace xs n a = Some xs' ->
  nth_error xs' n = Some a.
Proof.
  induction n.
  intros.
  destruct xs; disc.
  inversion H.
  trivial.
  intros.
  destruct xs; disc.
  simpl in H.
  remember (replace xs n a).
  destruct o; disc.
  inversion H.
  simpl.
  apply IHn with xs.
  auto.
Qed.

Lemma length_replace: forall A (L : list A) i v L',
  replace L i v = Some L' ->
  length L = length L'.
Proof.
  induction L; intros.
  icase i.
  icase i.
  inversion H.
  trivial.
  simpl in H.
  remember (replace L i v).
  icase o.
  spec IHL i v l.
  inversion H.
  simpl.
  rewrite IHL; auto.
Qed.

Lemma replace_nth2 : forall A n (xs : list A) a xs' n',
  replace xs n a = Some xs' ->
  n <> n' ->
  nth_error xs' n' = nth_error xs n'.
Proof.
  induction n.
  intros.
  destruct xs; disc.
  inversion H.
  destruct n'.
  contradiction H0.
  trivial.
  trivial.
  intros.
  destruct xs; disc.
  simpl in H.
  remember (replace xs n a).
  destruct o; disc.
  inversion H.
  destruct n'.
  trivial.
  simpl.
  apply IHn with a.
  auto.
  omega.
Qed.

Lemma replace_inveq: forall A l1 l1' i (x x' : A) l2,
  replace l1 i x = Some l2 ->
  replace l1' i x' = Some l2 ->
  nth_error l1 i = nth_error l1' i ->
  l1 = l1'.
Proof.
  intros.
  apply nth_error_eq.
  intro.
  case (eq_nat_dec i i0); intro.
  subst i.
  trivial.
  apply replace_nth2 with (n' := i0) in H; trivial.
  apply replace_nth2 with (n' := i0) in H0; trivial.
  congruence.
Qed.

Lemma can_replace : forall A L i (x : A),
  nth_error L i = Some x ->
    forall x',
      exists L',  
        replace L i x' = Some L'.
Proof.
  induction L; intros.
  icase i.
  icase i.
  simpl in H.
  inversion H.
  subst.
  exists (x' :: L).
  trivial.
  simpl in H.
  spec IHL i x H x'.
  destruct IHL as [L' ?].
  exists (a :: L').
  simpl.
  rewrite H0.
  trivial.
Qed.

Lemma replace_replace : forall A L i (x x' : A) L',
  nth_error L i = Some x ->
  replace L i x' = Some L' ->
  replace L' i x = Some L.
Proof.
  induction L; intros.
  icase i.
  icase i.
  simpl in H, H0.
  inversion H.
  subst x.
  inversion H0.
  simpl.
  trivial.
  simpl in H.
  spec IHL i x.
  simpl in H0.
  remember (replace L i x').
  icase o.
  inversion H0.
  subst L'.
  symmetry in Heqo.
  spec IHL x' l H Heqo.
  simpl.
  rewrite IHL.
  trivial.
Qed.

Lemma replace_eq : forall A L n (x : A) L',
  replace L n x = Some L' ->
  forall y,
    replace L n y = replace L' n y.
Proof.
  induction L; intros.
  icase n.
  destruct n.
  inversion H.
  subst.
  trivial.
  inversion H.
  remember (replace L n x).
  icase o.
  inversion H1; subst.
  simpl.
  spec IHL n x l.
  rewrite IHL; auto.
Qed.

(* END reimplement of parts of ecm/list_lemmas.v *)

(* From coqlib2.v *)
Lemma if_true: forall (A: Prop) (E: {A}+{~A}) (T: Type) (B C: T), A -> (if E then B else C) = B.
Proof.
intros.
destruct E; auto.
contradiction.
Qed.

Lemma if_false: forall (A: Prop) (E: {A}+{~A}) (T: Type) (B C: T), ~A -> (if E then B else C) = C.
Proof.
intros.
destruct E; auto.
contradiction.
Qed.

Tactic Notation "if_tac" := match goal with |- context [if ?a then _ else _] => destruct a as [?H | ?H] end.
Tactic Notation "if_tac" simple_intropattern(H) 
   := match goal with |- context [if ?a then _ else _] => destruct a as H end.
Tactic Notation "if_tac" "in" hyp(H) 
 := match type of H with context [if ?a then _ else _] => destruct a as [?H | ?H] end.
Tactic Notation "if_tac" simple_intropattern(H) "in" hyp(H1)
 := match type of H1 with context [if ?a then _ else _] => destruct a as H end.

(* END ecm/coqlib2.v *)

(* FROM CCM PROJECT: ecm/list_lemmas.v *)

Inductive fold_rel (A: Type) (f: A -> A -> A -> Prop):  A -> list A -> A -> Prop :=
| fold_rel_nil: forall w, fold_rel A f w nil w
| fold_rel_cons: forall a l w w1 w2,
            f w a w1 ->
            fold_rel A f w1 l w2 ->
            fold_rel A f w (a::l) w2.
Implicit Arguments fold_rel.

(* END ecm/list_lemmas.v *)

(* FROM CCM PROJECT: ecm/sepalg_list.v *)
(* Minor modifications by Aquinas Hobor *)

Set Implicit Arguments.

Hint Resolve @join_sub_refl.

Definition join_list `{sepalg A}: A -> list A -> A -> Prop :=
   fold_rel join.

Lemma join_list_assoc1 `{sepalg A}:
 forall {a b cl d e},
   join a b d -> join_list d cl e ->
  exists f, join_list b cl f /\ join a f e.
Proof.
intros A HA.
do 3 intro; revert a b; induction cl; intros; inv H0.
exists b; split; auto; constructor.
destruct (IHcl _ _ _ _ H4 H6) as [f [? ?]].
rename w1 into phi.
(* why is this change needed? *)
destruct (join_assoc _ _ _ _ _ H H1) as [g [? ?]].
exists g; split; auto.
(* why is this change needed? *)
destruct (join_assoc _ _ _ _ _ H H4) as [h [? ?]].
constructor 2 with h.
auto.
spec IHcl a0 h phi e. spec IHcl H7 H6.
destruct IHcl as [j [? ?]].
(* why is this change needed? *)
generalize (join_canc _ _ _ _ (join_com _ _ _ H3) (join_com _ _ _ H9)); intro; subst j.
auto.
Qed.

Lemma join_list_assoc2 `{sepalg A}:
  forall {a b cl e f},
   join_list b cl f -> join a f e ->
   exists d,  join a b d /\ join_list d cl e.
Proof.
intros A HA.
do 3 intro; revert a b; induction cl; intros; inv H.
exists e; split; auto; constructor.
destruct (IHcl _ _ _ _ H6 H0) as [d [? ?]].
(* why is this change needed? *)
destruct (join_assoc _ _ _ _ _ (join_com _ _ _ H4) (join_com _ _ _ H)) as [g [? ?]].
exists g; split; auto.
(* why is this change needed? *)
apply join_com; trivial.
constructor 2 with d; auto.
(* why is this change needed? *)
apply join_com; trivial.
Qed.

Lemma join_list_app `{sepalg A}:
  forall {a bl c dl e},
   join_list a bl c -> join_list c dl e -> join_list a (bl++dl) e.
Proof.
intros A HA.
intros a bl; revert a; induction bl; intros.
inv H.
simpl.
auto.
inv H.
rewrite <- app_comm_cons.
econstructor 2.
apply H4.
eapply IHbl.
apply H6.
auto.
Qed.

Lemma join_list_unapp `{sepalg A}:
  forall {a bl dl e},
  join_list a (bl++dl) e -> exists c, join_list a bl c /\ join_list c dl e.
Proof.
intros A HA.
intros a bl; revert a; induction bl; intros.
simpl in H. exists a; split; auto. constructor.
rewrite <- app_comm_cons in H.
inv H.
destruct (IHbl _ _ _ H5) as [c [? ?]]. clear IHbl.
exists c.
split; auto.
econstructor 2; eauto.
Qed.

Lemma join_list_1 `{sepalg A}:
  forall a b c, join a b c = join_list a (b::nil) c.
Proof.
intros A HA.
intros; apply prop_ext; split; intros.
econstructor 2; eauto.
constructor.
inv H.
inv H5.
auto.
Qed.

Lemma join_list_split_nth `{sepalg A}:
 forall n l phin phi phia phib phi2,
  nth_error l n = Some phin -> join_list phi l phi2 -> join phib phia phin ->
           exists phic, exists l',
                 replace l n phib = Some l' /\ join phi phia phic /\ join_list phic l' phi2.
Proof.
intros A saA.
induction n; intros.
inv H.
destruct l.
discriminate.
inv H3.
inv H0.
simpl.
destruct (join_assoc _ _ _ _ _ H1 (join_com _ _ _ H4))
      as [f [? ?]].
exists f; exists (phib::l).
repeat split; auto.
apply join_com; trivial.
econstructor 2; eauto.
apply join_com; trivial.
simpl in H.
destruct l.
inv H.
inv H0.
destruct (IHn _ _ _ _ _ _ H H7 H1) as [phic [l' [? [? ?]]]].
simpl. rewrite H0.
destruct (join_assoc _ _ _ _ _ (join_com _ _ _ H5) H2) as [f [? ?]].
exists f; exists (a::l'); repeat split; auto.
econstructor 2; eauto.
apply join_com.
trivial.
Qed.

Lemma join_list_join_nth `{sepalg A}:
 forall n l phin phi phia phib phi2,
  nth_error l n = Some phin -> join_list phi l phi2 -> join phi2 phia phib ->
           exists phic, exists l',
                 join phin phia phic /\ replace l n phic = Some l' /\ join_list phi l' phib.
Proof.
intros A saA.
induction n; intros.
inv H.
destruct l; try discriminate.
inv H0.
inv H3.
simpl.
destruct (join_list_assoc2 H7 (join_com _ _ _ H1)) as [f [? ?]].
destruct (join_assoc _ _ _ _ _ H5 (join_com _ _ _ H)) as [g [? ?]].
exists g; exists (g::l); repeat split; auto.
econstructor 2; eauto.
destruct l; simpl in H.
inv H.
inv H0.
simpl.
destruct (IHn _ _ _ _ _ _ H H7 H1) as [phic  [l' [? [? ?]]]].
exists phic. rewrite H2. exists (a::l').
repeat split; auto.
econstructor 2; eauto.
Qed.

(* END ecm/sepalg_list.v *)

(* New material *)

(* These kinds of facts are suprisingly hard to prove *)
Lemma join_list_cons `{sepalg A} : forall a L b b' b'',
  join_list a L b ->
  join b b' b'' ->
  join_list a (b' :: L) b''.
Proof.
  induction 1; intros.
  apply fold_rel_cons with b''.
  trivial.
  constructor.
  spec IHfold_rel H2.
  inversion IHfold_rel ; subst.
  apply join_com in H0.
  destruct (join_assoc _ _ _ _ _ H0 H6) as [? [? ?]].
  apply fold_rel_cons with x; trivial.
  apply fold_rel_cons with w3.
  apply join_com; trivial.
  trivial.
Qed.

Lemma join_list_cons2 `{sepalg A} : forall a L b a' a'',
  join_list a L b ->
  join a' a'' a ->
  join_list a' (a'' :: L) b.
Proof.
  intros.
  apply fold_rel_cons with a; trivial.
Qed.

Lemma join_list_uncons `{sepalg A} : forall a L b' b'',
  join_list a (b' :: L) b'' ->
  exists b,
    join_list a L b /\
    join b b' b''.
Proof.
  intros ? ? ? ?.
  revert a.
  induction L; intros.
  inversion H0; subst.
  inversion H6; subst.
  exists a.
  split; trivial.
  constructor.
  inversion H0; subst.
  spec IHL w1 a b'' H6.
  destruct IHL as [b [? ?]].
  apply join_com in H4.
  destruct (join_list_assoc1 H4 H6) as [f [? ?]].
  exists f.
  split; trivial.
  apply join_com.
  trivial.
Qed.

Lemma join_list_eq `{sepalg A} : forall a L a' a'',
  join_list a L a' ->
  join_list a L a'' ->
  a' = a''.
Proof.
  intros until L.
  revert a.
  induction L; intros.
  inversion H0.
  inversion H1.
  congruence.
  inversion H0.
  inversion H1.
  subst.
  generalize (join_eq _ _ _ _ H5 H11); intro.
  subst.
  spec IHL w3 a' a''.
  spec IHL H7 H13.
  trivial.
Qed.

Lemma same_silhouette_prod1 `{sepalg A} `{sepalg B} : forall (a a' : A),
  same_silhouette a a' ->
  forall (b : B),
  @same_silhouette _ (sa_prod _ _ ) (a, b) (a', b).
Proof.
  repeat intro.
  destruct c.
  split; intros [[? ?] ?]; destruct H2;
  spec H1 a0;
  destruct H1.
  assert (joins a0 a) by (exists a1; trivial).
  spec H1 H5.
  destruct H1 as [a'' ?].
  exists (a'', b1); split; trivial.
  assert (joins a0 a') by (exists a1; trivial).
  spec H4 H5.
  destruct H4 as [a'' ?].
  exists (a'', b1); split; trivial.
Qed.

Lemma same_silhouette_prod2 `{sepalg A} `{sepalg B} : forall (b b' : B),
  same_silhouette b b' ->
  forall (a : A),
  @same_silhouette _ (sa_prod _ _ ) (a, b) (a, b').
Proof.
  repeat intro.
  destruct c.
  split; intros [[? ?] ?]; destruct H2;
  spec H1 b0;
  destruct H1.
  assert (joins b0 b) by (exists b1; trivial).
  spec H1 H5.
  destruct H1 as [b'' ?].
  exists (a1, b''); split; trivial.
  assert (joins b0 b') by (exists b1; trivial).
  spec H4 H5.
  destruct H4 as [b'' ?].
  exists (a1, b''); split; trivial.
Qed.

Lemma join_list_same_silhouette `{sepalg A}: forall a aL b,
  join_list a aL b ->
  forall a',
  same_silhouette a a' ->
  exists b',
  same_silhouette b b' /\ join_list a' aL b'.
Proof.
  do 4 intro.
  revert a.
  induction aL; intros.
  inversion H0; subst.
  exists a'.
  split; trivial.
  constructor.
  inversion H0; subst.
  copy H1.
  spec H2 a.
  destruct H2.
  assert (joins a a0) by (exists w1; apply join_com; trivial).
  spec H2 H4.
  destruct H2 as [new_w1 ?].
  apply join_com in H2.
  generalize (same_silhouette_join _ _ _ H1 H5 H2); intro.
  spec IHaL w1 b H7 new_w1 H6.
  destruct IHaL as [b' [? ?]].
  exists b'.
  split; trivial.
  apply join_list_cons2 with new_w1; trivial.
Qed.

Lemma sub_silhouette_prod1 `{sepalg A} `{sepalg B} : forall (a a' : A),
  sub_silhouette a a' ->
  forall (b : B),
  @sub_silhouette _ (sa_prod _ _ ) (a, b) (a', b).
Proof.
  repeat intro.
  destruct c.
  destruct H2 as [[? ?] [? ?]];
  spec H1 a0;
  unfold fst, snd in *.
  destruct H1.
  exists a1; trivial.
  exists (x, b1); split; trivial.
Qed.

Lemma sub_silhouette_prod2 `{sepalg A} `{sepalg B} : forall (b b' : B),
  sub_silhouette b b' ->
  forall (a : A),
  @sub_silhouette _ (sa_prod _ _ ) (a, b) (a, b').
Proof.
  repeat intro.
  destruct c.
  destruct H2 as [[? ?] [? ?]];
  spec H1 b0;
  unfold fst, snd in *.
  destruct H1.
  exists b1; trivial.
  exists (a1, x); split; trivial.
Qed.

Lemma sub_silhouette_join `{sepalg A} : forall phi phi' phiy phiz phiz',
  sub_silhouette phi phi' ->
  join phi phiy phiz ->
  join phi' phiy phiz' -> 
  sub_silhouette phiz phiz'.
Proof.
  intros.
  intros ? [? ?].
  apply join_com in H3.
  destruct (join_assoc _ _ _ _ _ H2 H3) as [? [? ?]].
  spec H0 x0.
  detach H0.
  2: exists x; apply join_com; trivial.
  destruct H0.
  apply join_com in H4.
  destruct (join_assoc _ _ _ _ _ H4 H0) as [? [? ?]].
  apply join_com in H6.
  apply (join_eq _ _ _ _ H1) in H6.
  subst x2.
  exists x1.
  trivial.
Qed.

Lemma join_list_sub_silhouette `{sepalg A}: forall a aL b,
  join_list a aL b ->
  forall a',
  sub_silhouette a' a ->
  exists b',
  sub_silhouette b' b /\ join_list a' aL b'.
Proof.
  do 4 intro.
  revert a.
  induction aL; intros.
  inversion H0; subst.
  exists a'.
  split; trivial.
  constructor.
  inversion H0; subst.
  copy H1.
  spec H2 a.
  detach H2.
  2: exists w1; apply join_com; trivial.
  destruct H2 as [new_w1 ?].
  apply join_com in H2.
  generalize (sub_silhouette_join _ _ H1 H2 H5); intro.
  spec IHaL w1 b H7 new_w1 H3.
  destruct IHaL as [b' [? ?]].
  exists b'.
  split; trivial.
  apply join_list_cons2 with new_w1; trivial.
Qed.
