(* CSL with Barriers, v0.1
 * 
 * It is our intention to release the files under a BSD-style license shortly
 * after some badly-needed cleanup work,  
 * 
 * Until then:
 * (c) 2010 Aquinas Hobor, Cristian Gherghina, and Christopher Chak
 * All Rights Reserved
 *
 * Permission granted to use these proof scripts only for the purpose of 
 * evaluating "Barriers in Concurrent Separation Logic."
 *)

Require Import SLB_Base.
Require Import SLB_Lang.
Require Import SLB_BarDefs.
Require Import SLB_CLang.
Require Import SLB_SSem.
Require Import SLB_Sem.
Require Import SLB_OSem.

Definition Gamma : Type := list BarDef.

Definition can_step c :=
  exists c', seq_step c c'.

Definition safely_halted (c : config) :=
  snd c = Unit.

Definition safe_ora (st: state) (o : O.t) (k: cmd) : Prop :=
    forall st' o' k', seq_stepstar (st, o, k) (st', o', k') ->
    can_step (st', o', k') \/
    safely_halted (st', o', k').

Definition safe (G : Gamma) (st: state) (k: cmd) : Prop :=
  forall o (ora_ok : O.ora_barDefs o G),
    safe_ora st o k.

Definition guards (G : Gamma) (P : assert) (k : cmd) : Prop :=
  forall st, (sepconj P pred_brk) st -> safe G st k.
Notation "P [] k " := (guards P k)(at level 40).

Definition Htuple (G : Gamma) (P : assert) (C : cmd) (Q : assert) : Prop :=
  forall F k, 
    closedWith F C -> 
      guards G (sepconj F Q) k -> 
      guards G (sepconj F P) (Seq C k).

Lemma safe_ora_step_safe_ora: forall st c o st' o' c',
  safe_ora st o c ->
  seq_step (st, o, c) (st', o', c') ->
  safe_ora st' o' c'.
Proof.
  repeat intro.
  spec H st'0 o'0 k'.
  apply H.
  apply SeqStepStarS with (st', o', c'); auto.
Qed.

Lemma Unit_safe_ora : forall st o,
  safe_ora st o Unit.
Proof.
  repeat intro.
  inversion H.
  right.
  red.
  trivial.
  inversion H0.
Qed.

Lemma Unit_safe : forall G st,
  safe G st Unit.
Proof.
  do 4 intro.
  apply Unit_safe_ora.
Qed.

Lemma Unit_guards : forall G P, 
  guards G P Unit.
Proof.
  unfold guards.
  intros.
  apply Unit_safe.
Qed.

Lemma Htuple_unit_guards : forall G P c Q,
  Htuple G P c Q ->
  guards G P (Seq c Unit).
Proof.
  intros.
  spec H emp Unit.
  rewrite sc_comm in H.
  rewrite emp_simpl in H.
  rewrite sc_comm in H.
  rewrite emp_simpl in H.
  apply H.
  apply emp_closedWith.
  apply Unit_guards.
Qed.

(* Hoare Rules *)

Module Type HRULES.

(*1assignment*)
Axiom Htuple_Assignment: forall G v e val,
  Htuple G  (eval_assert e val) (Assign v e) (eval_assert (expVar v) val).

(*2sequential*)
Axiom Htuple_Sequential: forall G cmd cmd' P Q R,
  Htuple G P cmd Q ->
  Htuple G Q cmd' R ->
  Htuple G P (Seq cmd cmd') R.

Definition derives (P P' : assert) : Prop :=
  forall st, P st -> P' st.

Axiom Htuple_Consequence : forall G P P' Q Q' c,
  Htuple G P c Q ->
  derives P' P ->
  derives Q Q' ->
  Htuple G P' c Q'.

Axiom Htuple_Frame : forall G P c Q F,
  Htuple G P c Q ->
  closedWith F c ->
  Htuple G (sepconj F P) c (sepconj F Q).

(*3if*)
Axiom Htuple_If: forall G cmdT cmdE  v P Q,
Htuple G (sepconj P (isTru_assert v)) cmdT Q ->
Htuple G (sepconj P (not (isTru_assert v))) cmdE Q ->
(* ------------------------------------------------------------*)
Htuple G P (If v cmdT cmdE) Q.

(*5skip*)
Axiom Htuple_Unit : forall G P,
  Htuple G P Unit P.

(*6free*)
Axiom Htuple_Free : forall G exp x,
  Htuple G (points2 exp pfullshare x)
                  (Free exp)
                   emp.

(*7new*)
Axiom Htuple_New : forall G v val e,
  Htuple G (eval_assert e val)
                  (New v e)
                  (points2 (expVar v) pfullshare (expConst val)).

(*8store*)
Axiom Htuple_Store : forall G e1 olde2 e2,
  Htuple G (points2 e1 pfullshare olde2)
                   (Store e1 e2)
                   (points2 e1 pfullshare e2).

(*9load*)
Axiom Htuple_Load : forall G e e' v vale vale' someshare,
  Htuple G (sepconj (points2 e someshare e') 
                   (sepconj (eval_assert e vale) (eval_assert e' vale')))
                  (Load v e)
                  (sepconj (points2 (expConst vale) someshare (expConst vale'))
                    (eval_assert (expVar v) vale')).

(*4while*)
Axiom Htuple_While: forall G cmd' I (v : expr),
(Htuple G (sepconj I (isTru_assert v)) cmd' I) ->
(* ------------------------------------------------------------*)
Htuple G I (While v cmd') (sepconj I (not (isTru_assert v))).

Axiom Htuple_Barrier : forall G bn bd cs dir mv pre post,
  nth_error G bn = Some bd ->
  lookup_move bd cs dir mv pre post ->
  Htuple G pre (Barrier bn) post.

End HRULES.

