(* CSL with Barriers, v0.1
 * 
 * It is our intention to release the files under a BSD-style license shortly
 * after some badly-needed cleanup work,  
 * 
 * Until then:
 * (c) 2010 Aquinas Hobor, Cristian Gherghina, and Christopher Chak
 * All Rights Reserved
 *
 * Permission granted to use these proof scripts only for the purpose of 
 * evaluating "Barriers in Concurrent Separation Logic."
 *)

Require Import SLB_Base.
Require Import SLB_Lang.
Require Import SLB_BarDefs.
Require Import SLB_CLang.
Require Import SLB_SSem.
Require Import SLB_Sem.
Require Import SLB_OSem.
Require Import SLB_HRules.

Module HRules <: HRULES.

(* Add safe halt states if needed *)

Lemma Htuple_Barrier : forall G bn bd cs dir mv pre post,
  nth_error G bn = Some bd ->
  lookup_move bd cs dir mv pre post ->
  Htuple G pre (Barrier bn) post.
Proof.
  repeat intro.
  remember (st, o, Seq (Barrier bn) k).
  remember (st', o', k').
  induction H4; subst.
  inversion Heqp0; subst.
  left.
  destruct o' as [sch bp othds].
  destruct sch.
  eexists.
  apply SeqStep_BarLoop.
  constructor 1.
  trivial.
  LEM (exists cs, O.Ora2CM (O.ora (n :: sch) bp othds) st' (Running (Seq (Barrier bn) k)) n cs).
  destruct H4 as [cst ?].
  assert (exists cst', O.CM.conc_step cst cst').
    generalize (O.stdy_match_lookup _ _ ora_ok bn); intro.
    rewrite H in H5.
    simpl in H5.
    eapply O.O2CM_can_step; eauto.
  destruct H5 as [cst' ?].
  LEM (exists cst'', O.stepOthers n cst' cst'').
  destruct H6 as [cst'' ?].
  destruct (O.Ora2CM_threadn _ _ _ _ _ H4) as [t0 ?].
  destruct (O.CM.cs_threads_preserved_barrier _ _ _ _ _ _ _ _ H7 H5).
  inversion H4; subst.
  trivial.
  destruct H8 as [h' [bm' [dir' [mv' ?]]]].
  destruct (O.thread_so_constants2 _ _ _ _ _ _ _ _ _ _ H8 H6) as [h'' [bm'' ?]].
  destruct (O.Ora2CM_tot2 _ _ _ _ _ _ H9) as [o' [st'' ?]].
  exists (st'', o', k).
  apply SeqStep_BarStep.
  eapply O.os_stepgood; eauto.
  simpl.
  reflexivity.
  destruct H8 as [h' [bs' ?]].
  generalize (O.thread_so_constants1 _ _ _ _ _ _ _ H8 H6); intro.
  destruct (O.Ora2CM_tot2 _ _ _ _ _ _ H9) as [o' [st'' ?]].
  exists (st'', o', k).
  apply SeqStep_BarStep.
  eapply O.os_stepgood; eauto.
  simpl.
  reflexivity.
  eexists.
  apply SeqStep_BarLoop.
  eapply O.os_steploop; eauto.
  simpl.
  reflexivity.  
  eexists.
  apply SeqStep_BarLoop.
  apply O.os_badora2 with n sch; trivial.
  
  (* stage 2 *)
  inversion H4; subst.
  2: apply IHseq_stepstar; auto.
  clear IHseq_stepstar.
  spec H2 st'0.
  detach H2.
  assert (O.ora_barDefs o'0 G) by (eapply O.barDefs_ostep_preserved; eauto).
  spec H2 o'0 H6 st' o' k'.
  apply H2.
  apply O.ostep_preserves_ctl in H11.
  subst k.
  trivial.
  clear -H11 H3 H0 H H1 ora_ok.
  rewrite sc_comm in *.
  rewrite (sc_comm F _).
  rewrite (sc_comm F _) in H3.
  eapply O.OStep_Barrier; eauto.
Qed.

Print Assumptions Htuple_Barrier.

(*The rules*)

(*4while*)

(*auxiliary n versions for safe guards and htuple*)

Definition safen (G : Gamma)(n:nat) (st: state) (k: cmd) : Prop :=
  forall o (ora_ok : O.ora_barDefs o G),
    forall st' o' k', seq_stepn n (st, o, k) (st', o', k') ->
    can_step (st', o', k') \/ safely_halted (st', o', k').

Definition guardsn (G : Gamma) (n:nat)(P : assert) (k : cmd) : Prop :=
  forall st, (sepconj P pred_brk) st -> safen G n st k.


Definition Htuplen (G : Gamma)(n:nat) (P : assert) (C : cmd) (Q : assert) : Prop :=
  forall F k, 
    closedWith F C -> 
      guards G (sepconj F Q) k -> 
      guardsn G n (sepconj F P) (Seq C k).

Lemma stepstar_stepn_eq : forall c c' ,
	(seq_stepstar c c') = (exists n, seq_stepn n c c').
Proof.
  intros.
  apply prop_ext.
  split; intro.
  apply stepstar_stepn.
  trivial.
  destruct H.
  apply stepn_stepstar with x.
  trivial.
Qed.

Lemma htuple_htuplen : forall G P C Q ,
      (Htuple G P C Q) = (forall n, Htuplen G n P C Q).
Proof.
   intros.
   apply prop_ext; split.
   repeat intro.
   spec H F k H0.
   spec H H1.
   spec H st H2.
   spec H o ora_ok st' o' k'.
   apply H.
   apply stepn_stepstar in H3.
   trivial.
   repeat intro.
   rewrite stepstar_stepn_eq  in H3.
   destruct H3.
   spec H x F k H0 H1.
   spec H st H2 o ora_ok st'.
   spec H o' k' H3 ;trivial.   
Qed.

Lemma Htuple_While_n: forall G cnt cmd' I (v : expr),
Htuplen G cnt (sepconj I (isTru_assert v)) cmd' I ->
(* ------------------------------------------------------------*)
Htuple G I (While v cmd') (sepconj I (not (isTru_assert v))).
Proof.
  repeat intro.
  apply stepstar_stepn in H3.
  destruct H3 as [x ?].
  remember x.
  assert (n <= x) by omega.
  clear Heqn.
  revert n H4 G I F st H2 o ora_ok v cmd' H0 cnt H k H1 st' o' k' H3.
  induction x.
  left.
  intros.
  assert (n = 0) by omega.
  subst n.
  inversion H3;subst.
  LEM (evalExpr (st_stk st') v = truE).
  exists (st', o', Seq cmd' (Seq (While v cmd') k)).
  destruct st' as [[? ?] ?].
  apply SeqStep_WhileTru; trivial.
  exists (st', o', k).
  destruct st' as [[? ?] ?].
  apply SeqStep_WhileBrk; trivial.
  
  (* stage 2 *)

  intros.
  inversion H3; subst.
  left.
(* blah *)
  LEM (evalExpr (st_stk st') v = truE).
  exists (st', o', Seq cmd' (Seq (While v cmd') k)).
  destruct st' as [[? ?] ?].
  apply SeqStep_WhileTru; trivial.
  exists (st', o', k).
  destruct st' as [[? ?] ?].
  apply SeqStep_WhileBrk; trivial.
(* end blah *)
  rename n0 into n.
  assert (n <= x) by omega.
  clear H4.
  inversion H5; subst.

  Focus 2.
(* stage 2b *)
  spec H1 (stk, h, bs).
  detach H1.
  spec H1 o ora_ok st' o' k'.
  detach H1;trivial.
  apply stepn_stepstar in H6; trivial.
  rewrite (sc_comm I).
  rewrite <- sc_assoc.
  rewrite (sc_comm F).
  repeat (rewrite sc_assoc in *).
  exists (stk, empty_heap, empty_bm).
  exists (stk, h, bs).
  split.
  split.
  split.
  split; trivial.
  apply join_emptyheap.
  unfold snd; apply join_emptybm.
  split; trivial.
   intro.
  destruct H1.
  unfold st_stk, fst in H1.
  rewrite H1 in H13.
  contradiction H13.
  trivial.
  
  (*Stage 2a*)
  clear H3.
  assert (closedWith F cmd').
    clear -H0. 
    unfold closedWith.
    intro.
    spec H0 x.
    intros.
    detach H0.
    spec H0 st v0.
    spec H0 H1; trivial.
    simpl;trivial.
  spec H F (Seq (While v cmd') k).
  spec H H3.
  admit.


(*
  detach H.
  unfold guardsn in H.
  spec H (stk,h,bs).
  detach H.
  unfold safen in H.
  spec H o.
  spec H ora_ok.
  spec H st' o' k'.
  detach H;trivial.
  clear H5.
  apply  stepn_split in H6.
  rewrite <- sc_assoc.
  rewrite <- sc_comm.
  rewrite <- sc_assoc.
  exists (stk, h, bs).
  exists (stk,empty_heap, empty_bm).
  split.
  split; unfold fst,snd.
  split; unfold fst,snd.
  split;trivial.
  apply join_com.
  apply join_emptyheap.
  apply join_com.
  apply join_emptybm.
  split.
  destruct H2.
  destruct H.
  exists x1.
  exists x0.
  split;trivial.
  destruct H.
  destruct H2.
  apply join_com;trivial.
  destruct H as [? [? ? ]].
  split; trivial.
  split.
  unfold st_stk,fst;trivial.
  unfold st_hp, st_bsm,fst,snd.
  split; trivial.
  unfold guards.
  
  repeat intro.
  apply stepstar_stepn in H4.
  destruct H4.
  assert (x<0).
  spec IHx x0.
  detach IHx.
  spec IHx G I F st H.
  spec IHx o0 ora_ok0.
  spec IHx v cmd'.
  spec IHx H3.
  detach IHx.
  spec IHx k H1.
  spec IHx st'0 o'0 k'0 H4;trivial.
  unfold Htuple.
*)
Qed.

Lemma Htuple_While: forall G cmd' I (v : expr),
Htuple G (sepconj I (isTru_assert v)) cmd' I ->
(* ------------------------------------------------------------*)
Htuple G I (While v cmd') (sepconj I (not (isTru_assert v))).
Proof.
  do 5 intro.
  rewrite htuple_htuplen .
  intro n.
  rewrite htuple_htuplen  in H.
  spec H n.
  apply  Htuple_While_n in H.
  rewrite htuple_htuplen  in H.
  spec H n;  trivial.
Qed.

(*1assignment*)
Lemma Htuple_Assignment: forall G v e val,
(* ------------------------------------------------------------*)
  Htuple G  (eval_assert e val) (Assign v e) (eval_assert (expVar v) val).
Proof.
  repeat intro.
  inversion H2;subst.
  left.
  unfold can_step.
  destruct st' as [[? ?] ?].
  exists ((upd_stack s v val, h, b), o', k).
  apply SeqStep_Assign.
  rewrite (sc_comm F) in H1.
  rewrite sc_assoc in H1.
  inv_sc H1.
  simpl in H1, H3.
  subst.
  unfold eval_assert,st_stk,st_hp,fst,snd in H6.
  simpl in H6.
  destruct H6.
  rewrite H1.
  trivial.

  inversion H3; subst.
  spec H0 (upd_stack stk v (evalExpr stk e), hp,bs).
  detach H0.
  spec H0 o ora_ok st' o' k'.
  auto.
  
  inv_sc H1.
  simpl in H0, H1.
  subst s0 s.
  exists (upd_stack stk v (evalExpr stk e), h,b).
  exists (upd_stack stk v (evalExpr stk e), h0,b0).
  split.
  split;trivial. 
  split;trivial.
  split;trivial.
  split.
  inv_sc H7.
  simpl in H0, H1.
  subst s0 stk.
  exists (upd_stack s v (evalExpr s e), h1,b1).
  exists (upd_stack s v (evalExpr s e), h2,b2).
  split.
  split;trivial.
  split;trivial.
  split;trivial.
  split;trivial.
  spec H v.
  detach H.
  2: simpl; trivial.
  spec H (s, h1,b1) (evalExpr s e).
  unfold st_upd_stk_var,st_stk,fst,st_upd_stk,snd,fst in H.
  apply H.
  trivial.
  unfold eval_assert,st_hp,st_stk,st_bsm, fst,snd in *.
  destruct H11 as [? [? ?]].
  split;trivial.
  simpl.
  rewrite H0.
  apply upd_stk_eq.
  split;trivial.
  unfold pred_brk in *.
  simpl in *.
  trivial.
Qed.

(*2sequential*)
Lemma Htuple_Sequential: forall G cmd cmd' P Q R,
  Htuple G P cmd Q ->
  Htuple G Q cmd' R ->
(* ------------------------------------------------------------*)
  Htuple G P (Seq cmd cmd') R.
Proof.
  repeat intro.
  inversion H4; subst.
  left.
  exists (st', o', Seq cmd (Seq cmd' k)).
  constructor.
  inversion H5; subst.
  spec H0 F k.
  detach H0.
  spec H0 H2.
  spec H F (Seq cmd' k).
  detach H.
  spec H H0.
  spec H st H3.
  spec H o ora_ok st' o' k'.
  apply H; trivial.
  repeat intro.
  spec H1 x.
  simpl in H1.
  auto.
  repeat intro.
  spec H1 x.
  simpl in H1.
  auto.
Qed.

(*3if*)
Lemma Htuple_If: forall G cmdT cmdE  v P Q ,
Htuple G (sepconj P (isTru_assert v)) cmdT Q ->
Htuple G (sepconj P (not (isTru_assert v))) cmdE Q ->
(* ------------------------------------------------------------*)
Htuple G P (If v cmdT cmdE) Q.
Proof.
  repeat intro.
  inversion H4; subst.
  left.
  destruct st' as [[? ?] ?].
  destruct (classic (evalExpr s v = truE)).
  exists ((s, h, b), o', Seq cmdT k).
  constructor; trivial.
  exists ((s, h, b), o', Seq cmdE k).
  constructor; trivial.
  
  inversion H5; subst.
  spec H F k.
  detach H.
  spec H H2.
  spec H (stk, h, bs).
  detach H.
  eapply H; eauto.
  inv_sc H3.
  simpl in H, H3.
  subst s0 stk.
  exists (s, h0, b).
  exists (s, h1, b0).
  split.
  split; trivial.
  split; trivial.
  split; trivial.
  split; trivial.
  rewrite <- sc_assoc.
  exists (s, h0, b).
  exists (s, empty_heap, empty_bm).
  split.
  split; unfold fst,snd.
  split; unfold fst,snd.
  split;trivial.
  apply join_com.
  apply join_emptyheap.
  apply join_com.
  apply join_emptybm.
  split; trivial.
  split; trivial.
  split; trivial.
  repeat intro.
  spec H1 x.
  detach H1; auto.
  simpl.
  auto.

  spec H0 F k.
  detach H0.
  spec H0 H2.
  spec H0 (stk, h,bs).
  detach H0.
  eapply H0; eauto.
  inv_sc H3.
  simpl in H0, H3.
  subst s0 stk.
  exists (s, h0,b).
  exists (s, h1,b0).
  split.
  split; trivial.
  split; trivial.
  split; trivial.
  split; trivial.
  rewrite <- sc_assoc.
  exists (s, h0, b).
  exists (s, empty_heap,empty_bm).

  split.
  split;unfold fst, snd.
  split;unfold fst, snd.
  split; trivial.
  apply join_com.
  apply join_emptyheap.
  apply join_com.
  apply join_emptybm.

  split; trivial.
  intro.
  destruct H0.
  simpl in H0.
  contradiction.
  repeat intro.  
  spec H1 x.
  detach H1; auto.
  simpl.
  auto.
Qed.

(*5skip*)
Lemma Htuple_Unit : forall G P,
(* ------------------------------------------------------------*)
  Htuple G P Unit P.
Proof.
  repeat intro.
  inversion H2;subst.
  left.
  exists (st', o', k).
  constructor.
  inversion H3;subst.
  unfold guards in H0.
  spec H0 st.
  detach H0.
  unfold safe in H0.
  spec H0 o ora_ok st' o' k'.
  detach H0.
  trivial.
  trivial.
  trivial.
Qed.

Definition derives (P P' : assert) : Prop :=
  forall st, P st -> P' st.


Lemma Htuple_Consequence : forall G P P' Q Q' c,
  Htuple G P c Q ->
  derives P' P ->
  derives Q Q' ->
  Htuple G P' c Q'.
Proof.
  repeat intro.
  unfold Htuple in H.
  spec H  F k H2.
  detach H.
  unfold guards in H.
  spec H st.
  detach H.
  spec H o ora_ok st' o' k'.
  spec H H5.
  trivial.  
  rewrite <- sc_comm.
  rewrite <- sc_assoc.
  rewrite <- sc_comm.
  rewrite sc_assoc in H4.
  rewrite sc_comm in H4.
  rewrite sc_assoc in H4.
  destruct H4.
  destruct H.
  exists  x.
  exists  x0.
  split; try tauto.
  split; try tauto.
  apply H0; try tauto.
  unfold guards in *.
  intro.
  spec H3 st0.
  intro.
  apply H3.
  destruct H.
  destruct H.
  exists x.
  exists x0.
  destruct H.
  destruct H6.
  split; try auto.
  split; try auto.
  destruct H6.
  destruct H6.
  exists x1.
  exists x2.
  destruct H6.
  destruct H8.
  split;try tauto.
  split; try auto.
Qed.

Lemma Htuple_Frame : forall G P c Q F,
  Htuple G P c Q ->
  closedWith F c ->
  Htuple G (sepconj F P) c (sepconj F Q).
Proof.
  repeat intro.
  unfold Htuple in H.
  spec H (sepconj F F0) k.
  detach H.
  detach H.
  unfold guards in *.
  spec H st.
  detach H.
  spec H o ora_ok st' o' k'.
  spec H H4.
  trivial.
  destruct H3.
  destruct H.
  destruct H.
  destruct H3.
  exists x.
  exists x0.
  split; try tauto.
  split; try tauto.
  rewrite <- sc_comm.
  rewrite <- sc_assoc.
  rewrite <-sc_comm.
  destruct H3.
  destruct H3.
  destruct H3.
  destruct H6.
  exists x1.
  exists x2.
  split; try auto.
  split; try auto.
  rewrite sc_comm in H7.
  trivial.
  unfold guards in *.
  intro.
  spec H2 st0.
  intro.
  apply H2.
  destruct H.
  destruct H.
  exists x.
  exists x0.
  destruct H.
  destruct H5.
  split; trivial.
  split; try auto.
  destruct H5.
  destruct H5.
  destruct H5.
  destruct H7.
  destruct H7.
  destruct H7.
  destruct H7.
  destruct H9.
  exists x4.
  apply join_com in H7.
  generalize(join_assoc _ _ _ _ _ H7 H5).
  intro.
  destruct X.
  destruct a.
  exists x5.
  split;trivial.
  split; trivial.
  exists x3.
  exists x2.
  split;trivial.
  split;trivial.
  unfold closedWith in *.
  repeat intro.
  destruct H5.
  destruct H5.
  spec H0 x H x0 v.
  spec H1 x H x1 v.
  unfold sepconj.
  destruct H5.
  destruct H6.
  spec H0 H6.
  spec H1 H7.  
  exists(st_upd_stk_var x0 x v).
  exists(st_upd_stk_var x1 x v).
  split; try tauto.
  clear -H5.
  destruct x0 as[[? ?] ?].
  destruct x1 as[[? ?] ?].
  destruct st0 as [[? ?] ?].
  destruct H5.
  unfold fst,snd in *.
  split; unfold st_upd_stk_var,st_upd_stk, st_stk,fst,snd;trivial.
  destruct H.
  unfold fst,snd in *.
  split; unfold fst,snd;trivial.
  destruct H; subst.
  split;trivial.
Qed.

(*6free*)
Lemma Htuple_Free : forall G exp x,
(* ------------------------------------------------------------*)
  Htuple G (points2 exp pfullshare x)
                  (Free exp)
                   emp.
Proof.
  repeat intro.
  inversion H2; subst.
  left.
  destruct st' as[[? ?] ?].
  destruct H1.  destruct H1.  destruct H1.
  destruct H3.  destruct H3.  destruct H3.  destruct H3.
  destruct H5.
  unfold points2 in H6.
  remember (evalExpr (st_stk x3) exp).
  icase v.
  exists (s, upd_heap h a None ,b, o', k).
  destruct h.
  destruct b0.
  destruct x3 as[[? ?] ?].
  destruct x2 as[[? ?] ?].
  destruct x1 as[[? ?] ?].
  destruct x0 as[[? ?] ?].
  destruct H1.
  destruct H3.
  destruct h.
  destruct H1.
  destruct H3.
  unfold st_stk,st_brk,st_hf,st_hp,st_bsm,fst,snd in *.
  destruct H1. destruct H3. 
  subst.
  destruct h2.
  destruct h1.
  destruct h0.
  destruct H6. destruct H3.
  assert (a<n).  
    unfold pred_brk,st_brk,st_hf,st_hp,st_bsm,fst,snd in H4.
    destruct H4.
    destruct H11.
    destruct H11.
    destruct H9.
    unfold fst, snd in *.
    icase b6.
    icase b5.
    inversion H11;subst.
    inversion H9;subst.
    destruct (classic (a < n)); trivial.
    assert (a >= n) by omega.
    spec H12 a.
    destruct H12.
    spec H12 H4.
    destruct H10;unfold fst,snd in *.
    destruct (join_assoc _ _ _ _ _ H15 H13) as [?[? ?]].
    spec H16 a.
    clear -H3 H12 H16.
    inversion H16; subst.
    rewrite <- H0 in H3.
    inversion H3;subst.
    rewrite <- H1 in H12.
    inversion H12;subst.
    unfold res_builder in H3, H12.
    unfold pfullshare, mk_pshare in *.
    rewrite <- H in *.
    rewrite <- H0 in *.
    elimtype False.
    destruct H16.
    inversion H3.
    inversion H12.
    inversion H3.
    inversion H12.
    subst.
    destruct z0.
    clear -H4.
    simpl in *.
    destruct x2; simpl in *.
    inversion H4;subst.
    unfold fst,snd in *.
    change Share.top with (projT1 pfullshare) in H0.
    pfullshare_join.
  apply SeqStep_Free with n a (evalExpr s x).
  auto.
  simpl.
  trivial.
  trivial.
  unfold dref_heap, snd.
  destruct H9.
  destruct H10.
  destruct H4.
  destruct H14.
  unfold st_brk,st_hf,st_hp,fst,snd in *.
  destruct H14.
  icase b6.
  icase b5.
  inversion H14;subst.
  inversion H9;subst.
  spec H15 a.
  destruct H15.
  spec H15 H11.
  spec H13 a.
  spec H12 a.
  destruct (join_assoc _ _ _ _ _ H13 H12) as [?[? ?]].
  rewrite H15 in H16.
  rewrite H3 in H16.
  inversion H16;subst.
  inversion H17;subst.
  trivial.
  unfold pjoin in H21.
  destruct x0.
  destruct z.
  destruct x0.
  destruct x1.
  destruct H21.
  simpl in H19, H21.
  destruct H19.
  subst v v0.
  assert (t0 = Share.top).
    LEM (identity t).
    generalize (H19 _ _ H21).
    auto.
    change t with (projT1 (mk_pshare t H19)) in H21.
    change Share.top with (projT1 pfullshare) in H21.
    pfullshare_join.
  subst t0.
  unfold res_builder, pfullshare, mk_pshare.
  replace (@n2_identity_prod value Share.t sa_value Share.sa (evalExpr s x) Share.top top_share_nonidentity) with n1 by apply proof_irr; trivial.
  trivial.

  destruct x0 as [[? [? ?]] ?].
  destruct x1 as [[? [? ?]] ?].
  destruct x3 as [[? [? ?]] ?].
  destruct H1 as [[[? ?] [? ?]]?]; unfold fst,snd in *.
  subst.
  destruct H6 as[? [? [? ?]]].
  unfold st_stk, st_brk, st_hf,st_hp,st_bsm,fst,snd in *.

  simpl in *.
  unfold ojoin in H8.
  icase b0; icase b2.
  inversion H4;subst.
  unfold st_stk,st_brk, st_hf,st_hp, fst,snd in *.
  destruct H13.
  destruct H1.
  inversion H1.


  inversion H3;subst.
  spec H0 (stk, (upd_heap hp addr None), bs).
  detach H0.
  spec H0 o ora_ok st' o' k'.
  apply H0; trivial.
  rewrite sc_comm in *.
  rewrite <- sc_assoc in *.
  inv_sc H1.
  subst s0 stk.
  exists (s, h, b).
  exists (s, upd_heap h0 addr None, b0).
  split.
  split; unfold fst,snd;trivial.
  split; unfold fst,snd.
  split; trivial.
  simpl in H5.
  unfold join_prod in H5.
  destruct H5.
  destruct h. destruct h0. destruct hp. unfold fst, snd in *.
  unfold dref_heap,snd in H13.
  simpl.
  split;trivial.
  unfold upd_heap, snd.
  simpl.
  intro.
  spec H1 x0.
  case (eq_nat_dec x0 addr).
  intro.
  subst x0.
  unfold points2 in H8.
  unfold st_stk,fst in H8.
  icase (evalExpr s exp).
  destruct H8.
  destruct H8.
  unfold st_brk, st_hp, st_hf,fst,snd,st_hp,st_bsm,fst,snd in *.
  inversion H9;subst.
  clear -H13 H8 H1.
  unfold pfullshare , res_builder in H13,H8.
  rewrite H8, H13 in H1.
  unfold mk_pshare in *.
  inversion H1;subst.
  constructor.
  destruct x0.
  simpl in H3.
  destruct H3.
  simpl in H2.
  destruct x0.
  simpl in H2.
  change t with (projT1 (mk_pshare t (nidentity_pair2 n))) in H2.
  pfullshare_join.
  trivial.
  simpl.
  split;trivial.

  unfold points2,st_stk,fst in H8.
  rewrite H9 in H8.
  destruct H8 as[?[? [? ?]]].
  intro.
  intro.
  intro.
  destruct H14.
  destruct a as [[? ?] ?].
  destruct b1 as [[? ?] ?].
  destruct H14.
  unfold upd_heap,fst,snd in H16.
  unfold upd_heap,fst,snd in H14.
  destruct H14;subst.
  destruct H16. unfold fst,snd in *.
  apply injective_projections;simpl;trivial.
  apply injective_projections;simpl;trivial.
  destruct h1 as [? ?].
  destruct h2 as [? ?].
  destruct h0 .
  unfold st_brk,st_hf,st_hp, fst,snd in *.
  subst b5.
  apply injective_projections;simpl;trivial.
  icase b3.
  icase b4.
  icase b4.
  clear -H16 H8.
  extensionality x.
  spec H16 x.
  spec H8 x.
  case (eq_nat_dec x addr);intro.
  subst.
  icase (eq_nat_dec addr addr);trivial.
  inversion H16;subst;trivial.
  contradiction n.
  trivial.
  spec H8 n.
  
  revert H16; case (eq_nat_dec x addr); intros; contr.
  rewrite H8 in H16.
  inversion H16;
  trivial.

  unfold st_bsm, fst,snd in H11;subst.
  clear -H15.
  extensionality.
  spec H15 x.
  trivial.
  inversion H15;trivial.

Qed.

(*7new*)
Lemma Htuple_New : forall G v val e,
(* ------------------------------------------------------------*)
  Htuple G (eval_assert e val)
                  (New v e)
                  (points2 (expVar v) pfullshare (expConst val)).
Proof.
  repeat intro.
  inversion H2; subst.
  left.
  clear H H0 H2.
  destruct st' as [[? ?] ?].
  inv_sc H1.
  subst s0 s1.
  destruct h1.
  destruct H4 as [bk [? [? ?]]].
  compute in H.
  subst b2.
  destruct H1.
  destruct h0.
  destruct h.
  unfold fst, snd in *.
  icase b2.
  icase b3.
  simpl in H.
  unfold ojoin in H.
  subst n.
  remember (new_heap (Some x, r1) (res_builder val pfullshare)).
  icase o.
  destruct p.
  exists ((upd_stack s v (V_addr x), h, b), o', k).
  inv_sc H3.
  subst s0 s1 .
  destruct H7.
  unfold st_stk,fst in *.
  unfold st_hp,fst,snd in *.
  destruct H3.
  subst h1.
  apply join_com in H4.
  generalize (empty_heap_identity _ _ H4); intro.
  subst h0.
  clear H4.
  apply SeqStep_New with val x; auto.
  inversion Heqo.
  trivial.
  
(* STAGE 2 *)
  
  inversion H3; subst.
  spec H0 (upd_stack stk v (V_addr addr), hp', bs).
  detach H0.
  spec H0 o ora_ok st' o' k'.
  spec H0 H4.
  trivial.
  clear H2 H4 H3.
  rewrite (sc_comm F) in H1.
  rewrite sc_assoc in H1.
  inv_sc H1.
  subst s0 stk.
  destruct H4.
  unfold st_stk,st_hp, fst, snd in *.
  destruct H1.
  subst h.
  generalize (empty_heap_identity _ _ H2); intro.
  subst h0.
  clear H2.
  inv_sc H5.
  subst s0 s1.
  destruct H8 as [? [bk [? ?]]].
  unfold st_brk, st_hf,st_hp, fst, snd in *.
  destruct h0.
  subst b3.
  destruct h.
  destruct hp.
  destruct H5.
  unfold fst, snd in *.
  icase b3.
  icase b4.
  compute in H2.
  subst n.
  rewrite sc_assoc.
  spec H v.
  detach H; unfold modifies; trivial.
  generalize (H (s, (None, r0),b1) (V_addr addr) H7); intro.
  clear H.
  unfold st_upd_stk_var, st_upd_stk, st_stk, fst, snd in H2.
  unfold st_bsm, snd in H1;subst.
  unfold st_bsm,snd in H4; subst.
  apply join_com in H6.
  generalize (empty_bm_identity _ _ H6); intro.
  subst.
  clear H6.
  generalize (empty_bm_identity _ _ H3); intro.
  subst.
  exists (upd_stack s v (V_addr addr), (None, r0),bs).
(* a bit painful but it gets the job done *)
  exists (upd_stack s v (V_addr addr), (Some (bk + 1), 
    (fun addr' => if eq_nat_dec addr' bk 
                          then res_builder (evalExpr s e) pfullshare else
                          if le_gt_dec addr' bk then None else
                          res_builder falsE pfullshare)),empty_bm).
  inversion H12.
  (* Check out the size of the proof term! *)
  subst addr.
  split.
  2: split; trivial.
  split;unfold fst,snd.
  2:apply join_com;apply join_emptybm.
  split;unfold fst,snd. 
  split;trivial.
  split;unfold fst,snd.
  reflexivity.

  intro.
  case(eq_nat_dec x bk); intro.
  subst x.
  assert (r0 bk = None).
    spec H8 bk.
    spec H5 bk.
    destruct H8.
    assert (bk >= bk) by omega.
    spec H H4.
    rewrite H in H5.
    clear -H5.
    inversion H5; trivial.
    destruct x.
    destruct z.
    simpl in H2.
    destruct H2.
    simpl in H3.
    destruct x.
    simpl in H3.
    change t with (projT1 (mk_pshare t (nidentity_pair2 n))) in H3.
    change Share.top with (projT1 pfullshare) in H3.
    pfullshare_join.
  rewrite H.
  constructor.
  spec H8 x.
  destruct H8.
  spec H5 x.
  case (le_gt_dec x bk); intro.
  assert (x < bk) by omega.
  spec H1 H4.
  rewrite H1 in H5.
  trivial.
  assert (x >= bk) by omega.
  spec H H4.
  rewrite H in H5.
  trivial.
  exists (upd_stack s v (V_addr bk), (None, 
                  fun addr' => if eq_nat_dec addr' bk then 
                                       res_builder (evalExpr s e) pfullshare else
                                       None),empty_bm).
  exists (upd_stack s v (V_addr bk), (Some (bk + 1),
                  fun addr' => if le_gt_dec addr' bk then 
                                       None else
                                       res_builder falsE pfullshare),empty_bm).
  split.
  split;unfold fst, snd. 
  split;unfold fst, snd; trivial.
  split;trivial.
  split;unfold fst, snd.
  reflexivity.
  intro.
  simpl.
  case (eq_nat_dec x bk); intro.
  subst x.
  case (le_gt_dec bk bk); intro.
  constructor.
  omegac.
  case (le_gt_dec x bk); intro.
  constructor.
  constructor.
  constructor.
  unfold points2, st_stk, st_brk, st_hf, fst, snd, upd_stack.
  simpl.
  case (eq_nat_dec v v); intro.
  split; trivial.
  split;trivial.
  case (eq_nat_dec bk bk); intro.
  split;  trivial.
  split;trivial.  
  intros.
  icase (eq_nat_dec x bk).
  contradiction n;  trivial.
  split;trivial.
  simpl.
  unfold st_brk, st_hf,st_hp,fst,snd.
  destruct H0.
  exists (bk+1).
  split; trivial.
  intros.
  split.
  intros.
  case (le_gt_dec n bk).
  intro.
  omegac.
  trivial.
  intros.
  case (le_gt_dec n bk).
  trivial.
  intro.
  omegac.
  split.
  contradiction n;trivial.
  split; unfold st_bsm;trivial.
  unfold st_brk, st_hf,st_hp, upd_stack, fst, snd in *.
  exists (bk+1).
  split; trivial.
  contradiction n.
  trivial.
Qed.
  
(*8store*)
Lemma Htuple_Store : forall G e1 olde2 e2,
(* ------------------------------------------------------------*)
  Htuple G (points2 e1 pfullshare olde2)
                   (Store e1 e2)
                   (points2 e1 pfullshare e2).
Proof.
  repeat intro.
  inversion H2; subst.
  left.
  destruct st' as[[? ?] ?].
  rewrite sc_comm in H1.
  rewrite <- sc_assoc in H1.
  inv_sc H1.
  subst s0 s1.
  unfold points2, st_stk, fst in H7.
  remember (evalExpr s e1).
  icase v.
  destruct H7 as [? [? [? ?]]].
  unfold st_bsm,snd in H8. subst b1.
  apply join_com in H5.
  generalize (empty_bm_identity _ _ H5);intro;subst b0;clear H5.
  destruct h1.
  unfold st_brk,st_hf,st_hp,st_hf,fst, snd in *.
  subst b0.
  destruct H4.
  destruct h0.
  destruct h.
  simpl in H1, H4.
  unfold ojoin in H1.
  icase b0; icase b1.
  subst n0.
  eexists.
  econstructor; try reflexivity.
  rewrite <- Heqv.
  reflexivity.
  destruct (classic (a < n)); trivial.
  inv_sc H6.
  subst s0 s1.
  simpl in H8.
  destruct h.
  destruct h0.
  destruct H8.
  simpl in H5, H6.
  unfold pred_brk in H10.
  destruct H10 as [? [? [? ?]]].
  unfold st_brk,st_hp,st_hf, fst, snd in *.
  unfold st_hp,snd,fst in *.
  subst b2.
  unfold ojoin in H5.
  icase b3.
  subst x.
  spec H12 a.
  destruct H12.
  assert (a >= n) by omega.
  spec H5 H12.
  unfold m_join in H6.
  spec H6 a.
  spec H4 a.
  apply join_com in H6.
  destruct (join_assoc _ _ _ _ _ H6 H4) as [? [? ?]].

  remember (r2 a).
  remember (r a).
  inversion H13;subst.
  rewrite H5 in H15.
  inversion H15.
  rewrite H3 in H16.
  inversion H16.
  rewrite <- H16 in H13.
  rewrite <- H17 in H13.
  rewrite H5 in H16.
  rewrite H3 in H17.
  clear -H15 H17 H16.
  elimtype False.
  unfold res_builder in H16, H17.
  unfold pfullshare, mk_pshare in *.
  inversion H17.
  inversion H16.
  clear H17 H16.
  subst.
  destruct z.
  destruct H15.
  simpl in *.
  destruct x; simpl in H0.
  clear -H0.
  change Share.top with (projT1 pfullshare) in H0.
  pfullshare_join.
  spec H4 a.
  unfold dref_heap, snd.
  remember (r a).
  remember (r1 a).
  inversion H4; subst.
  rewrite <- H8.
  apply H3.
  rewrite H3 in H5.
  inversion H5.
  rewrite H3 in H8.
  inversion H8.
  subst y.
  unfold pjoin in H5.
  destruct x.
  destruct z.
  destruct x.
  destruct x0.
  destruct H5.
  simpl in H5, H10.
  destruct H5.
  subst v v0.
  clear -H10.
  assert (t0 = Share.top).
    LEM (identity t).
    generalize (H _ _ H10).
    auto.
    change t with (projT1 (mk_pshare t H)) in H10.
    change Share.top with (projT1 pfullshare) in H10.
    pfullshare_join.
  subst t0.
  unfold res_builder, pfullshare, mk_pshare.
  replace (@n2_identity_prod value Share.t sa_value Share.sa (evalExpr s olde2) Share.top top_share_nonidentity) with n1 by apply proof_irr; trivial.
  inv_sc H6.
  destruct H8.
  destruct h.
  destruct h0.
  simpl in *.
  unfold ojoin in H8.
  icase b2; icase b3.
  destruct H10 as [_ [? [? _]]].
  inversion H10.
  
  inversion H3; subst.
  spec H0 (stk, upd_heap hp addr (res_builder (evalExpr stk e2) pfullshare),bs).
  detach H0.
  spec H0 o ora_ok st' o' k'.
  apply H0; trivial.
  clear -H1 H10.
  rewrite sc_comm in *.
  rewrite <- sc_assoc in *.
  inv_sc H1.
  subst s0 stk.
  exists (s, h, b).
  exists (s, upd_heap h0 addr (res_builder (evalExpr s e2) pfullshare),b0).
  split.
  split;unfold fst,snd;trivial.
  split;unfold fst,snd.
  split;trivial.
  destruct H1.
  destruct h.
  destruct h0.
  destruct hp.
  unfold fst, snd in *.
  split; trivial.
  unfold upd_heap.
  unfold snd.
  intro.
  spec H0 x.
  case (eq_nat_dec x addr); intro; trivial.
  subst x.
  unfold points2, st_stk, st_hf, st_brk, fst, snd,st_hp,fst,snd in H4.
  rewrite H10 in H4.
  destruct H4 as [? [? [? _]]].
  inversion H0; subst.
  constructor.
  rewrite H4 in H8.
  inversion H8.
  rewrite H4 in H7.
  unfold res_builder, pfullshare, mk_pshare in H7.
  inversion H7.
  subst.
  clear H7.
  elimtype False.
  clear -H9.
  destruct x as [[? ?] ?].
  destruct z as [[? ?] ?].
  destruct H9.
  destruct H.
  simpl in *.
  subst v v0.
  generalize (nidentity_pair2 n); intro.
  change t with (projT1 (mk_pshare t H)) in H0.
  change Share.top with (projT1 pfullshare) in H0.
  pfullshare_join.
  split; trivial.
  destruct h0.
  
  inv_sc H3.
  subst s1 s0.
  destruct h0.
  destruct h1.
  destruct h.
  destruct H3.
  simpl in H, H0.
  destruct H6 as [? [bk [? ?]]].
  unfold st_brk, fst, snd,st_hp,fst,snd in H6.
  subst b4.
  icase b5.
  unfold ojoin in H.
  icase b6.
  subst n.
  destruct H1.
  destruct hp.
  simpl in H, H1.
  icase b1.
  unfold ojoin in H.
  icase b4.
  subst n.
  unfold points2 in *.
  unfold st_brk, st_hf, st_stk, fst, snd,st_hp,fst,snd in *.
  rewrite H10 in *.
  unfold upd_heap,st_hp,fst,snd.
  destruct H4 as [? [? [? ?]]].
  split; trivial.
  split.
  case (eq_nat_dec addr addr); intros; auto.
  contradiction n.
  trivial.
  split.
  intros.
  case (eq_nat_dec x addr); contr; auto.
  unfold st_bsm, snd in *.
  trivial.  
Qed.

(*9load*)
Lemma Htuple_Load : forall G e e' v vale vale' someshare,
(* ------------------------------------------------------------*)
  Htuple G (sepconj (points2 e someshare e') 
                   (sepconj (eval_assert e vale) (eval_assert e' vale')))
                  (Load v e)
                  (sepconj (points2 (expConst vale) someshare (expConst vale'))
                    (eval_assert (expVar v) vale')).
Proof.
  repeat intro.
  inversion H2; subst.
  left.
  destruct st' as [[? ?] bs].
  exists ((upd_stack s v vale', h,bs), o', k).
  clear H0 H.
  rewrite (sc_comm (points2 _ _ _)) in H1.
  rewrite <- sc_assoc in H1.
  rewrite (sc_comm F) in H1.
  rewrite sc_assoc in H1.
  rewrite sc_assoc in H1.
  inv_sc H1.
  inv_sc H4.
  subst s0 s1 s2 s3.
  destruct H10 as [? [? ?]].
  destruct H9 as [? [? ?]].
  unfold st_bsm, eval_assert, st_stk, fst, snd,st_hp,snd,fst in *.
  subst h3 h2 b2 b1.
  generalize (empty_heap_identity empty_heap h0 H7); intro.
  generalize (empty_bm_identity _ _ H8); intro.
  subst h0 b.
  generalize (empty_heap_identity _ _ H1); intro.
  generalize (empty_bm_identity _ _ H3); intro.
  subst h1 b0.
  clear H8 H1 H3 H7.
  rewrite <- sc_assoc in H5.
  inv_sc H5.
  subst s0 s1.
  destruct h1.
  destruct H7 as [? [bk [? ?]]].
  destruct h0.
  destruct h.
  destruct H3.
  unfold st_bsm,st_brk,st_hp,st_hf,st_hp,fst,snd in *.
  subst b0 b1.
  apply join_com in H4.
  generalize (empty_bm_identity _ _ H4);intro.
  subst b.
  simpl in H3.
  icase b2.
  icase b3.
  unfold ojoin in H3.
  subst n.
  inv_sc H5.
  subst s0 s1.
  destruct h.
  destruct h0.
  destruct H3.
  unfold fst, snd in *.
  icase b1; icase b2.
  unfold points2 in H10.
  unfold st_stk, st_hf, st_brk, fst, snd, st_hp, fst, snd in H10.
  rewrite H6 in H10.
  icase vale.
  destruct H10 as [? [? [? ?]]].
  clear H3 H0.
  spec H7 a.
  destruct H7.
  spec H8 a.
  spec H1 a.
  destruct (classic (a >= bk)).
  spec H0 H7.
  inversion H8.
  rewrite <- H14 in H1.
  destruct someshare.
  rewrite H10 in H1.
  unfold res_builder in H1.
  inversion H1.
  rewrite H0 in H15.
  inversion H15.
  elimtype False.
  rewrite H0 in H14.
  inversion H14.
  subst.
  destruct x.
  destruct z.
  simpl in H16.
  clear -H16 n.
  destruct H16.
  destruct x.
  destruct x0.
  simpl in *.
  clear -H0 n.
  change Share.top with (projT1 pfullshare) in H0.
  change t with (projT1 (mk_pshare t (nidentity_pair2 n))) in H0.
  pfullshare_join.
  destruct someshare.
  assert (a < bk) by omega.
  spec H3 H13.
  rewrite H3 in H8.
  inversion H8.
  rewrite <- H15 in H1.
  inversion H1.
  rewrite H10 in H20.
  inversion H20.
  rewrite H10 in H19.
  inversion H19.
  clear H3 H7.
  unfold res_builder in H10.
  rewrite H10 in H1.
  inversion H1.
  apply SeqStep_Load with bk a vale' (existT _ x n); auto.
  simpl.
  rewrite <- H15.
  rewrite <- H17.
  rewrite H.
  trivial.
  rewrite H in H17.
  destruct x1.
  destruct z.
  simpl in H17.
  destruct x2.
  destruct x1.
  destruct H17.
  simpl in H17, H18.
  destruct H17.
  subst v1 v0.
  assert (~identity t).
    intro.
    apply n1.
    rewrite identity_unit_equiv.
    split.
     split; trivial.
    rewrite identity_unit_equiv in H17.
    apply H17.
  apply SeqStep_Load with bk a vale' (mk_pshare t H17); auto.
  simpl.
  rewrite <- H15.
  rewrite <- H16.
  replace n1 with (@n2_identity_prod value Share.t sa_value Share.sa vale' t H17); trivial.
  apply proof_irr.
  
  (* Stage 2 *)
  inversion H3; subst.
  spec H0 (upd_stack stk v val, hp,bs).
  detach H0.
  spec H0 o ora_ok st' o' k'.
  auto.
  rewrite (sc_comm (points2 _ _ _)).
  rewrite <- sc_assoc.
  rewrite (sc_comm F).
  repeat rewrite sc_assoc.
  rewrite (sc_comm (points2 _ _ _)) in H1.
  rewrite <- sc_assoc in H1.
  rewrite (sc_comm F) in H1.
  do 2 rewrite sc_assoc in H1.
  inv_sc H1.
  subst s0 s.
  exists (upd_stack stk v val, empty_heap,empty_bm).
  exists (upd_stack stk v val, hp,bs).
  split.
  split;unfold fst,snd;trivial.
  split; unfold fst,snd;trivial.
  split;trivial.
  apply join_emptyheap.
  apply join_emptybm.
  unfold eval_assert, st_stk, fst, snd.
  unfold evalExpr.
  rewrite upd_stk_eq.
  inv_sc H7.
  subst s0 stk.
  destruct H15 as [? [? ?]].
  destruct H11 as [? [? ?]].
  unfold st_bsm,st_stk,st_hp,fst,snd in *.
  subst h1 h2 b1 b2.
  generalize (empty_heap_identity _ _ H7); intro.
  subst h.
  clear H7.
  generalize (empty_bm_identity _ _ H9); intro.
  subst b.
  clear H9.
  generalize (empty_bm_identity _ _ H6); intro.
  subst b0.
  clear H6.
  generalize (empty_heap_identity _ _ H5); intro.
  subst h0.
  clear H5.
  rewrite (sc_comm (points2 _ _ _)) in H8.
  rewrite <- sc_assoc in H8.
  inv_sc H8.
  subst s0 s1.
  unfold points2,st_stk,fst in H9.
  rewrite H10 in H9.
  destruct H9 as [? [? [? ?]]].
  destruct h0.
  unfold st_bsm,st_brk,st_hf, snd,st_hp,fst,snd in *.
  destruct hp.
  destruct h.
  destruct H6.
  unfold fst, snd in *.
  unfold dref_heap, brk_heap, fst, snd,st_hf,st_hp,fst,snd in *.
  subst b0 b1.
  simpl in H6.
  apply join_com in H7.
  generalize (empty_bm_identity _ _ H7);intro.
  subst b.
  clear H7.
  icase b3.
  unfold ojoin in H6.
  icase b2.
  subst n.
  clear H3 H2 H4.
  generalize (H16 addr); intro.
  rewrite H12 in H1.
  rewrite H5 in H1.
  inv_sc H8.
  simpl in H2, H3, H4.
  subst s0 s1.
  destruct H4.
  simpl in H2.
  destruct h0.
  destruct h.
  simpl in H2.
  destruct H8 as [? [bk [? ?]]].
  unfold st_brk, st_hf, st_hp, fst, snd in *.
  subst b1.
  icase b2.
  unfold ojoin in H2.
  subst bk.
  rewrite H0 in H1.
  assert (val = vale').
    destruct someshare.
    destruct sh.
    unfold res_builder in H1.
    inversion H1; auto.
    destruct x1.
    simpl in H18.
    destruct H18.
    destruct H18.
    simpl in *.
    rewrite H20.
    trivial.
  subst val.
  split.
  split; trivial.
  split; trivial.
  rewrite <- sc_assoc.
  rewrite (sc_comm F).
  rewrite sc_assoc.
  exists (upd_stack s v vale', (None, r),empty_bm).
  exists (upd_stack s v vale', (Some brk, r1),bs).
  split.
  split;unfold fst,snd.
  2:apply join_emptybm.
  split; unfold fst,snd;trivial.
  split;trivial.
  split;unfold fst,snd.
  inversion H13.
  split;trivial.
  apply join_com.
  trivial.
  split.
  unfold points2.
  unfold st_stk, st_hf, st_brk,st_hp, fst, snd,evalExpr.
  rewrite <- H11.
  rewrite H10.
  split; trivial.
  rewrite H0 in H5.
  split; trivial.
  split; trivial.
  unfold st_bsm;trivial.
  exists (upd_stack s v vale', (None, r3),b).
  exists (upd_stack s v vale', (Some brk, r2),b0).
  split.
  split;unfold fst,snd;trivial.
  split;unfold fst,snd; trivial.
  split;trivial.
  split;unfold fst,snd;trivial.
  reflexivity.
  split.
  spec H v.
  detach H.
  generalize (H (s, (None, r3),b) vale' H7); intro.
  apply H2.
  reflexivity.
  inversion H13;subst.
  split.
  unfold st_bsm,snd.
  unfold st_bsm,snd in H4; trivial.
  exists brk.
  split; trivial.
  icase b2.
Qed.

End HRules.
