(* CSL with Barriers, v0.1
 * 
 * It is our intention to release the files under a BSD-style license shortly
 * after some badly-needed cleanup work,  
 * 
 * Until then:
 * (c) 2010 Aquinas Hobor, Cristian Gherghina, and Christopher Chak
 * All Rights Reserved
 *
 * Permission granted to use these proof scripts only for the purpose of 
 * evaluating "Barriers in Concurrent Separation Logic."
 *)

Require Import SLB_Base.
Require Import SLB_Lang.
Require Import SLB_BarDefs.
Require Import SLB_CLang.
Require Import SLB_SSem.
Require Import SLB_Sem.

Module ConcOracle <: ORACLES.

Module CM := ConcMachine.

Record Ora : Type := ora {
  ora_sched : sched;
  ora_barpool : Barpool;
  ora_thds : list Thread}.
(* The alloc pool will be part of this thread's heap
    The thread-id will be the start of the scheduler
    We will replace this thread at position i (dummy thread) *)

Definition t : Type := Ora.

Fixpoint stdy_match (stbs : list BarDef) (dybs : list DyBarStatus) : Prop :=
  match (stbs, dybs) with
    | (nil, nil) => True
    | (nil, dy :: dybs') => False
    | (st :: stbs', nil) => False
    | (bard :: stbs', (dyBarStatus _ _ _ dybd _) :: dybs') =>
             bard = dybd /\ stdy_match stbs' dybs'
  end.

Lemma stdy_match_lookup: forall s d,
  stdy_match s d ->
  forall i,
    nth_error s i = option_map dbs_bd (nth_error d i).
Proof.
  induction s; intros.
  icase d.
  icase i.
  icase d.
  simpl in H.
  destruct d.
  destruct H.
  subst dbs_bd.
  destruct i; simpl.
  trivial.
  auto.
Qed.

Definition ora_barDefs (o : t) (sbs : list SLB_BarDefs.BarDef) : Prop :=
  stdy_match sbs (bp_bars (ora_barpool o)).

Definition dummyThd : Thread := 
  thread empty_stack empty_heap empty_bm (Running Unit).

Inductive Ora2CM : Ora -> state -> thd_state -> nat -> Cstate -> Prop :=
  | o2cm : forall i sched bp othds cthds stk hp hp' c ap bs pf,
    nth_error othds i = Some dummyThd ->
    replace othds i (thread stk hp' bs c)  = Some cthds ->
    pred_brk (empty_stack, ap, empty_bm) ->
    join hp' ap hp ->
    Ora2CM (ora sched bp othds) (stk, hp, bs) c i (cstate sched cthds bp ap pf).

Lemma Ora2CM_fun1 : forall o st c cs i cs',
  Ora2CM o st c i cs ->
  Ora2CM o st c i cs' ->
  cs = cs'.
Proof.
  intros.
  inversion H.
  inversion H0.
  subst.
  clear H H0.
  inversion H15.
  subst stk0 hp0 bs0.
  clear H15.
  inversion H14.
  subst sched0 bp0 othds0.
  generalize (pred_brk_precise (empty_stack, hp, empty_bm) (empty_stack, ap, empty_bm) (empty_stack, ap0, empty_bm)); intro.
  detach H.
  detach H.
  spec H H3 H12.
  inversion H.
  subst ap0.
  generalize (join_canc _ _ _ _ H4 H13); intro.
  subst hp'0.
  rewrite H11 in H2.
  inversion H2.
  subst cthds0.
  replace pf0 with pf by apply proof_irr.
  trivial.
  exists (empty_stack, hp'0, empty_bm).  
  apply join_com.
  repeat (split; trivial).
  apply join_emptybm.
  exists (empty_stack, hp', empty_bm).
  apply join_com.
  repeat (split; trivial).
  apply join_emptybm.
Qed.

Lemma Ora2CM_fun2 : forall o st c i cs o' st' c',
  Ora2CM o st c i cs ->
  Ora2CM o' st' c' i cs ->
  o = o' /\ st = st' /\ c = c'.
Proof.
  intros.
  inversion H.
  inversion H0.
  subst.
  clear H H0.
  inversion H18; subst.
  clear H18.
  generalize (replace_inveq _ _ _ _ _ _ _ H2 H11); intro.
  rewrite H10, H1 in H.
  detach H; trivial.
  subst othds0.
  split; trivial.
  generalize (replace_nth _ i othds _ _ H11).
  generalize (replace_nth _ i othds _ _ H2).
  intros.
  rewrite H in H0.
  inversion H0.
  subst.
  split; trivial.
  apply injective_projections; simpl; trivial.
  apply injective_projections; simpl; trivial.
  apply (join_eq _ _ _ _ H4 H13).
Qed.

Lemma Ora2CM_tot2 : forall i cs stk h bs c,
  nth_error (cs_thds cs) i = Some (thread stk h bs c) ->
  exists o, exists st,
    Ora2CM o st c i cs.
Proof.
  intros.
  destruct cs.
  simpl in H.
  destruct (can_replace _ _ _ _ H dummyThd) as [othds ?].
  exists (ora cs_sched cs_barpool othds).
  copy cs_wf.
  destruct cs_wf as [? [? [bptl [? [tw [? ?]]]]]].
  generalize (tl_states_lookup _ _ _ _ _ _ H); intro.
  assert (join empty_state (empty_stack, h, bs) (empty_stack, h, bs)).
    apply join_emptystate.
    trivial.
  destruct (@join_list_split_nth _ _ _ _ _ _ _ _ _ H6 H3 H7) as 
    [st [stl' [? [? ?]]]].
  destruct (@join_list_assoc2 _ _ _ _ _ _ _ H10 H4) as [st' [? ?]].
  apply join_com in H11.
  destruct (join_assoc _ _ _ _ _ H9 H11) as [? [? ?]].
  destruct x as [[stk' hp'] bs'].
  exists (stk, hp', bs').
  destruct H13 as [[[? ?] ?] ?].
  unfold fst, snd in *.
  subst.
  apply join_com in H17.
  generalize (empty_bm_identity _ _ H17); intro.
  subst bs'.
  apply o2cm with h; auto.
  eapply replace_nth; eauto.
  eapply replace_replace; eauto.
Qed.

Lemma Ora2CM_threadn : forall o st c n cs,
  Ora2CM o st c n cs ->
  exists h,
    nth_error (cs_thds cs) n = Some (thread (st_stk st) h (st_bsm st) c).
Proof.
  intros.
  inversion H; subst.
  exists hp'.
  simpl.
  rewrite (replace_nth _ _ _ _ _ H1).
  trivial.
Qed.

Definition ready_at (i : nat) (cs : Cstate) : Prop :=
  match cs_sched cs with
    | nil => False
    | tid :: sched' => 
        tid = i /\ 
        exists t, nth_error (cs_thds cs) i = Some t /\
        exists c, th_ctl t = Running c
  end.

Inductive stepOthers : nat -> Cstate -> Cstate -> Prop :=
  | so_done: forall i c,
      ready_at i c ->
      stepOthers i c c
  | so_step: forall i c c' c'',
      (~ready_at i c) ->
      CM.conc_step c c' ->
      stepOthers i c' c'' ->
      stepOthers i c c''.

Lemma thread_so_constants1 : forall cs n stk h bs k cs',
  nth_error (cs_thds cs) n = Some (thread stk h bs (Running k)) ->
  stepOthers n cs cs' ->
  nth_error (cs_thds cs') n = Some (thread stk h bs (Running k)).
Proof.
  intros.
  revert H.
  induction H0; auto.
  intro.
  assert (head (cs_sched c) <> Some i).
  intro.
  apply H.
  unfold ready_at.
  unfold head in H3.
  icase (cs_sched c).
  inversion H3.
  split; trivial.
  exists (thread stk h bs (Running k)).
  split; trivial.
  exists k.
  trivial.
  generalize (CM.cs_threads_preserved_notrunning _ _ _ _ _ _ _ H2 H3 H0); intro.
  auto.
Qed.

Lemma thread_so_constants2 : forall cs n stk h bs bn dir mv k cs',
  nth_error (cs_thds cs) n = Some (thread stk h bs (Waiting bn dir mv k)) ->
  stepOthers n cs cs' ->
  exists h', exists bs',
    nth_error (cs_thds cs') n = Some (thread stk h' bs' (Running k)).
Proof.
  intros.
  revert H.
  induction H0.
  intro.
  unfold ready_at in H.
  remember (cs_sched c).
  icase s.
  destruct H.
  subst.
  destruct H1 as [t' [? [ctl ?]]].
  rewrite H in H0.
  inversion H0.
  subst.
  disc.
  intros.
  destruct (CM.cs_threads_preserved_barrier2 _ _ _ _ _ _ _ _ _ _ H2 H0).
  auto.
  destruct H3 as [h' [bs' ?]].
  exists h'.
  exists bs'.
  eapply thread_so_constants1; eauto.
Qed.

Lemma barDefs_stepo_preserved : forall cs i cs',
  stepOthers i cs cs' ->
  bp_bardefs (cs_barpool cs) = bp_bardefs (cs_barpool cs').
Proof.
  induction 1; auto.
  rewrite CM.barDefs_cstep_preserved with c c'; auto.
Qed.

Lemma so_barrier_stuck: forall n cs cs',
  stepOthers n cs cs' ->
  forall stk hp bs bn dir mv k,
  nth_error (cs_thds cs) n = Some (thread stk hp bs (Waiting bn dir mv k)) ->
  forall pre post mv' pre' post',
  mv <> mv' ->
  lookup_move_bp (cs_barpool cs) bn dir mv pre post ->
  lookup_move_bp (cs_barpool cs) bn dir mv' pre' post' ->
  forall st,
  (sepconj pre Tru) st ->
  (sepconj pre' Tru) st ->
  forall slt,
  join (empty_stack, hp, bs) (slot2state slt) st ->
  lookup_slot_bp (cs_barpool cs) bn dir mv slt ->
  False.
Proof.
  induction 1; intros.
  clear -H H0.
  red in H.
  remember (cs_sched c).
  icase s.
  destruct H as [? [t' [? [c' ?]]]].
  subst i.
  rewrite H1 in H0.
  inversion H0.
  subst t'.
  disc.
  
  (* case 2 *)
  destruct (lookup_move_bp_lookup_state_bp _ _ _ _ _ _ H4) as [cst ?].
  destruct (CM.cs_barrier_stuck _ _ H0 _ _ _ _ _ _ _ _ H2 _ _ _ _ _ H3 H4 H5 _ H6 H7 _ H10 _ H8 H9) as [? [? ?]].
  spec IHstepOthers stk hp bs bn dir.
  spec IHstepOthers mv k H11 pre post.
  spec IHstepOthers mv' pre' post' H3.
  detach IHstepOthers.
  2: eapply CM.cs_lookup_move_bp; eauto.
  detach IHstepOthers.
  2: eapply CM.cs_lookup_move_bp; eauto.
  spec IHstepOthers st H6 H7 slt H8.
  apply IHstepOthers.
  trivial.
Qed.

Lemma so_waiting : forall i cs cs',
  stepOthers i cs cs' ->
  forall stk hp bs bn dir mv k pre post,
  lookup_move_bp (cs_barpool cs) bn dir mv pre post ->
  nth_error (cs_thds cs) i = Some (thread stk hp bs (Waiting bn dir mv k)) ->
   exists hp', exists bs', exists hp'', exists bs'',
     join bs bs' bs'' /\
     join hp hp' hp'' /\
     post (empty_stack, hp', bs') /\
     nth_error (cs_thds cs') i = Some (thread stk hp'' bs'' (Running k)).
Proof.
  induction 1; intros.
  red in H.
  rewrite H1 in H.
  icase (cs_sched c).
  destruct H as [? [? [? [? ?]]]].
  inversion H2.
  rewrite <- H5 in H3.
  disc.
  destruct (CM.cstep_waiting_inv c i stk hp bs bn dir mv k pre post c'); auto.
  destruct H4.
  spec IHstepOthers stk hp bs bn dir.
  spec IHstepOthers mv k pre post.
  apply IHstepOthers; auto.
  destruct H4 as [hp' [bs' [hp'' [bs'' [? [? [? ?]]]]]]].
  exists hp'.
  exists bs'.
  exists hp''.
  exists bs''.
  split; trivial.
  split; trivial.
  split; trivial.
  eapply thread_so_constants1; eauto.
Qed.

Inductive ostep : t -> state -> cmd -> option (t * state * cmd) -> Prop :=
  | os_badora1 : forall o st bn,
      ora_sched o = nil ->
      ostep o st bn None
  | os_badora2 : forall o st i sched c,
      ora_sched o = i :: sched ->
      (~exists cs, Ora2CM o st (Running c) i cs) ->
      ostep o st c None
  | os_steploop : forall o st c i sched cs cs',
      ora_sched o = i :: sched ->
      Ora2CM o st (Running c) i cs ->
      CM.conc_step cs cs' ->
      (~ exists cs'', stepOthers i cs' cs'') ->
      ostep o st c None
  | os_stepgood : forall o o' st st' c c' i sched cs cs' cs'',
      ora_sched o = i :: sched ->
      Ora2CM o st (Running c) i cs ->
      CM.conc_step cs cs' ->
      stepOthers i cs' cs'' ->
      Ora2CM o' st' (Running c') i cs'' ->
      ostep o st c (Some (o', st', c')).

Definition step (x : t * state * cmd) (y : option (t * state * cmd)) : Prop :=
  ostep (fst (fst x)) (snd (fst x)) (snd x) y.

Lemma stdy_match_lookup1: forall s d,
    stdy_match s d ->
    forall i, nth_error s i = option_map dbs_bd (nth_error d i).
Proof.
  induction s; intros.
  icase d.
  icase i.
  icase d.
  destruct i.
  simpl.
  destruct d; simpl in *.
  destruct H.
  subst.
  trivial.
  simpl.
  apply IHs.
  simpl in H.
  destruct d.
  tauto.
Qed.

Lemma stdy_match_lookup2: forall s d,
  (forall i, nth_error s i = option_map dbs_bd (nth_error d i)) ->
  stdy_match s d.
Proof.
  induction s; intros.
  spec H 0.
  icase d.
  red.
  trivial.
  destruct d.
  spec H 0.
  disc.
  simpl.
  destruct d.
  split.
  spec H 0.
  simpl in H.
  inversion H.
  trivial.
  apply IHs.
  intros.
  spec H (S i).
  apply H.
Qed.

Lemma ora_barDefs_exists: forall ora,
  exists G,
    ora_barDefs ora G.
Proof.
  intro.
  exists (map dbs_bd (bp_bars (ora_barpool ora0))).
  destruct ora0 as [sch bp othd].
  unfold ora_barDefs.
  simpl.
  destruct bp as [bars ? ?].
  simpl.
  clear.
  induction bars.
  exact I.
  simpl.
  destruct a.
  split; trivial.
Qed.

Lemma barDefs_ostep_preserved: forall o G st bn k o' st' k',
  ora_barDefs o G ->
  step (o, st, Seq (Barrier bn) k) (Some (o', st', k')) ->
  ora_barDefs o' G.
Proof.
  intros.
  inversion H0; subst.
  generalize (CM.barDefs_cstep_preserved _ _ H9); intro.
  generalize (barDefs_stepo_preserved _ _ _ H10); intro.
  unfold ora_barDefs in *.
  apply stdy_match_lookup2.
  intros.
  generalize (stdy_match_lookup _ _ H); intro.
  rewrite H3.
  destruct o.
  destruct o'.
  simpl in *.
  do 2 rewrite <- nth_error_map.
  unfold bp_bardefs in H1, H2.
  inversion H5; subst.
  inversion H11; subst.
  simpl in *.
  rewrite <- H2.
  rewrite <- H1.
  trivial.
Qed.

Lemma ostep_preserves_ctl : forall o st bn k o' st' k',
  step (o, st, Seq (Barrier bn) k) (Some (o', st', k')) ->
  k = k'.
Proof.
  intros.
  inversion H; subst.
  inversion H4; subst.
  unfold ora_sched, fst, snd in *.
  subst.
  subst.
  apply replace_nth in H1.
  change cthds with (cs_thds (cstate (i::sched) cthds bp ap pf)) in H1.
  assert (head (cs_sched (cstate (i :: sched) cthds bp ap pf)) = Some i) by auto.
  apply (CM.cs_threads_preserved_barrier _ _ _ _ _ _ _ _ H1 H8) in H3.
  destruct H3.
  destruct H3 as [? [? [? [? ?]]]].
  apply (thread_so_constants2 _ _ _ _ _ _ _ _ _ _ H3) in H9.
  destruct H9 as [? [? ?]].
  inversion H10; subst.
  apply replace_nth in H9.
  simpl in H6.
  rewrite H9 in H6.
  inversion H6.
  trivial.
  destruct H3 as [? [? ?]].
  apply (thread_so_constants1 _ _ _ _ _ _ _ H3) in H9.
  inversion H10; subst.
  apply replace_nth in H7.
  simpl in H9.
  rewrite H9 in H7.
  inversion H7.
  trivial.
Qed.  

Lemma O2CM_can_step: forall n sch bp thds st bn k cs bd cst dir mv pre post F,
  Ora2CM (ora (n :: sch) bp thds) st (Running (Seq (Barrier bn) k)) n cs ->
  option_map dbs_bd (nth_error (bp_bars bp) bn) = Some bd ->
  lookup_move bd cst dir mv pre post ->
  (sepconj (sepconj F pre) pred_brk) st ->
  exists cs', CM.conc_step cs cs'.
Proof.
  intros.
  inversion H; subst; clear H.
  inv_sc H2.
  subst s0 s.
  rewrite <- pred_brk_ubiquitous in H12.
  spec H12 stk.
  unfold st_upd_stk, fst, snd in H12.
  
  generalize (pred_brk_precise (stk, hp, bs) (stk, ap, empty_bm) (stk, h0, b0)); intro.
  detach H.
  2: exists (stk, hp', bs); split; try apply join_emptybm; apply join_com; split; trivial; split; trivial.
  detach H.
  2: exists (stk, h, b); apply join_com; split; trivial; split; trivial; split; trivial.
  spec H H12 H8.
  inversion H.
  subst h0 b0.
  clear H H12.
  generalize (join_canc _ _ _ _ H13 H3); intro.
  subst hp'.
  clear H3.    
  apply join_com in H4.
  generalize (empty_bm_identity _ _ H4); intro.
  subst b.
  clear H4.
  assert (nth_error cthds n = Some (thread stk h bs (Running (Seq (Barrier bn) k)))) by (apply replace_nth with thds; trivial).
  remember (nth_error (bp_bars bp) bn).
  icase e.
  destruct d as [bn' cst' wp bd' wfbd].
  simpl in H0.
  inversion H0.
  subst bd'.
  clear H0.
  assert (bn = bn').
    destruct bp.
    copy bp_wf.
    destruct bp_wf0.
    simpl in Heqe.
    spec H2 bn (dyBarStatus bn' cst' wp bd wfbd).
    rewrite Heqe in H2.
    detach H2; trivial.
    simpl in H2.
    auto.
  subst bn'.
  assert (cst' = cst).
    inv_sc H5.
    subst s0 s.
    destruct (lookup_move_facts _ _ _ _ _ _ H1) as [? [? [? [ps ?]]]].
    assert ((sepconj (barrier_assert (bd_bn bd) ps cst) Tru) (stk, h1, b0)) by auto.
    inv_sc H12.
    subst s s0.
    copy pf.
    destruct pf0 as [? [? [bptp [? [tw [? ?]]]]]].
    spec H21 bn ps cst.
    rewrite <- Heqe in H21.
    simpl in H21.
    detach H21.
    inversion H21.
    trivial.
    exists (empty_stack, h2, b1).
    move H12 after n.
    generalize (tl_states_lookup _ _ _ _ _ _ H); intro.
    rewrite <- ubiquitous_barrier_assert in H17.
    spec H17 empty_stack.
    unfold st_upd_stk, fst, snd in H17.
    apply join_com in H3.
    destruct (join_assoc _ _ _ _ _  H15 H3) as [hf1 [? ?]].
    apply join_com in H4.
    destruct (join_assoc _ _ _ _ _  H16 H4) as [bf1 [? ?]].
    assert (join (empty_stack, h2, b1) (empty_stack, hf1, bf1) (empty_stack, h, bs)) by (repeat (split; trivial)).
    apply join_com in H26.
    destruct (join_list_split_nth _ _ _ H21 H19 H26) as [sf1 [nl [? [? ?]]]].
    apply join_com in H28.
    destruct (join_list_assoc1 H28 H29) as [sf2 [? ?]].
    apply join_com in H20.
    destruct (join_assoc _ _ _ _ _ H31 H20) as [sf3 [? ?]].
    exists sf3.
    split; trivial.
    split; trivial.
    replace bn with (bd_bn bd); trivial.
    destruct bp as [bars bpst bpwf].
    copy wfbd.
    unfold dbs_wellformed in wfbd0.
    destruct wp.
    destruct wfbd0 as [? [? ?]].
    auto.
  subst cst'.
  assert (lookup_move_bp bp bn dir mv pre post).
    unfold lookup_move_bp.
    rewrite <- Heqe.
    trivial.
  inv_sc H5.
  subst s0 s.
  assert (sepconj (and (exact (stk, h1, b0)) pre) (exact (stk, h0, b)) (stk, h, bs)).
    exists (stk, h1, b0).
    exists (stk, h0, b).
    split.
    apply join_com.
    repeat (split; trivial).
    split.
    split; trivial.
    reflexivity.
    reflexivity.
  assert (exists bp', bp_inc_waitpool bp bn dir mv (n, (h1, b0)) bp').
    destruct (lookup_move_facts _ _ _ _ _ _ H1) as [? [? [? [ps ?]]]].
    rewrite <- H3 in H10.
    spec H10 empty_stack.
    unfold st_upd_stk, fst, snd in H10.
    assert (joins (empty_stack, h1, b0) (bp_st bp)).
      copy pf.
      destruct pf0 as [? [? [st [? ?]]]].
      generalize (tl_states_lookup _ _ _ _ _ _ H); intro.
      assert (join (empty_stack, h0,b) (empty_stack, h1, b0) (empty_stack, h, bs)) by (repeat (split; trivial)).
      destruct (join_list_split_nth _ _ _ H19 H17 H20) as [st' [? [? [? ?]]]].
      exists st'.
      apply join_com.
      trivial.
    apply (can_bp_inc_wp bp bn wp dir bd cst mv pre post h1 b0 n wfbd); auto.
    (* Should this be here?? Maybe into can_bp_inc_wp?? *)
    destruct wp as [diro slts lim wpst wfwp].
    simpl.
    destruct diro; auto.
    right.
    case (eq_nat_dec d dir); intro; auto.
    elimtype False.
    copy wfwp.
    destruct wfwp0 as [? [? ?]].
    assert (exists sn, exists slt, nth_error slts sn = Some (Some slt)).
      clear -H17.
      revert H17.
      induction slts; intros.
      simpl in H17.
      assert (0 = 0) by reflexivity.
      rewrite <- H17 in H.
      disc.
      destruct a.
      exists 0.
      exists s.
      trivial.
      simpl in H17.
      spec IHslts H17.
      destruct IHslts as [? [? ?]].
      exists (S x).
      exists x0.
      apply H.
    destruct H19 as [sn [slt ?]].
    copy wfbd.
    destruct wfbd0 as [? [? ?]].
    spec H22 sn.
    rewrite H19 in H22.
    destruct H22 as [pre' [post' [? ?]]].
    generalize (lookup_move_only1DirOK _ _ _ _ _ _ _ _ _ _ H1 H22); intro.
    destruct H24.
    apply n0; auto.
    destruct H15 as [nbpst ?].
    destruct bp.
    copy bp_wf.
    destruct bp_wf0.
    simpl in Heqe.
    assert (nth_error (dbs_getsts bp_bars) bn = Some wpst).
      unfold dbs_getsts.
      rewrite nth_error_map.
      rewrite <- Heqe.
      compute.
      trivial.
    unfold SLB_CLang.bp_st in H15.
    destruct (join_list_assoc2 H25 H15) as [st [? ?]].
    apply join_com in H28.
    generalize (empty_state_identity _ _ H28); intro.
    subst st.
    assert (join empty_state wpst wpst).
      apply join_emptystate.
      eapply join_list_empty_state; eauto.
    destruct (join_list_split_nth _ _ _ H27 H29 H30) as [f [? [? [? ?]]]].
    destruct (join_list_assoc2 H18 H32) as [st [? ?]].
    apply join_com in H34.
    generalize (empty_state_identity _ _ H34); intro.
    subst st.
    assert (nth_error (wp_getstates slts) sn = Some (slot2state slt)).
      unfold wp_getstates.
      rewrite nth_error_map.
      change (prod tid (prod heap BarstMap)) with slot.
      rewrite H19.
      simpl.
      unfold slot2state.
      trivial.
    assert (join empty_state (slot2state slt) (slot2state slt)).
      apply join_emptystate.
      unfold slot2state.
      destruct slt.
      destruct p.
      trivial.
    destruct (join_list_split_nth _ _ _ H36 H35 H37) as [f' [? [? [? ?]]]].
    change False with (Fls f').
    rewrite <- H24.
    split.
    exists (empty_stack, h1, b0).
    exists (slot2state slt).
    split; trivial.
    split; trivial.
    compute.
    trivial.
    exists (slot2state slt).
    exists (empty_stack, h1, b0).
    apply join_com in H39.
    split; trivial.
    split; trivial.
    compute.
    trivial.
    (* end should this be here?? *)
  destruct H3 as [bp' ?].
  destruct (can_replace _ _ _ _ H (thread stk h0 b (Waiting bn dir mv k))) as [cthds' ?].
  LEM (bp_ready bp' bn).
  (* case 1 *)
  destruct (can_bp_transition _ _ H12) as [bp'' [out ?]].
  copy pf.
  move pf after pf0.
  destruct pf as [? [? ?]].
  rename pf0 into pf.
  assert (exists cthds'', transition_threads out cthds' cthds'').
    generalize (can_transition_threads _ _ _ _ H14); intro.
    detach H18.
    2: eapply bpi_uni_nin_uni; eauto; change bp with (cs_barpool (cstate _ _ _ _ pf)); try apply cs_slots_unique; eapply cs_running_nin_bp; eauto.
    copy pf.
    destruct pf0 as [? [? [bpthst [? ?]]]].
    spec H18 cthds' bpthst.
    apply H18.
    (* BR1 *)
    clear -H21 H3 H11 H2 H.
    destruct bp as [bars bst wfbp].
    destruct bp' as [bars' bst' wfbp'].
    destruct H3 as [? _].
    unfold slot2state, fst, snd, bp_st in *.
    inv_sc H2.
    subst s s0.
    destruct H5.
    inversion H6.
    inversion H1.
    subst h3 h2 b1 b2.
    clear H1 H6.
    assert (nth_error (tl_states cthds) n = Some (empty_stack, h, bs)).
      unfold tl_states.
      rewrite nth_error_map.
      rewrite H.
      trivial.
    assert (join (empty_stack, h0, b) (empty_stack, h1, b0) (empty_stack, h, bs)).
      apply join_com.
      repeat (split; trivial).
    destruct (join_list_split_nth _ _ _ H1 H21 H5) as [f [thds' [? [? ?]]]].
    assert (thds' = tl_states cthds').
      apply nth_error_eq.
      unfold tl_states.
      intro j.
      rewrite nth_error_map.
      case (eq_nat_dec n j); intro.
      subst j.
      rewrite (replace_nth _ _ _ _ _ H6).
      rewrite (replace_nth _ _ _ _ _ H11).
      trivial.
      rewrite (replace_nth2 _ _ _ _ _ _ H6 n0).
      rewrite (replace_nth2 _ _ _ _ _ _ H11 n0).
      unfold tl_states.
      rewrite nth_error_map.
      trivial.
    subst thds'.
    apply join_com in H7.
    apply (join_eq _ _ _ _ H0) in H7.
    subst f.
    trivial.
    (* BR2 *)
    intros.
    case (eq_nat_dec n i); intro.
    subst i.
    rewrite (replace_nth _ _ _ _ _ H11).
    trivial.
    rewrite (replace_nth2 _ _ _ _ _ _ H11 n0).
    assert (dir = dir0).
      apply bpiwp_luplstbp in H3.
      red in H3, H23.
      icase (nth_error (bp_bars bp') bn).
      destruct d.
      destruct H3.
      destruct H23.
      congruence.
    subst dir0.
    assert (mv <> mv0).
      case (eq_nat_dec mv mv0); auto.
      intro.
      subst mv0.
      apply bpiwp_luplstbp in H3.
      apply (lookup_slot_bp_fun _ _ _ _ _ _ H3) in H23.
      inversion H23.
      contradiction.
    rewrite <- (bp_inc_wp_constant _ _ _ _ _ _ H3) in H23; auto.
    spec H19 bn dir mv0 i.
    assert (exists h, exists b, lookup_slot_bp bp bn dir mv0 (i, (h, b))).
      destruct sx.
      exists h2.
      exists b1.
      trivial.
    rewrite H19 in H25.
    destruct H25 as [stkx [hpx [bmx [cx ?]]]].
    rewrite H25.
    trivial.
  destruct H18 as [cthds'' ?].
  assert (cs_wellformed sch cthds'' bp'' ap).
    split.
    (* prop1 *)
    intros.
    apply prop_ext.
    split; intro.
    (* BR1 *)
    destruct H19 as [hx [bx ?]].
    case (eq_nat_dec bn bn0); intro.
    subst bn0.
    elimtype False.
    eapply bptrans_bpnin; eauto.
    apply (bptrans_slt2 _ _ _ _ H14 _ n0) in H19.
    rewrite <- (bp_inc_wp_constant _ _ _ _ _ _ H3) in H19; auto.
    copy pf.
    destruct pf0 as [? _].
    spec H20 bn0 dir0 mv0 tid.
    assert (exists h, exists b, lookup_slot_bp bp bn0 dir0 mv0 (tid, (h, b))) by (exists hx; exists bx; trivial).
    rewrite H20 in H21.
    destruct H21 as [stkx [hx' [bx' [cx ?]]]].
    case (eq_nat_dec n tid); intro.
    subst tid.
    rewrite H in H21.
    disc.
    rewrite <- (replace_nth2 _ _ _ _ _ _ H11 n1) in H21.
    exists stkx.
    exists hx'.
    exists bx'.
    exists cx.
    rewrite (transition_threads_constant _ _ _ H18) in H21; auto.
    intro.
    apply (bptrans_out_tid _ _ _ _ H14) in H22.
    destruct H22 as [dirx [mvx [[? ?] [? ?]]]].
    move H19 after H22.
    assert (dirx = dir).
      apply bpiwp_luplstbp in H3.
      clear - H22 H3.
      red in H3, H22.
      icase (nth_error (bp_bars bp') bn).
      destruct d.
      destruct H3.
      destruct H22.
      congruence.
    subst dirx.
    case (eq_nat_dec mv mvx); intro.
    subst mvx.
    apply bpiwp_luplstbp in H3.
    apply (lookup_slot_bp_fun _ _ _ _ _ _ H3) in H22.
    compute in H23.
    subst t0.
    inversion H22.
    contradiction.
    rewrite <- (bp_inc_wp_constant _ _ _ _ _ _ H3) in H22; auto.
    compute in H23.
    subst t0.
    destruct p.
    copy pf.
    destruct pf0 as [? _].
    spec H23 bn dir mvx tid.
    assert (exists h, exists b, lookup_slot_bp bp bn dir mvx (tid, (h, b))) by (exists h2; exists b1; trivial).
    assert (exists h, exists b, lookup_slot_bp bp bn0 dir0 mv0 (tid, (h, b))) by (exists hx; exists bx; trivial).
    rewrite H23 in H24.
    rewrite H15 in H25.
    destruct H24 as [? [? [? [? ?]]]].
    destruct H25 as [? [? [? [? ?]]]].
    rewrite H24 in H25.
    inversion H25.
    contradiction.
    (* BR2 *)
    destruct H19 as [stkx [hpx [bmx [cx ?]]]].
    copy H19.
    apply (trans_thds_rev _ _ _ H18) in H20.
    case (eq_nat_dec bn bn0); intro.
    subst bn0.
    elimtype False.
    (* case a *)
    generalize (transition_threads_changes _ _ _ H18); intro.
    detach H21.
    assert (In tid (map slot2tid (dbsto_slots out))).
      clear H21.
      case (eq_nat_dec n tid); intro.
      subst tid.
      rewrite (replace_nth _ _ _ _ _ H11) in H20.
      inversion H20.
      subst stkx hpx bmx dir0 mv0 cx.
      copy H3.
      apply bpiwp_luplstbp in H21.
      assert (exists pre, exists post, lookup_move_bp bp' bn dir mv pre post).
        eapply lupslt_lupmv; eauto.
      destruct H22 as [prex [postx ?]].
      destruct (lookup_slot_out _ _ _ _ H14 _ _ _ _ H22 _ H21) as [sltx [? [? _]]].
      destruct sltx.
      compute in H24.
      subst t0.
      apply nth_error_In with mv.
      rewrite nth_error_map.
      rewrite H23.
      trivial.
      rewrite (replace_nth2 _ _ _ _ _ _ H11 n0) in H20.
      copy pf.
      destruct pf0 as [? _].
      spec H21 bn dir0 mv0 tid.
      assert (exists stk, exists hp, exists bm, exists c, nth_error cthds tid = Some (thread stk hp bm (Waiting bn dir0 mv0 c))).
        exists stkx.
        exists hpx.
        exists bmx.
        exists cx.
        trivial.
      rewrite <- H21 in H22.
      clear H21.
      destruct H22 as [hx [bx ?]].
      rewrite (bp_inc_wp_constant _ _ _ _ _ _ H3) in H21.
      assert (exists pre, exists post, lookup_move_bp bp' bn dir0 mv0 pre post).
        eapply lupslt_lupmv; eauto.
      destruct H22 as [prex [postx ?]].
      destruct (lookup_slot_out _ _ _ _ H14 _ _ _ _ H22 _ H21) as [sltx [? [? _]]].
      destruct sltx.
      compute in H24.
      subst t0.
      apply nth_error_In with mv0.
      rewrite nth_error_map.
      rewrite H23.
      trivial.
      case (eq_nat_dec dir dir0); auto.
      case (eq_nat_dec mv mv0); auto.
      intros.
      subst dir0 mv0.
      elimtype False.
      generalize (bp_lupmvslt_pre _ _ _ _ _ _ _ H0 H21); intro.
      destruct (lookup_move_bp_facts _ _ _ _ _ _ H0) as [cstx [_ [? [_ [? [ps ?]]]]]].
      destruct (lupmvbp_lupstbp _ _ _ _ _ _ H0) as [bst ?].
      generalize (bp_lupsltst _ _ _ _ _ _ H26 H21); intro.
      change cthds with (cs_thds (cstate _ _ _ _ pf)) in H.
      change bp with (cs_barpool (cstate _ _ _ _ pf)) in H26.
      generalize (lupst_joins_thd _ _ _ H26 _ _ _ _ _ H); intro.
      destruct H28 as [tst ?].
      spec H24 tst (empty_stack, hx, bx) (empty_stack, h1, b0).
      unfold slot2state in *.
      assert (join_sub bst tst) by (eexists; eauto).
      detach H24.
      2: eapply join_sub_trans; eauto.
      assert (join_sub (empty_stack, h, bs) tst) by (apply join_com in H28; eexists; eauto).
      assert (join_sub (empty_stack, h1,b0) (empty_stack, h, bs)) by (exists (empty_stack, h0, b); apply join_com; repeat (split; trivial)).
      detach H24.
      2: eapply join_sub_trans; eauto.
      rewrite <- H23 in H10.
      spec H10 empty_stack.
      spec H24 H22 H10.
      inversion H24.
      subst h1 b0.
      destruct H27 as [ob ?].
      apply join_com in H27.
      destruct (join_assoc _ _ _ _ _ H27 H28) as [f [? ?]].
      apply join_com in H32.
      destruct H31 as [ot ?].
      apply join_com in H31.
      destruct (join_assoc _ _ _ _ _ H31 H32) as [f' [? ?]].
      generalize (join_self _ _ H34); intro.
      subst f'.
      assert (identity (empty_stack, hx, bx)) by (rewrite identity_unit_equiv; auto).
      apply identity_emptystate in H36; auto.
      inversion H36.
      subst hx bx.
      apply H25 in H22.
      clear -H22.
      inv_sc H22.
      generalize (split_identity _ _ H2 empty_bm_identity); intro.
      apply identity_emptybm in H5.
      subst b.
      destruct H3.
      destruct ps.
      disc.
    apply In_nth_error in H22.
    destruct H22 as [j ?].
    rewrite nth_error_map in H22.
    remember (nth_error (dbsto_slots out) j).
    icase e.
    spec H21 j s.
    detach H21; auto.
    spec H21 stkx hpx bmx bn dir0.
    destruct s.
    inversion H22.
    subst t0.
    spec H21 mv0 cx H20.
    destruct H21 as [hx [bx [? ?]]].
    simpl in H23.
    rewrite H19 in H23.
    disc.
    eapply bpt_uni_out; eauto.
    eapply bpi_uni_nin_uni; eauto.
    change bp with (cs_barpool (cstate _ _ _ _ pf)).
    apply cs_slots_unique.
    change bp with (cs_barpool (cstate _ _ _ _ pf)).
    eapply cs_running_nin_bp; eauto.
    (* case b *)
    case (eq_nat_dec n tid); intro.
    subst tid.
    rewrite (replace_nth _ _ _ _ _ H11) in H20.
    inversion H20.
    contradiction.
    rewrite (replace_nth2 _ _ _ _ _ _ H11 n1) in H20.
    copy pf.
    destruct pf0 as [? _].
    spec H21 bn0 dir0 mv0 tid.
    assert (exists stk, exists hp, exists bm, exists c, nth_error cthds tid = Some (thread stk hp bm (Waiting bn0 dir0 mv0 c))).
      exists stkx.
      exists hpx.
      exists bmx.
      exists cx.
      trivial.
    rewrite <- H21 in H22.
    clear H21.
    destruct H22 as [hx [bx ?]].
    exists hx.
    exists bx.
    rewrite (bp_inc_wp_constant _ _ _ _ _ _ H3) in H21; auto.
    eapply bptrans_slt; eauto.
    (* prop2 *)
    split; auto.
    (* prop3 *)
    destruct H17 as [bptl [? [tw [? ?]]]].
    destruct (lookup_move_bp_lookup_state_bp _ _ _ _ _ _ H0) as [cs ?].
    apply (bpiwp_lupbst _ _ _ _ _ _ H3) in H21.
    assert (exists cs', lookup_bstate_bp bp'' bn cs').
      eapply bptrans_lupmvbp_lupbst; eauto.
      eapply bpiwp_lupmvbp; eauto.
    destruct H22 as [cs' ?].
    destruct (bp_trans_st _ _ _ _ H14 _ _ H21 H22).
    exists (st_upd_bm_bn bptl bn (barst_builder cs' pfullshare)).
    split.
    generalize (transthds_join _ _ _ H18); intro.
    move H3 after H25.
    move H17 after H3.    
    assert (nth_error (tl_states cthds) n = Some (empty_stack, h, bs)).
      unfold tl_states.
      rewrite nth_error_map.
      rewrite H.
      trivial.
    assert (join (empty_stack, h0, b) (empty_stack, h1, b0) (empty_stack, h, bs)).
      repeat (split; trivial).
    destruct (join_list_split_nth _ _ _ H26 H17 H27) as [f [tlx [? [? ?]]]].
    assert (tlx = tl_states cthds').
      apply nth_error_eq.
      intro.
      case (eq_nat_dec n i); intro.
      subst i.
      apply replace_nth in H28.
      rewrite H28.
      unfold tl_states.
      rewrite nth_error_map.
      apply replace_nth in H11.
      rewrite H11.
      trivial.
      rewrite (replace_nth2 _ _ _ _ _ _ H28 n0).
      unfold tl_states.
      do 2 rewrite nth_error_map.
      rewrite (replace_nth2 _ _ _ _ _ _ H11 n0).
      trivial.
    subst tlx.
    generalize (bpiwp_st _ _ _ _ _ _ H3); intro.
    unfold slot2state in H31.
    apply (join_eq _ _ _ _ H29) in H31.
    subst f.
    assert (join_list (st_upd_bm_bn (bp_st bp') bn (barst_builder cs' pfullshare)) (tl_states cthds') (st_upd_bm_bn bptl bn (barst_builder cs' pfullshare))).
      destruct bp' as [bars' bpst' wpbp'].
      unfold bp_st in *.
      clear -H30 H23.
      destruct bpst' as [[? ?] ?].
      inv_sc H23.
      subst s0 s1.
      clear H3.
      assert (join (s, h1, b1) (s, h0, b0) (s, h, b)) by (apply join_com; repeat (split; trivial)).
      destruct (join_list_assoc1 H H30) as [f [? ?]].
      assert (join (st_upd_bm_bn (s, h1, b1) bn (barst_builder cs' pfullshare)) f (st_upd_bm_bn bptl bn (barst_builder cs' pfullshare))).
        clear -H4 H3.
        destruct H4 as [? [? ?]].
        unfold st_bsm, st_hp, fst, snd in *.
        destruct f as [? ?].
        destruct bptl as [? ?].
        destruct H3.
        split; trivial.
        unfold st_upd_bm_bn, upd_bm, snd.
        intro bx.
        case (eq_nat_dec bx bn); intro.
        subst bx.
        unfold snd in H3.
        spec H3 bn.
        rewrite H in H3.
        unfold pfullshare, barst_builder in H3.
        simpl in H3.
        inversion H3.
        constructor.
        destruct y.
        destruct z.
        destruct H7.
        unfold fst, snd in *.
        destruct x0.
        destruct x1.
        change Share.top with (projT1 pfullshare) in H8.
        generalize (nidentity_pair2 n); intro.
        change t0 with (projT1 (existT (fun t => ~identity t) _ H9)) in H8.
        pfullshare_join.
        apply H3.
      destruct (join_list_assoc2 H0 H5) as [f' [? ?]].
      replace (st_upd_bm_bn (s, h, b) bn (barst_builder cs' pfullshare)) with f'; auto.
      clear -H6 H.
      destruct f' as [[? ?] ?].
      destruct H6 as [[? ?] ?].
      destruct H as [[? ?] ?].
      unfold st_upd_bm_bn, fst, snd in *.
      apply (join_eq _ _ _ _ H3) in H1.
      destruct H0.
      subst s0 h2.
      apply injective_projections; unfold snd; auto.
      extensionality bx.
      unfold upd_bm in *.
      spec H2 bx.
      spec H4 bx.
      revert H2.
      case (eq_nat_dec bx bn); intros.
      subst bx.
      unfold barst_builder, pfullshare in *.
      simpl in *.
      inversion H2; auto.
      destruct y.
      destruct z.
      destruct H7.
      destruct x0.
      destruct x1.
      simpl in H8.
      change Share.top with (projT1 pfullshare) in H8.
      generalize (nidentity_pair2 n); intro.
      change t0 with (projT1 (existT (fun t => ~identity t) _ H9)) in H8.
      pfullshare_join.
      apply (join_eq _ _ _ _ H2) in H4.
      trivial.
    destruct (join_list_assoc1 H24 H31) as [f [? ?]].
    spec H25 f H32.
    destruct (join_list_assoc2 H25 H33) as [f' [? ?]].
    apply join_com in H34.
    apply empty_state_identity in H34.
    subst f'.
    trivial.
    exists (st_upd_bm_bn tw bn (barst_builder cs' pfullshare)).
    split.
    clear -H19.
    destruct H19.
    split; trivial.
    destruct bptl.
    destruct tw.
    unfold st_upd_bm_bn, fst, snd, upd_bm in *.
    intro bnx.
    case (eq_nat_dec bnx bn); intro.
    subst bnx.
    unfold empty_bm.
    constructor.
    apply H0.
    (* prop4 *)
    intros.
    case (eq_nat_dec bn bn0); intro.
    (* BR1 *)
    subst bn0.
    clear -H25 H22.
    red in H22.
    icase (nth_error (bp_bars bp'') bn).
    destruct d.
    simpl.
    subst dbs_cs.
    destruct tw as [[? ?] ?].
    unfold st_upd_bm_bn, fst, snd in H25.
    inv_sc H25.
    subst.
    spec H2 bn.
    unfold upd_bm in H2.
    revert H2.
    case (eq_nat_dec bn bn); intro hx.
    2: destruct hx; trivial.
    intro.
    destruct H3.
    simpl in H.
    rewrite H in H2.
    unfold barst_builder in H2.
    destruct pfullshare.
    destruct ps.
    inversion H2; subst; auto.
    destruct y.
    destruct H7.
    destruct H3.
    unfold fst in *.
    congruence.
    (* BR2 *)
    spec H20 bn0 ps cst0.
    assert ((sepconj (barrier_assert bn0 ps cst0) Tru) tw).
      clear -H25 n0.
      unfold st_upd_bm_bn, fst, snd in H25.
      destruct tw as [[? ?] ?].
      inv_sc H25.
      subst s0 s1.
      exists (s, h0, b0).
      exists (s, h1, upd_bm b1 bn (b bn)).
      repeat (split; trivial).
      intro bnx.
      spec H2 bnx.
      unfold snd.
      unfold upd_bm in *.
      revert H2.
      case (eq_nat_dec bnx bn); auto; intros.
      subst bnx.
      destruct H3 as [? [? ?]].
      spec H0 bn n0.
      simpl in H0.
      rewrite H0.
      constructor.
    spec H20 H26.
    assert (lookup_bstate_bp bp bn0 cst0).
      red.
      remember (nth_error (bp_bars bp) bn0).
      icase e.
      destruct d.
      inversion H20.
      trivial.
    apply (bpiwp_lupbst _ _ _ _ _ _ H3) in H27.
    apply (bptrans_bstate _ _ _ _ H14) in H27; auto.
    red in H27.
    icase (nth_error (bp_bars bp'') bn0).
    destruct d.
    subst.
    trivial.
  exists (cstate sch cthds'' bp'' ap H19).
  eapply ConcMachine.c_step_bar2; eauto.
  (* case 2 *)
  assert (cs_wellformed sch cthds' bp' ap).
    copy pf.
    destruct pf0 as [? [? ?]].
    split.
    (* BEGIN thread/wp consists *)
    intros.
    case (eq_nat_dec n tid); intro.
    (* case 1 *)
    subst n.
    apply prop_ext; split; intro.
    (* case 1a *)
    destruct H17 as [hx [bx ?]].
    assert (bn <> bn0 \/ dir <> dir0 \/ mv <> mv0 -> False).
      intro.
      rewrite <- (bp_inc_wp_constant _ _ _ _ _ _ H3) in H17; auto.
      assert (exists h, exists b, lookup_slot_bp bp bn0 dir0 mv0 (tid, (h, b))) by (exists hx; exists bx; trivial).
      rewrite H14 in H19.
      destruct H19 as [stkx [hpx [bmx [cx ?]]]].
      rewrite H in H19.
      disc.
    case (eq_nat_dec bn bn0); intro.
    2: destruct H18; auto.
    case (eq_nat_dec dir dir0); intro.
    2: destruct H18; auto.
    case (eq_nat_dec mv mv0); intro.
    2: destruct H18; auto.
    subst bn0 dir0 mv0.
    apply replace_nth in H11.
    rewrite H11.
    exists stk.
    exists h0.
    exists b.
    exists k.
    trivial.
    (* case 1b *)
    destruct H17 as [stkx [hpx [bmx [cx ?]]]].
    apply replace_nth in H11.
    rewrite H11 in H17.
    inversion H17.
    subst stkx hpx bmx bn0 dir0 mv0 cx.
    apply bpiwp_luplstbp in H3.
    exists h1.
    exists b0.
    trivial.
    (* case 2 *)
    rewrite (replace_nth2 _ _ _ _ _ _ H11 n0).
    rewrite <- H14.
    apply prop_ext; 
    split; 
    intros [hx [bx ?]]; 
    exists hx; 
    exists bx; 
    rewrite (bp_inc_wp_constant _ _ _ _ _ _ H3) in *; 
    auto;
    case (eq_nat_dec bn bn0); auto;
    case (eq_nat_dec dir dir0); auto;
    case (eq_nat_dec mv mv0); auto;
    intros; subst bn0 dir0 mv0; elimtype False.
    (* case 2a *)
    apply bpiwp_luplstbp in H3.
    apply (lookup_slot_bp_fun _ _ _ _ _ _ H3) in H17.
    inversion H17.
    contradiction.
    (* case 2b *)
    generalize (bp_lupmvslt_pre _ _ _ _ _ _ _ H0 H17); intro.
    destruct (lookup_move_bp_facts _ _ _ _ _ _ H0) as [cstx [_ [? [_ [? [ps ?]]]]]].
    destruct (lupmvbp_lupstbp _ _ _ _ _ _ H0) as [bst ?].
    generalize (bp_lupsltst _ _ _ _ _ _ H22 H17); intro.
    change bp with (cs_barpool (cstate _ _ _ _ pf)) in H22.
    change cthds with (cs_thds (cstate _ _ _ _ pf)) in H.
    generalize (lupst_joins_thd _ _ _ H22 _ _ _ _ _ H); intro.
    destruct H24 as [tst ?].
    spec H20 tst (empty_stack, hx, bx) (empty_stack, h1, b0).
    unfold slot2state in *.
    assert (join_sub bst tst) by (eexists; eauto).
    detach H20.
    2: eapply join_sub_trans; eauto.
    assert (join_sub (empty_stack, h, bs) tst) by (apply join_com in H24; eexists; eauto).
    assert (join_sub (empty_stack, h1,b0) (empty_stack, h, bs)) by (exists (empty_stack, h0, b); apply join_com; repeat (split; trivial)).
    detach H20.
    2: eapply join_sub_trans; eauto.
    rewrite <- H19 in H10.
    spec H10 empty_stack.
    spec H20 H18 H10.
    inversion H20.
    subst h1 b0.
    destruct H23 as [ob ?].
    apply join_com in H23.
    destruct (join_assoc _ _ _ _ _ H23 H24) as [f [? ?]].
    apply join_com in H28.
    destruct H27 as [ot ?].
    apply join_com in H27.
    destruct (join_assoc _ _ _ _ _ H27 H28) as [f' [? ?]].
    generalize (join_self _ _ H30); intro.
    subst f'.
    assert (identity (empty_stack, hx, bx)) by (rewrite identity_unit_equiv; auto).
    apply identity_emptystate in H32; auto.
    inversion H32.
    subst hx bx.
    apply H21 in H18.
    clear -H18.
    inv_sc H18.
    generalize (split_identity _ _ H2 empty_bm_identity); intro.
    apply identity_emptybm in H5.
    subst b.
    destruct H3.
    destruct ps.
    disc.
    (* END thread/wp consists *)
    split; trivial.
    destruct H16 as [bptl [? [tw [? ?]]]].
    exists bptl.
    split.
(* hmm *)
    clear -H16 H11 H3 H4 H5 H.
    generalize (tl_states_lookup _ _ _ _ _ _ H); intro.
    assert (join (empty_stack, h0, b) (empty_stack, h1, b0) (empty_stack, h, bs)) by repeat (split; trivial).
    destruct (join_list_split_nth _ _ _ H0 H16 H1) as [bpst' [? [? [? ?]]]].
    assert (x = (tl_states cthds')).
      apply nth_error_eq.
      intro.
      case (eq_nat_dec n i); intro.
      subst.
      rewrite (replace_nth _ _ _ _ _ H2).
      generalize (replace_nth _ _ _ _ _ H11); intro.
      rewrite (tl_states_lookup _ _ _ _ _ _ H8).
      trivial.
      rewrite (replace_nth2 _ _ _ _ _ _ H2 n0).
      generalize (replace_nth2 _ _ _ _ _ _ H11 n0); intro.
      unfold tl_states.
      do 2 rewrite nth_error_map.
      rewrite H8.
      trivial.
    subst x.
    replace (bp_st bp') with bpst'; trivial.
    clear -H6 H3.
    destruct bp as [bars st wfbp].
    destruct bp' as [bars' st' wfbp'].
    unfold bp_st in *.
    destruct H3.
    apply join_com in H.
    apply (join_eq _ _ _ _ H6 H).
(* end hmm *)
    exists tw.
    split; trivial.
    intros.
    spec H18 bn0 ps cst0 H19.
    move H3 after H18.
    destruct bp.
    destruct bp'.
    simpl.
    simpl in H18.
    destruct H3.
    remember (nth_error bp_bars bn).
    icase e.
    remember (nth_error bp_bars0 bn).
    icase e.
    destruct H20.
    case (eq_nat_dec bn bn0); intro.
    subst bn0.
    rewrite (replace_nth _ _ _ _ _ H21).
    simpl.
    rewrite <- Heqe0 in H18.
    simpl in H18.
    inversion H18.
    subst cst0.
    destruct d.
    destruct d0.
    simpl in *.
    red in H20.
    destruct dbs_wp.
    destruct dbs_wp0.
    destruct H20 as [_ [? _]].
    subst; trivial.
    rewrite (replace_nth2 _ _ _ _ _ _ H21 n0).
    trivial.
  exists (cstate sch cthds' bp' ap H14).
  eapply ConcMachine.c_step_bar1; eauto.
Qed.

(* The Big Kahuna... going to be a royal pain *)

Lemma OStep_Barrier: forall o G bn bd cs dir mv pre post F st k o' st' k',
  ora_barDefs o G ->
  nth_error G bn = Some bd ->
  lookup_move bd cs dir mv pre post ->
  (sepconj pred_brk (sepconj pre F)) st ->
  step (o, st, Seq (Barrier bn) k) (Some (o', st', k')) ->
  (sepconj pred_brk (sepconj post F)) st'.
(* H1 : closedWith F (Barrier bn) ?? *)
Proof.
  intros.
  generalize (stdy_match_lookup1 _ _ H bn); intro.
  rewrite H0 in H4.
  clear H H0.
  inversion H3; clear H3; subst.
  inversion H7; clear H7; subst.
  unfold fst, snd in *.
  subst.
  simpl in H4, H6.
  subst sched0.
  apply replace_nth in H0.
  clear H.
  inversion H11.
  (* cstep-1, impossible *)
  subst.
  rewrite H0 in H10.
  inversion H10; subst.
  inversion H16; inversion H6.
  (* cstep-2, impossible *)
  subst i0 thds bp0 ap0 sched.
  rewrite H0 in H14.
  inversion H14.
  (* cstep-3 *)
  subst.
  clear pf1 pf2 pf3 pf4 pf5.
  rewrite H0 in H10.
  inversion H10; clear H10.
  subst stk0 hp0 bs0 bn0 c.
  unfold lookup_move_bp in H14.
  remember (nth_error (bp_bars bp) bn).
  icase e.
  inversion H4; clear H4.
  destruct d.
  simpl in H6.
  subst dbs_bd.
  inv_sc H2.
  subst s s0.
  rewrite <- pred_brk_ubiquitous in H3.
  spec H3 stk.
  generalize (pred_brk_precise (stk, hp, bs) (stk,h,b) (stk, ap, empty_bm)); intro.
  detach H3.
  2: exists (stk, h0, b0); repeat (split; trivial).
  detach H3.
  2: exists (stk, hp', bs); split; try apply join_emptybm; apply join_com; split; trivial; split; trivial.
  spec H H7 H3.
  inversion H.
  subst h b.
  clear H.
  apply empty_bm_identity in H6.
  subst b0.
  apply join_com in H4.
  apply (join_canc _ _ _ _ H5) in H4.
  subst h0.
  clear H3.
  assert (cs = dbs_cs).
    generalize (lookup_move_facts _ _ _ _ _ _ H14); intro.
    generalize (lookup_move_facts _ _ _ _ _ _ H1); intro.
    destruct H as [_ [_ [_ [ps1 ?]]]].
    destruct H2 as [_ [_ [_ [ps2 ?]]]].
    clear - H H2 H8 H16.
    inv_sc H8.
    subst s s0.
    inv_sc H16.
    subst s s0.
    destruct H9.
    inversion H10.
    inversion H0.
    subst.
    clear H0 H10.
    spec H (stk, h1, b1) H1.
    spec H2 (stk, h, b) H5.
    inv_sc H.
    inv_sc H2.
    subst s s0 s1 s2.
    clear H12 H17.
    clear H5 H1 H6.
    remember (bd_bn bd) as bn.
    destruct H11 as [? _].
    destruct H16 as [? _].
    simpl in H, H0.
    spec H15 bn.
    spec H10 bn.
    rewrite H in H10.
    rewrite H0 in H15.
    clear - H15 H10 H4 H8.
    spec H8 bn.
    spec H4 bn.
    destruct (join_assoc _ _ _ _ _ H15 H4) as [f [? ?]].
    destruct (join_assoc _ _ _ _ _ H10 H8) as [f2 [? ?]].
    clear - H0 H2.
    remember (bs bn).
    unfold barst_builder in *.
    destruct ps2.
    destruct ps1.
    destruct b.
    2: inversion H0.
    inversion H0;
    inversion H2;
    subst.
    inversion H7.
    trivial.
    destruct y.
    destruct H7 as [[? ?] _].
    rewrite H1 in H.
    simpl in H.
    inversion H.
    trivial.
    destruct y.
    destruct H4 as [[? ?] _].
    rewrite H1 in H.
    simpl in H.
    inversion H.
    trivial.
    destruct p.
    destruct y.
    destruct y0.
    destruct H4 as [[? ?] _].
    destruct H8 as [[? ?] _].
    rewrite H4 in H3.
    rewrite H1 in H.
    simpl in H3, H.
    subst.
    trivial.
  subst dbs_cs.
  assert (rightdirection = dir).
    destruct (lookup_move_only1DirOK _ _ _ _ _ _ _ _ _ _ H1 H14); auto.
    elimtype False.
    change False with (Fls (stk, hp', bs)).
    rewrite <- H.
    split.
    inv_sc H8.
    subst s s0.
    exists (stk, h, b).
    exists (stk, h0, b0).
    repeat (split; trivial).
    inv_sc H16.
    subst s s0.
    destruct H9.
    exists (stk, h, b).
    exists (stk, h0, b0).
    repeat (split; trivial).
  subst rightdirection.
  assert (rightmove = mv).
    case (eq_nat_dec rightmove mv); trivial; intro.
    elimtype False.
    generalize (so_barrier_stuck i (cstate sched thds' bp' ap pf') cs'' H12); intro.
    spec H stk hp'' bs'' bn dir.
    spec H rightmove k.
    detach H.
    2: eapply replace_nth; eauto.
    spec H pre0 post0 mv pre post.
    spec H n.
    unfold cs_barpool in H.
    detach H.
    2: eapply bpiwp_lupmvbp; eauto; red; rewrite <- Heqe; trivial.
    detach H.
    2: eapply bpiwp_lupmvbp; eauto; red; rewrite <- Heqe; trivial.
    spec H (empty_stack, hp', bs).
    detach H.
    2: inv_sc H16; destruct H6; exists (empty_stack, h, b); exists (empty_stack, h0, b0); repeat (split; trivial).
    2: apply lookup_move_facts in H14; destruct H14 as [? _]; rewrite <- H14 in H10; spec H10 empty_stack; trivial.
    detach H.
    2: inv_sc H8; exists (empty_stack, h, b); exists (empty_stack, h0, b0); repeat (split; trivial).
    2: apply lookup_move_facts in H1; destruct H1 as [? _]; rewrite <- H1 in H6; spec H6 empty_stack; trivial.
    spec H (i, (hp'0, bs')).
    detach H.
    2: apply join_com; inv_sc H16; destruct H6; inversion H6; inversion H9; subst s h b s0 h0 b0; repeat (split; trivial).
    apply H.
    eapply bpiwp_luplstbp; eauto.
  subst rightmove.
  apply (lookup_move_fun _ _ _ _ _ _ _ _ H1) in H14.
  destruct H14.
  subst pre0 post0.
  inv_sc H16.
  subst s s0.
  inversion H9; clear H9.
  subst h0 b0.
  destruct H6.
  inversion H; clear H.
  subst h b.
  inv_sc H8.
  subst s s0.
  assert (hp'0 = h /\ bs' = b).
    apply lookup_move_facts in H1.
    destruct H1 as [_ [_ [? _]]].
    spec H (stk, hp', bs) (stk, h, b) (stk, hp'0, bs').
    detach H.
    2: exists (stk, h0, b0); repeat (split; trivial).
    detach H.
    2: exists (stk, hp'', bs''); repeat (split; trivial).
    spec H H10 H2.
    inversion H.
    auto.
  destruct H.
  subst h b.
  apply join_com in H3.
  apply join_com in H8.
  apply join_com in H9.
  apply join_com in H4.
  apply (join_canc _ _ _ _ H3) in H8.
  apply (join_canc _ _ _ _ H4) in H9.
  subst h0 b0.
  clear H10.
  generalize (so_waiting _ _ _ H12 stk hp'' bs'' bn dir mv k pre post); intro.
  detach H.
  2: eapply bpiwp_lupmvbp; eauto; red; rewrite <- Heqe; trivial.
  detach H.
  2: eapply replace_nth; eauto.
  clear -H1 H14 H H13.
  inversion H13; subst.
  apply replace_nth in H2.
  destruct H as [hx [bx [hx' [bx' [? [? [? ?]]]]]]].
  simpl in H7.
  rewrite H7 in H2.
  inversion H2; subst.
  exists (stk0, ap, empty_bm).
  exists (stk0, hp', bs).
  split.
  split.
  2: apply join_emptybm.
  apply join_com.
  do 2 (split; trivial).
  split.
  rewrite <- pred_brk_ubiquitous in H3.
  spec H3 stk0.
  trivial.
  exists (stk0, hx, bx).
  exists (stk0, hp'', bs'').
  split.
  apply join_com.
  repeat (split; trivial).
  split; trivial.
  apply lookup_move_facts in H1.
  destruct H1 as [_ [? _]].
  rewrite <- H1 in H6.
  spec H6 stk0.
  trivial.
  
  (* cstep-4 *)
  (* begin similar to cstep-3 *)
  subst.
  clear pf1 pf2 pf3 pf4 pf5.
  rewrite H0 in H10.
  inversion H10; clear H10.
  subst stk0 hp0 bs0 bn0 c.
  unfold lookup_move_bp in H14.
  remember (nth_error (bp_bars bp) bn).
  icase e.
  inversion H4; clear H4.
  destruct d.
  simpl in H6.
  subst dbs_bd.
  inv_sc H2.
  subst s s0.
  rewrite <- pred_brk_ubiquitous in H3.
  spec H3 stk.
  generalize (pred_brk_precise (stk, hp, bs) (stk,h,b) (stk, ap, empty_bm)); intro.
  detach H3.
  2: exists (stk, h0, b0); repeat (split; trivial).
  detach H3.
  2: exists (stk, hp', bs); split; try apply join_emptybm; apply join_com; split; trivial; split; trivial.
  spec H H7 H3.
  inversion H.
  subst h b.
  clear H.
  apply empty_bm_identity in H6.
  subst b0.
  apply join_com in H4.
  apply (join_canc _ _ _ _ H5) in H4.
  subst h0.
  clear H3.
  assert (cs = dbs_cs).
    generalize (lookup_move_facts _ _ _ _ _ _ H14); intro.
    generalize (lookup_move_facts _ _ _ _ _ _ H1); intro.
    destruct H as [_ [_ [_ [ps1 ?]]]].
    destruct H2 as [_ [_ [_ [ps2 ?]]]].
    clear - H H2 H8 H15. (* vs H16 *)
    inv_sc H8.
    subst s s0.
    inv_sc H15. (* vs H16 *)
    subst s s0.
    destruct H9.
    inversion H10.
    inversion H0.
    subst.
    clear H0 H10.
    spec H (stk, h1, b1) H1.
    spec H2 (stk, h, b) H5.
    inv_sc H.
    inv_sc H2.
    subst s s0 s1 s2.
    clear H12 H17.
    clear H5 H1 H6.
    remember (bd_bn bd) as bn.
    destruct H11 as [? _].
    destruct H16 as [? _].
    simpl in H, H0.
    spec H15 bn.
    spec H10 bn.
    rewrite H in H10.
    rewrite H0 in H15.
    clear - H15 H10 H4 H8.
    spec H8 bn.
    spec H4 bn.
    destruct (join_assoc _ _ _ _ _ H15 H4) as [f [? ?]].
    destruct (join_assoc _ _ _ _ _ H10 H8) as [f2 [? ?]].
    clear - H0 H2.
    remember (bs bn).
    unfold barst_builder in *.
    destruct ps2.
    destruct ps1.
    destruct b.
    2: inversion H0.
    inversion H0;
    inversion H2;
    subst.
    inversion H7.
    trivial.
    destruct y.
    destruct H7 as [[? ?] _].
    rewrite H1 in H.
    simpl in H.
    inversion H.
    trivial.
    destruct y.
    destruct H4 as [[? ?] _].
    rewrite H1 in H.
    simpl in H.
    inversion H.
    trivial.
    destruct p.
    destruct y.
    destruct y0.
    destruct H4 as [[? ?] _].
    destruct H8 as [[? ?] _].
    rewrite H4 in H3.
    rewrite H1 in H.
    simpl in H3, H.
    subst.
    trivial.
  subst dbs_cs.
  assert (rightdirection = dir).
    destruct (lookup_move_only1DirOK _ _ _ _ _ _ _ _ _ _ H1 H14); auto.
    elimtype False.
    change False with (Fls (stk, hp', bs)).
    rewrite <- H.
    split.
    inv_sc H8.
    subst s s0.
    exists (stk, h, b).
    exists (stk, h0, b0).
    repeat (split; trivial).
    inv_sc H15. (* vs H16 *)
    subst s s0.
    destruct H9.
    exists (stk, h, b).
    exists (stk, h0, b0).
    repeat (split; trivial).
  subst rightdirection.
  assert (rightmove = mv).
    (* not the same! *)
    inv_sc H15.
    subst s s0.
    destruct H6.
    inversion H.
    inversion H9.
    subst h b h0 b0.
    clear H H9.
    assert (lookup_move_bp bp bn dir mv pre post) by (red; rewrite <- Heqe; trivial).
    assert (lookup_move_bp bp bn dir rightmove pre0 post0) by (red; rewrite <- Heqe; trivial).
    generalize (bpiwp_lupmvbp _ _ _ _ _ _ _ _ _ _ _ H16 H); intro.
    generalize (bpiwp_lupmvbp _ _ _ _ _ _ _ _ _ _ _ H16 H6); intro.
    generalize (bpiwp_luplstbp _ _ _ _ _ _ H16); intro.
    destruct (lupmvbp_lupstbp _ _ _ _ _ _ H) as [st ?].
    change bp with (cs_barpool (cstate _ _ _ _ pf)) in H17.
    change cthds with (cs_thds (cstate _ _ _ _ pf)) in H0.
    generalize (lupst_joins_thd _ _ _ H17 _ _ _ _ _ H0); intro.
    unfold cs_barpool, cs_thds in *.
    destruct H22 as [stx ?].
    assert (join (empty_stack, hp'0, bs') (empty_stack, hp'', bs'') (empty_stack, hp', bs)) by repeat (split; trivial).
    apply join_com in H22.
    apply join_com in H23.
    destruct (join_assoc _ _ _ _ _ H23 H22) as [st'' [? ?]].
    apply join_com in H24.
    generalize (bpiwp_state _ _ _ H17 (i, (hp'0, bs')) _ H24 _ _ _ H16); intro.
    clear -H9 H10 H15 H26 H25 H24 H2 H3 H4 H8 H19.
    inv_sc H8.
    subst s s0.
    generalize (stuck_not_bp_ready _ _ _ _ _ _ _ _ _ H19 H10 H9 _ H26 _ H15 (empty_stack, hp'', bs'') (empty_stack, hp', bs)); intro.
    detach H.
    2: apply join_com; repeat (split; trivial).
    detach H.
    2: exists stx; trivial.
    apply H.
    exists (empty_stack, h, b).
    exists (empty_stack, h0, b0).
    repeat (split; trivial).
    apply lookup_move_bp_facts in H9.
    destruct H9 as [? [? [? ?]]].
    rewrite <- H8 in H6.
    spec H6 empty_stack.
    auto.
    (* end not the same *)
  subst rightmove.
  apply (lookup_move_fun _ _ _ _ _ _ _ _ H1) in H14.
  destruct H14.
  subst pre0 post0.
  inv_sc H15. (* vs H16 *)
  subst s s0.
  inversion H9; clear H9.
  subst h0 b0.
  destruct H6.
  inversion H; clear H.
  subst h b.
  inv_sc H8.
  subst s s0.
  assert (hp'0 = h /\ bs' = b).
    apply lookup_move_facts in H1.
    destruct H1 as [_ [_ [? _]]].
    spec H (stk, hp', bs) (stk, h, b) (stk, hp'0, bs').
    detach H.
    2: exists (stk, h0, b0); repeat (split; trivial).
    detach H.
    2: exists (stk, hp'', bs''); repeat (split; trivial).
    spec H H10 H2.
    inversion H.
    auto.
  destruct H.
  subst h b.
  apply join_com in H3.
  apply join_com in H8.
  apply join_com in H9.
  apply join_com in H4.
  apply (join_canc _ _ _ _ H3) in H8.
  apply (join_canc _ _ _ _ H4) in H9.
  subst h0 b0.
  clear H10.
  (* end similar to cstep-3 *)
  generalize (transition_threads_changes _ _ _ H21); intro.
  detach H.
  assert (lookup_move_bp bp bn dir mv pre post).
    red.
    rewrite <- Heqe.
    trivial.
  generalize (bpiwp_lupmvbp _ _ _ _ _ _ _ _ _ _ _ H16 H6); intro.
  generalize (bpiwp_luplstbp _ _ _ _ _ _ H16); intro.
  destruct (lookup_slot_out _ _ _ _ H20 _ _ _ _ H8 _ H9) as [sltx [? [? ?]]].
  spec H mv sltx H10.
  destruct sltx as [? [? ?]].
  unfold slot2tid, fst in *.
  subst i.
  apply replace_nth in H18.
  rewrite H18 in H.
  spec H stk hp'' bs'' bn.
  spec H dir mv k.
  detach H; auto.
  destruct H as [hx [bx [? ?]]].
  change thds'' with (cs_thds (cstate _ _ _ _ pf')) in H15.
  generalize (thread_so_constants1 _ _ _ _ _ _ _ H15 H12); intro.
  inversion H13; subst.
  simpl in H22.
  apply replace_nth in H24.
  rewrite H24 in H22.
  inversion H22.
  subst stk0 hp'1 bs0 k'.
  apply join_com in H26.
  exists (stk, ap0, empty_bm).
  exists (stk, hx, bx).
  rewrite <- pred_brk_ubiquitous in H25.
  spec H25 stk.
  split; trivial.
  repeat (split; trivial).
  apply join_emptybm.
  split; trivial.
  exists (stk, h, b).
  exists (stk, hp'', bs'').
  split.
  unfold slot2state in H.
  destruct H as [[? ?] ?].
  unfold fst, snd in *.
  repeat (split; trivial).
  split; trivial.
  apply lookup_move_bp_facts in H6.
  destruct H6 as [? [_ [_ [? _]]]].
  rewrite <- H6 in H17.
  spec H17 stk.
  trivial.
  eapply bpt_uni_out; eauto.
  eapply bpi_uni_nin_uni; eauto.
  change bp with (cs_barpool (cstate _ _ _ _ pf)).
  apply cs_slots_unique.
  change bp with (cs_barpool (cstate _ _ _ _ pf)).
  eapply cs_running_nin_bp; eauto.
Qed.

End ConcOracle.

Module O := ConcOracle.

Module OM := SeqStep(O).
Export OM.

Lemma ostep_sstep: forall st O cmd st' O' cmd',
  seq_step (st, O, cmd) (st', O', cmd') ->
  SM.seq_step (st, tt, cmd) (st', tt, cmd') \/ exists bn, exists cmd'', cmd = Seq (Barrier bn) cmd''.
Proof.
  inversion 1; subst; 
  try (left; econstructor; eauto; discriminate);
  right; eexists; eexists; eauto.
Qed.

Lemma sstep_ostep: forall st cmd st' cmd',
  SM.seq_step (st,tt,cmd) (st',tt,cmd') -> 
  forall O,
    seq_step (st, O, cmd) (st', O, cmd').
Proof.
  inversion 1; subst; econstructor; eauto; contr.
Qed.
