(* CSL with Barriers, v0.1
 * 
 * It is our intention to release the files under a BSD-style license shortly
 * after some badly-needed cleanup work,  
 * 
 * Until then:
 * (c) 2010 Aquinas Hobor, Cristian Gherghina, and Christopher Chak
 * All Rights Reserved
 *
 * Permission granted to use these proof scripts only for the purpose of 
 * evaluating "Barriers in Concurrent Separation Logic."
 *)

Require Import SLB_Base.
Require Import SLB_Lang.
Require Import SLB_BarDefs.
Require Import SLB_CLang.
Require Import SLB_SSem.
Require Import SLB_Sem.
Require Import SLB_OSem.
Require Import SLB_HRules.

Import ConcMachine.
Import ConcOracle.

Inductive conc_stepstar : Cstate -> Cstate -> Prop :=
  | cst_Z : forall cs,
      conc_stepstar cs cs
  | cst_S : forall cs cs' cs'',
      conc_step cs cs' ->
      conc_stepstar cs' cs'' ->
      conc_stepstar cs cs''.

Definition sched_halted (cs : Cstate) : Prop :=
  match cs_sched cs with
   | nil => True
   | i :: sch => 
      match nth_error (cs_thds cs) i with
       | None => True
       | Some t => 
         match th_ctl t with
           | Running _ => False
           | Waiting _ _ _ _ => True
         end
      end
  end.

(*
Definition thds_halted (cs : Cstate) : Prop :=
  forall i thd,
    nth_error (cs_thds cs) i = Some thd ->
      th_ctl thd = Running Unit.
*)

Definition can_cstep (cs : Cstate) : Prop :=
  exists cs',
    conc_step cs cs'.

Definition isafe (cs : Cstate) : Prop :=
  sched_halted cs \/
(*  thds_halted cs \/ *)
  can_cstep cs.

Definition cstate_safe (cs : Cstate) : Prop :=
  forall cs',
    conc_stepstar cs cs' ->
      isafe cs'.

Definition all_safe (cs : Cstate) : Prop :=
  forall i cs',
    stepOthers i cs cs' ->
      forall ora st ctl,
        Ora2CM ora st (Running ctl) i cs' ->
	  safe_ora st ora ctl.

Lemma progress: forall cs,
  all_safe cs ->
  isafe cs.
Proof.
  intros.
  destruct cs as [sch thds bp ap wfcs].
  destruct sch.
  left.
  exact I.
  remember (nth_error thds n).
  destruct e.
  2: left; red; simpl; rewrite <- Heqe; trivial.
  destruct t0 as [stk hp bs ctl].
  destruct ctl.
  2: left; red; simpl; rewrite <- Heqe; simpl; trivial.
  spec H n (cstate (n :: sch) thds bp ap wfcs).
  symmetry in Heqe.
  detach H.
  2: constructor; split; trivial; eexists; split; eauto; eexists; reflexivity.
  change thds with (cs_thds (cstate (n :: sch) thds bp ap wfcs)) in Heqe.
  destruct (Ora2CM_tot2 _ _ _ _ _ _ Heqe) as [O [st ?]].
  simpl in Heqe.
  spec H O st c H0.
  spec H st O c.
  detach H.
  2: constructor.
  right.
  destruct H.
  2: compute in H; subst c; eexists; eapply c_step_texit; eauto.
  destruct H as [[[st' O'] cmd'] ?].
  destruct (ostep_sstep _ _ _ _ _ _ H).
  (* Regular steps *)
  clear H.
  inversion H0; subst.
  clear pf0 pf1 pf2 pf3.
  generalize (replace_nth _ _ _ _ _ H10); intro.
  rewrite Heqe in H.
  inversion H.
  subst stk0 hp' bs0.
  clear H.
  destruct st' as [[? ?] ?].
  destruct (smsstep_heap _ _ _ _ _ _ _ _ _ _ H11 H12 H1) as [hp' [ap' [? [? ?]]]].
  destruct (can_replace _ _ _ _ Heqe (thread s hp' b (Running cmd'))) as [thds' ?].
  generalize (smsstep_bs _ _ _ _ _ _ _ _ H1); intro.
  subst b.
  assert (cs_wellformed (n:: sch) thds' bp ap').
    copy wfcs.
    destruct wfcs0 as [? [? ?]].
    split.
    intros.
    spec H5 bn dir mv tid.
    rewrite H5.
    case (eq_nat_dec n tid); intro.
    subst n.
    apply replace_nth in H4.
    rewrite H4.
    rewrite Heqe.
    apply prop_ext; split; intros [? [? [? [? ?]]]]; disc.
    rewrite (replace_nth2 _ _ _ _ _ _ H4 n0).
    trivial.
    split; trivial.
    destruct H7 as [bptl [? [tw [? ?]]]].
    generalize (tl_states_lookup _ _ _ _ _ _ Heqe); intro.
    assert (join empty_state (empty_stack, hp, bs) (empty_stack, hp, bs)).
      apply join_emptystate; trivial.
    destruct (join_list_split_nth _ _ _ H14 H7 H15) as [s1 [nl [? [? ?]]]].
    destruct (join_list_assoc2 H18 H8) as [s2 [? ?]].
    apply join_com in H19.
    destruct (join_assoc _ _ _ _ _ H17 H19) as [s3 [? ?]].
    destruct s3 as [[es hpx] bs'].
    destruct H21.
    apply join_com in H23.
    apply empty_bm_identity in H23.
    destruct H21 as [[? ?] ?].
    apply (join_eq _ _ _ _ H12) in H25.
    unfold fst, snd in *.
    subst es bs' hp0.
    clear H21. (* H21 : empty_stack = empty_stack *)
    remember (bp_st bp).

    assert (sub_silhouette (empty_stack, h, bs) (empty_stack, hpx, bs)) by
      (apply sub_silhouette_prod1; apply sub_silhouette_prod2; trivial).
    assert (joins s0 (empty_stack, hpx, bs)) by (eexists; eauto).
    copy H21.
    spec H21 s0.
    spec H21 H23.
    destruct H21 as [s3 ?].
    apply join_com in H21.
    assert (join (empty_stack, hp', bs) (empty_stack, ap', empty_bm) (empty_stack, h, bs)).
      split.
      2: apply join_com; apply join_emptybm.
      split; trivial.
      split; trivial.
    generalize (sub_silhouette_join _ _ H24 H21 (join_com _ _ _ H22)); intro.
    destruct (join_list_sub_silhouette H20 H26) as [tw' [? ?]].
    apply join_com in H25.
    destruct (join_assoc _ _ _ _ _ H25 H21) as [s4 [? ?]].
    destruct (join_list_assoc1 H30 H28) as [s5 [? ?]].
    generalize H16; intro H16cpy.
    apply replace_nth in H16.
    destruct (join_list_assoc1 H29 H31) as [s6 [? ?]].
    apply join_com in H34.
    destruct (join_list_join_nth _ _ _ H16 H33 H34) as [stx [nthds'' [? [? ?]]]].
    apply empty_state_identity in H35.
    subst stx.
    assert (tl_states thds' = nthds'').
      clear -H36 H16cpy H4.
      (* probably better to do with lemmas *)
      apply nth_error_eq; intro.
      case (eq_nat_dec n i); intro.
      subst i.
      apply replace_nth in H4.
      apply replace_nth in H16cpy.
      apply replace_nth in H36.
      rewrite H36.
      unfold tl_states.
      rewrite nth_error_map.
      rewrite H4.
      trivial.
      generalize (replace_nth2 _ _ _ _ _ _ H36 n0).
      generalize (replace_nth2 _ _ _ _ _ _ H16cpy n0).
      generalize (replace_nth2 _ _ _ _ _ _ H4 n0).
      intros.
      rewrite H1, H0.
      unfold tl_states.
      do 2 rewrite nth_error_map.
      rewrite H.
      trivial.
      (* end do with lemmas *)
    subst nthds''.
    exists s5.
    split; trivial.
    exists tw'.
    split; trivial.
    intros.
    apply H13 with ps.
    clear -H32 H37 H7 H8 H4 Heqe H35 H6 H3.
    (* probably this is not the most elegant way, e.g., cross-split, but 
    it gets the job done *)
    destruct tw' as [[ts' th'] tb'].
    destruct tw as [[ts th] tb].
    replace tb with tb'.
    inv_sc H35.
    subst s1 s2.
    exists (ts, empty_heap, b).
    exists (ts, th, b0).
    split.
    split; trivial.
    split.
    split; trivial.
    apply join_emptyheap.
    split; trivial.
    rewrite <- ubiquitous_barrier_assert in H5.
    spec H5 ts.
    destruct H5 as [? [? ?]].
    split; trivial.
    split; trivial.
    destruct bptl as [[? ?] tb''].
    destruct s5 as [[? ?] tb'''].
    destruct H32 as [[[? ?] ?] ?].
    destruct H8 as [[[? ?] ?] ?].
    apply empty_bm_identity in H2.
    apply empty_bm_identity in H10.
    unfold fst, snd in *.
    subst s2 ts' s1 ts tb''' tb''.
    clear -H37 H7 H4 Heqe.
    destruct s0 as [[? ?] ?].
    generalize (join_list_stk _ _ _ H37).
    unfold st_stk, fst.
    intro.
    subst s0.
    apply joinlistState_joinlistBm in H37.
    apply joinlistState_joinlistBm in H7.
    unfold tl_states in *.
    rewrite map_comp' in *.
    change (fun x => snd (empty_stack, th_hp x, th_bs x)) with th_bs in H37.
    change (fun x => snd (empty_stack, th_hp x, th_bs x)) with th_bs in H7.
    replace (map th_bs thds') with (map th_bs thds) in H37.
    eapply join_list_eq; eauto.
    apply nth_error_eq; intro.
    do 2 rewrite nth_error_map.
    case (eq_nat_dec n i); intro.
    subst i.
    apply replace_nth in H4.
    rewrite H4.
    rewrite Heqe.
    trivial.
    rewrite (replace_nth2 _ _ _ _ _ _ H4 n0).
    trivial.
    (* end not the most elegant *)
  exists (cstate (n::sch) thds' bp ap' H5).
  eapply c_step_seq; eauto.

  (* Oracle steps *)
  destruct H1 as [bn [cmd'' ?]].
  subst c.
  inversion H; subst.
  (* ostep1 *)
  inversion H2; subst.
  unfold fst, snd in *.
  assert (i = n).
    destruct O.
    inversion H0; simpl in *; subst.
    inversion H7.
    trivial.
  subst i.
  apply (Ora2CM_fun1 _ _ _ _ _ _ H0) in H6.
  subst cs.
  exists cs'.
  trivial.
  (* ostep2 *)
  inversion H2; subst.
  destruct O'.
  inversion H0; subst.
  disc.
  unfold fst, snd in *.
  assert (i = n).
    destruct O'.
    inversion H0; subst; simpl in *.
    inversion H1.
    trivial.
  subst i.
  contradiction H3.
  eexists.
  apply H0.
  assert (i = n).
    destruct O'.
    inversion H0; subst; simpl in *.
    inversion H1.
    trivial.
  subst i.
  exists cs'.
  apply (Ora2CM_fun1 _ _ _ _ _ _ H0) in H3.
  subst cs.
  trivial.
Qed.

Lemma preservation: forall cs cs',
  all_safe cs ->
  conc_step cs cs' ->
  all_safe cs'.
Proof.
  intros.
  remember (cs_sched cs) as sch.
  symmetry in Heqsch.
  destruct sch.
  inversion H0; subst; disc.
  do 3 intro.
  case (eq_nat_dec n i); intro.
  2: spec H i cs'0; apply H; apply so_step with cs'; auto; unfold ready_at; rewrite Heqsch; intros [? _]; contr.

  subst i.
  destruct (cstep_head_running _ _ H0 n) as [c ?].
    rewrite Heqsch; reflexivity.
  remember (nth_error (cs_thds cs) n).
  symmetry in Heqe.
  icase e.
  inversion H2.
  clear H2.
  spec H n cs. (* no choice about cs *)
  detach H.
  2: constructor; red; rewrite Heqsch; split; trivial; eexists; split; eauto.
  destruct t0 as [stk hp bs cctl].
  destruct (Ora2CM_tot2 _ _ _ _ _ _ Heqe) as [ora1 [st1 ?]].
  simpl in H4.
  subst cctl.
  spec H ora1 st1 c H2.
  intros.
  inversion H2; subst.
  inversion H3; subst.
  inversion H0; subst; inversion Heqsch; clear Heqsch.

  (* cstep1 *)
  subst i omg.
  inversion H1; subst.
  2: destruct H12; split; trivial; apply replace_nth in H22; eexists; split; eauto; eexists; split; eauto; eexists; reflexivity.
  replace othds0 with othds.
  2: generalize (replace_replace _ _ _ _ _ _ H8 H9); generalize (replace_replace _ _ _ _ _ _ H4 H5); rewrite (replace_eq _ _ _ _ _ H22); congruence.
  eapply safe_ora_step_safe_ora; eauto.
  apply sstep_ostep.
  apply replace_nth in H9.
  simpl in Heqe.
  apply replace_nth in H22.
  apply replace_nth in H5.
  rewrite H22 in H9.
  inversion H9.
  subst.
  rewrite H16 in Heqe.
  inversion Heqe.
  subst.
  rewrite H16 in H5.
  inversion H5.
  subst.
  apply (join_eq _ _ _ _ H7) in H17.
  apply (join_eq _ _ _ _ H11) in H20.
  subst.
  trivial.
  (* cstep2 *)
  subst n.
  rename i into n.
  subst sch.
  apply replace_nth in H9.
  apply replace_nth in H5.
  simpl in Heqe.
  rewrite Heqe in H5.
  inversion H5.
  subst stk0 hp' bs0.
  rewrite H17 in Heqe.
  inversion Heqe.
  subst stk2 hp2 bs2 c.
  change cthds with (cs_thds (cstate omg cthds bp ap pf2)) in H17.
  generalize (thread_so_constants1 _ _ _ _ _ _ _ H17 H1); intro.
  simpl in H12.
  rewrite H9 in H12.
  inversion H12.
  subst.
  apply Unit_safe_ora.
  (* cstep3 *)
  subst i omg.
  apply replace_nth in H5.
  simpl in Heqe.
  rewrite Heqe in H5.
  inversion H5; clear H5.
  subst stk0 hp' bs0.
  rewrite H16 in Heqe.
  inversion Heqe; clear Heqe.
  subst.
  eapply safe_ora_step_safe_ora; eauto.
  apply SeqStep_BarStep.
  eapply os_stepgood; eauto.
  simpl.
  reflexivity.
  (* cstep4, identical to cstep3 *)
  subst i omg.
  apply replace_nth in H5.
  simpl in Heqe.
  rewrite Heqe in H5.
  inversion H5; clear H5.
  subst stk0 hp' bs0.
  rewrite H16 in Heqe.
  inversion Heqe; clear Heqe.
  subst.
  eapply safe_ora_step_safe_ora; eauto.
  apply SeqStep_BarStep.
  eapply os_stepgood; eauto.
  simpl.
  reflexivity.
Qed.

Theorem safety: forall cs,
  all_safe cs ->
  cstate_safe cs.
Proof.
  repeat intro.
  induction H0.
  apply progress; trivial.
  apply IHconc_stepstar.
  apply preservation with cs; trivial.
Qed.
