(* CSL with Barriers, v0.1
 * 
 * It is our intention to release the files under a BSD-style license shortly
 * after some badly-needed cleanup work,  
 * 
 * Until then:
 * (c) 2010 Aquinas Hobor, Cristian Gherghina, and Christopher Chak
 * All Rights Reserved
 *
 * Permission granted to use these proof scripts only for the purpose of 
 * evaluating "Barriers in Concurrent Separation Logic."
 *)

Require Import SLB_Base.
Require Import SLB_Lang.
Require Import SLB_BarDefs.
Require Import SLB_CLang.
Require Import SLB_SSem.

Module ConcMachine.

Inductive conc_step : Cstate -> Cstate -> Prop :=

  | c_step_seq : forall i omg stk stk' hp' hp'' ctl' hp hp''' bs ctl thds bp ap thds' ap' pf pf',
    nth_error thds i = Some (thread stk hp bs (Running ctl)) ->
    join hp ap hp' ->
    SM.seq_step ((stk, hp', bs), tt, ctl) ((stk', hp'', bs), tt, ctl') ->
    join hp''' ap' hp'' ->
    pred_brk (empty_stack, ap', empty_bm) ->
    replace thds i (thread stk' hp''' bs (Running ctl')) = Some thds' ->
(* ------------------------------------------------------------*)
    conc_step (cstate (i :: omg) thds bp ap pf)
                       (cstate (i :: omg) thds' bp ap' pf')

  | c_step_texit : forall i omg stk hp bs thds bp ap pf,
    nth_error thds i = Some (thread stk hp bs (Running Unit)) ->
(* ------------------------------------------------------------*)
    conc_step (cstate (i :: omg) thds bp ap pf)
                       (cstate omg thds bp ap pf)

  | c_step_bar1 : forall i omg thds bp ap pf stk hp bn c bs
                                     rightdirection rightmove pre post
                                     hp' hp'' bs' bs''
                                     thds' bp' pf', 
    nth_error thds i = Some (thread stk hp bs (Running (Seq (Barrier bn) c))) ->
    lookup_move_bp bp bn rightdirection rightmove pre post ->
    (sepconj (and (exact (stk, hp', bs')) pre) (exact (stk, hp'', bs''))) (stk, hp, bs) ->
    bp_inc_waitpool bp bn rightdirection rightmove (i, (hp', bs')) bp' ->
    replace thds i (thread stk hp'' bs'' (Waiting bn rightdirection rightmove c)) = Some thds' ->
    ~bp_ready bp' bn ->
(* ------------------------------------------------------------*)
    conc_step (cstate (i :: omg) thds bp ap pf)
                       (cstate omg thds' bp' ap pf')

  | c_step_bar2 : forall i omg thds bp ap pf stk hp hp' hp'' bn c bs bs' bs''
                                     rightdirection rightmove pre post
                                     thds' thds'' bp' bp'' out pf', 
    nth_error thds i = Some (thread stk hp bs (Running (Seq (Barrier bn) c))) ->
    lookup_move_bp bp bn rightdirection rightmove pre post ->
    (sepconj (and (exact (stk, hp', bs')) pre) (exact (stk, hp'', bs''))) (stk, hp, bs) ->
    bp_inc_waitpool bp bn rightdirection rightmove (i, (hp', bs')) bp' ->
    replace thds i (thread stk hp'' bs'' (Waiting bn rightdirection rightmove c)) = Some thds' ->
    bp_ready bp' bn ->
    bp_transition bp' bn bp'' out ->
    transition_threads out thds' thds'' ->
(* ------------------------------------------------------------*)
    conc_step (cstate (i :: omg) thds bp ap pf)
                       (cstate omg thds'' bp'' ap pf').

Lemma cs_threads_preserved_barrier : forall cs i stk h bs bn c cs',
  nth_error (cs_thds cs) i = Some (thread stk h bs (Running (Seq (Barrier bn) c))) ->
  conc_step cs cs' ->
  head (cs_sched cs) = Some i -> 
  (exists h', exists bs', exists dir, exists mv, nth_error (cs_thds cs') i = Some (thread stk h' bs' (Waiting bn dir mv c))) \/
  (exists h', exists bs', nth_error (cs_thds cs') i = Some (thread stk h' bs' (Running c))).
Proof.
  intros.
  inversion H0; subst.
  simpl in *.
  inversion H1.
  subst i.
  rewrite H in H2.  
  inversion H2; subst.
  inversion H4; contr.
  simpl in H, H1.
  inversion H1.
  subst i0.
  rewrite H in H2.
  inversion H2.  
  simpl in H1.
  inversion H1.
  subst i0.
  left.
  exists hp''.
  exists bs''.
  exists rightdirection.
  exists rightmove.
  simpl.
  rewrite (replace_nth _ _ _ _ _ H6).
  simpl in H.
  rewrite H in H2.
  inversion H2.
  subst.
  trivial.
  simpl in H.
  simpl in H1.
  inversion H1; subst.
  rewrite H in H2.
  inversion H2; subst.
  right.
  simpl.
  generalize (transition_threads_changes _ _ _ H9); intro.
  apply replace_nth in H6.
  detach H10.
  generalize (bpiwp_lupmvbp _ _ _ _ _ _ _ _ _ _ _ H5 H3); intro.
  generalize (bpiwp_luplstbp _ _ _ _ _ _ H5); intro.
  destruct (lookup_slot_out _ _ _ _ H8 _ _ _ _ H11 _ H12) as [sltx [? [? ?]]].
  spec H10 rightmove sltx H13.
  destruct sltx as [? [? ?]].
  unfold slot2tid, fst in *.
  subst i.
  rewrite H6 in H10.
  spec H10 stk0 hp'' bs'' bn0.
  spec H10 rightdirection rightmove c0.
  detach H10; auto.
  destruct H10 as [hx [bx [? ?]]].
  exists hx.
  exists bx.
  trivial.
  eapply bpt_uni_out; eauto.
  eapply bpi_uni_nin_uni; eauto.
  change bp with (cs_barpool (cstate _ _ _ _ pf)).
  apply cs_slots_unique.
  change bp with (cs_barpool (cstate _ _ _ _ pf)).
  eapply cs_running_nin_bp; eauto.
Qed.

Lemma bptrans_lupmv: forall bp bn dir mv pre post,
  lookup_move_bp bp bn dir mv pre post ->
  forall bn' bp' out,
  bn <> bn' ->
  bp_transition bp bn' bp' out ->
  lookup_move_bp bp' bn dir mv pre post.
Proof.
  intros.
  generalize (bptrans_bstate _ _ _ _ H1 bn); intro.
  apply barDefs_bp_transition_preserved in H1.
  apply (eq_nth_error bn) in H1.
  unfold bp_bardefs in H1.
  do 2 rewrite nth_error_map in H1.
  detach H2.
  2: intro; subst; apply H0; trivial.
  unfold lookup_move_bp, lookup_bstate_bp in *.
  icase (nth_error (bp_bars bp) bn).
  destruct d.
  spec H2 dbs_cs.
  detach H2; trivial.
  icase (nth_error (bp_bars bp') bn).
  destruct d.
  subst dbs_cs0.
  inversion H1.
  subst dbs_bd0.
  trivial.
Qed.

Lemma cstep_waiting_inv : forall cs i stk hp bs bn dir mv k pre post cs',
  lookup_move_bp (cs_barpool cs) bn dir mv pre post ->
  nth_error (cs_thds cs) i = Some (thread stk hp bs (Waiting bn dir mv k)) ->
  conc_step cs cs' ->
  ((lookup_move_bp (cs_barpool cs') bn dir mv pre post /\
    nth_error (cs_thds cs') i = Some (thread stk hp bs (Waiting bn dir mv k))) \/
   exists hp', exists bs', exists hp'', exists bs'',
     join bs bs' bs'' /\
     join hp hp' hp'' /\
     post (empty_stack, hp', bs') /\
     nth_error (cs_thds cs') i = Some (thread stk hp'' bs'' (Running k))).
Proof.
  intros.
  inversion H1; subst; unfold cs_thds, cs_barpool in *.
  (* seq *)
  left.
  split; trivial.
  case (eq_nat_dec i0 i); intro.
  subst i0.
  rewrite H0 in H2.
  disc.
  rewrite (replace_nth2 _ _ _ _ _ _ H7 n).
  trivial.
  (* exit *)
  auto.
  (* bar1 *)
  left.
  case (eq_nat_dec i0 i); intro.
  subst i0.
  rewrite H0 in H2.
  disc.
  rewrite (replace_nth2 _ _ _ _ _ _ H6 n).
  split; trivial.
  eapply bpiwp_lupmvbp; eauto.
  (* bar2 *)
  case (eq_nat_dec i0 i); intro.
  subst i0.
  rewrite H0 in H2.
  disc.
  copy H0.
  rewrite <- (replace_nth2 _ _ _ _ _ _ H6 n) in H0.
  case (eq_nat_dec bn bn0); intro.
  (* case A *)
  subst bn0.
  right.
  generalize (transition_threads_changes _ _ _ H9); intro.
  detach H11.
  generalize (bpiwp_lupmvbp _ _ _ _ _ _ _ _ _ _ _ H5 H); intro.
  copy pf.
  destruct pf0 as [? _].
  assert (exists stk, exists hp, exists bm, exists c, nth_error thds i = Some (thread stk hp bm (Waiting bn dir mv c))).
    exists stk.
    exists hp.
    exists bs.
    exists k.
    trivial.
  rewrite <- H13 in H14.
  clear H13.
  destruct H14 as [hx [bx ?]].
  assert (lookup_slot_bp bp' bn dir mv (i, (hx, bx))).
    rewrite <- (bp_inc_wp_constant _ _ _ _ _ _ H5); trivial.
    case (eq_nat_dec rightdirection dir); auto.
    case (eq_nat_dec rightmove mv); auto.
    intros.
    subst rightmove rightdirection.
    apply (lookup_move_bp_fun _ _ _ _ _ _ _ _ H3) in H.
    destruct H.
    subst pre0 post0.
    destruct (lookup_move_bp_facts _ _ _ _ _ _ H3) as [? [_ [? [_ [? ?]]]]].
    generalize (bp_lupmvslt_pre _ _ _ _ _ _ _ H3 H13); intro.
    destruct (lupmvbp_lupstbp _ _ _ _ _ _ H3) as [bst ?].
    generalize (bp_lupsltst _ _ _ _ _ _ H17 H13); intro.
    change bp with (cs_barpool (cstate _ _ _ _ pf)) in H17.
    change thds with (cs_thds (cstate _ _ _ _ pf)) in H2.
    generalize (lupst_joins_thd _ _ _ H17 _ _ _ _ _ H2); intro.
    destruct H19 as [tst ?].
    spec H14 tst (empty_stack, hx, bx) (empty_stack, hp', bs').
    assert (join_sub bst tst) by (exists (empty_stack, hp0, bs0); trivial).
    detach H14.
    2: eapply join_sub_trans; eauto.
    inv_sc H4.
    subst s s0.
    destruct H24.
    inversion H25.
    inversion H4.
    subst h0 b0 h b.
    clear H25 H4.
    rewrite <- H in H21.
    spec H21 empty_stack.
    assert (join_sub (empty_stack, hp', bs') (empty_stack, hp0, bs0)) by (exists (empty_stack, hp'', bs''); repeat (split; trivial)).
    assert (join_sub (empty_stack, hp0, bs0) tst) by (exists bst; apply join_com; trivial).
    detach H14.
    2: eapply join_sub_trans; eauto.
    spec H14 H16 H21.
    inversion H14.
    subst hx bx.
    destruct H18 as [obst ?].
    apply join_com in H18.
    destruct (join_assoc _ _ _ _ _ H18 H19) as [f [? ?]].
    apply join_com in H25.
    assert (join (empty_stack, hp'', bs'') (empty_stack, hp', bs') (empty_stack, hp0, bs0)) by (apply join_com; repeat (split; trivial)).
    destruct (join_assoc _ _ _ _ _ H27 H25) as [f' [? ?]].
    unfold slot2state in H28.
    generalize (join_self _ _ H28); intro.
    subst f'.
    assert (identity (empty_stack, hp', bs')) by (rewrite identity_unit_equiv; trivial).
    apply identity_emptystate in H30; auto.
    inversion H30.
    subst bs' hp'.
    destruct H15 as [ps ?].
    apply H15 in H16.
    clear -H16.
    unfold slot2state in H16.
    inv_sc H16.
    subst s s0.
    generalize (split_identity _ _ H2 empty_bm_identity); intro.
    apply identity_emptybm in H.
    subst b.
    destruct H3.
    destruct ps.
    disc.
  destruct (lookup_slot_out _ _ _ _ H8 _ _ _ _ H12 _ H14) as [sltx [? [? ?]]].
  spec H11 mv sltx H15.
  destruct sltx as [? [? ?]].
  unfold slot2tid, fst in *.
  subst i.
  rewrite H0 in H11.
  spec H11 stk hp bs bn dir.
  spec H11 mv k.
  detach H11; trivial.
  destruct H11 as [hx' [bx' [? ?]]].
  exists h.
  exists b.
  exists hx'.
  exists bx'.
  destruct H11 as [[? ?] ?].
  unfold slot2state, fst, snd in *.
  apply join_com in H18.
  apply join_com in H19.
  repeat (split; trivial).
  eapply bpt_uni_out; eauto.
  eapply bpi_uni_nin_uni; eauto.
  change bp with (cs_barpool (cstate _ _ _ _ pf)).
  apply cs_slots_unique.
  change bp with (cs_barpool (cstate _ _ _ _ pf)).
  eapply cs_running_nin_bp; eauto.
  (* case B *)
  left.
  generalize (transition_threads_constant _ _ _ H9 i); intro.
  rewrite <- H11.
  split; trivial.
  generalize (bpiwp_lupmvbp _ _ _ _ _ _ _ _ _ _ _ H5 H); intro.
  eapply bptrans_lupmv; eauto.
  intro.
  assert (slots_unique_tid_bp bp').
    eapply bpi_uni_nin_uni; eauto.
    change bp with (cs_barpool (cstate _ _ _ _ pf)).
    apply cs_slots_unique.
    change bp with (cs_barpool (cstate _ _ _ _ pf)).
    eapply cs_running_nin_bp; eauto.
  apply n0.
  generalize (bptrans_nio _ _ _ _ H13 H8 bn dir mv); intro.
  generalize (bp_inc_wp_constant _ _ _ _ _ _ H5 bn dir mv); intro.
  copy pf.
  destruct pf0 as [? _].
  spec H16 bn dir mv i.
  assert (exists stk, exists hp, exists bm, exists c, nth_error thds i = Some (thread stk hp bm (Waiting bn dir mv c))).
    exists stk.
    exists hp.
    exists bs.
    exists k.
    trivial.
  rewrite <- H16 in H17.
  destruct H17 as [hx [bx ?]].
  rewrite H15 in H17; auto.
  spec H14 (i, (hx, bx)) H17.
  rewrite H14; auto.
Qed.

Lemma cs_threads_preserved_barrier2 : forall cs i stk h bs bn dir mv c cs',
  nth_error (cs_thds cs) i = Some (thread stk h bs (Waiting bn dir mv c)) ->
  conc_step cs cs' ->
  (nth_error (cs_thds cs') i = Some (thread stk h bs (Waiting bn dir mv c))) \/
  (exists h', exists bs', nth_error (cs_thds cs') i = Some (thread stk h' bs' (Running c))).
Proof.
  intros.
  destruct (lookup_move_waiting _ _ _ _ _ _ _ _ _ H) as [pre [post ?]].
  generalize (cstep_waiting_inv cs i stk h bs bn dir mv c pre post cs'); intro.
  spec H2 H1 H H0.
  destruct H2.
  destruct H2; auto.
  destruct H2 as [hx [bx [hx' [bx' [? [? [? ?]]]]]]].
  right.
  exists hx'.
  exists bx'.
  trivial.
Qed.

Lemma cstep_head_running: forall cs cs',
  conc_step cs cs' ->
  forall n, 
    head (cs_sched cs) = Some n ->
      exists c,
        option_map th_ctl (nth_error (cs_thds cs) n) = Some (Running c).
Proof.
  intros; 
  inversion H; 
  subst; 
  simpl; 
  inversion H0; 
  subst; 
  rewrite H1; 
  simpl; 
  eexists; 
  reflexivity.
Qed.

Lemma barDefs_cstep_preserved : forall cs cs',
  conc_step cs cs' ->
  bp_bardefs (cs_barpool cs) = bp_bardefs (cs_barpool cs').
Proof.
  inversion 1; auto.
  simpl.
  eapply barDefs_bp_inc_waitpool_preserved; eauto.
  generalize (barDefs_bp_inc_waitpool_preserved _ _ _ _ _ _ H3); intro.
  simpl.
  rewrite H10.
  eapply barDefs_bp_transition_preserved; eauto.
Qed.

Lemma cs_barrier_stuck: forall cs cs',
  conc_step cs cs' ->
  forall i stk hp bs bn dir mv k,
  nth_error (cs_thds cs) i = Some (thread stk hp bs (Waiting bn dir mv k)) ->
  forall pre post mv' pre' post',
  mv <> mv' ->
  lookup_move_bp (cs_barpool cs) bn dir mv pre post ->
  lookup_move_bp (cs_barpool cs) bn dir mv' pre' post' ->
  forall st,
  (sepconj pre Tru) st ->
  (sepconj pre' Tru) st ->
  forall cst,
  lookup_bstate_bp (cs_barpool cs) bn cst ->
  forall slt,
  join (empty_stack, hp, bs) (slot2state slt) st ->
  lookup_slot_bp (cs_barpool cs) bn dir mv slt ->
  nth_error (cs_thds cs') i = Some (thread stk hp bs (Waiting bn dir mv k)) /\
  lookup_slot_bp (cs_barpool cs') bn dir mv slt /\
  lookup_bstate_bp (cs_barpool cs') bn cst.
Proof.
  intros.
  inversion H; subst.
  (* seq *)
  simpl in *.
  case (eq_nat_dec i0 i); intro.
  subst i0; rewrite H0 in H9; disc.
  generalize (replace_nth2 _ _ _ _ _ _ H14 n); intro.
  rewrite H15.
  rewrite H0.
  repeat (split; trivial).
  (* exit *)
  simpl in *.
  case (eq_nat_dec i0 i); intro.
  subst i0; rewrite H0 in H9; disc.
  rewrite H0.
  repeat (split; trivial).
  (* bar1 *)
  simpl in *.
  case (eq_nat_dec i0 i); intro.
  subst i0; rewrite H0 in H9; disc.
  generalize (replace_nth2 _ _ _ _ _ _ H13 n); intro.
  rewrite H15.
  rewrite H0.
  split; trivial.
  split.
  2: eapply bpiwp_lupbst; eauto.
  case (eq_nat_dec bn0 bn); intro.
  2: rewrite <- (bp_inc_wp_constant _ _ _ _ _ _ H12 bn dir mv slt); auto.
  case (eq_nat_dec rightdirection dir); intro.
  2: rewrite <- (bp_inc_wp_constant _ _ _ _ _ _ H12 bn dir mv slt); auto.
  case (eq_nat_dec rightmove mv); intro.
  2: rewrite <- (bp_inc_wp_constant _ _ _ _ _ _ H12 bn dir mv slt); auto.
  subst rightmove rightdirection bn0.
  rewrite H0 in H15.
  copy pf'.
  destruct pf'0 as [? _].
  spec H16 bn dir mv i.
  assert (exists stk, exists hp, exists bm, exists c, nth_error thds' i = Some (thread stk hp bm (Waiting bn dir mv c))).
    exists stk.
    exists hp.
    exists bs.
    exists k.
    trivial.
  rewrite <- H16 in H17.
  destruct H17 as [h [b ?]].
  generalize (bpiwp_luplstbp _ _ _ _ _ _ H12); intro.
  apply (lookup_slot_bp_fun _ _ _ _ _ _ H17) in H18.
  inversion H18.
  destruct n; auto.
  (* bar2 *)
  simpl in *.
  case (eq_nat_dec i0 i); intro.
  subst i0; rewrite H0 in H9; disc.
  case (eq_nat_dec bn bn0); intro.
  (* case 1 *)
  subst bn0.
  destruct H1.
  destruct bp as [bars bpst wfbp].
  destruct bp' as [bars' bpst' wfbp'].
  copy H12.
  destruct H12.
  remember (nth_error bars bn).
  remember (nth_error bars' bn).
  icase e.
  icase e0.
  destruct H17.
  destruct d0.
  destruct dbs_wp.
  assert (rightdirection = dir).
    clear -H1 pf H0.
    destruct pf as [? _].
    spec H bn dir mv i.
    assert (exists stk, exists hp, exists bm, exists c, nth_error thds i = Some (thread stk hp bm (Waiting bn dir mv c))).
      do 4 eexists; eauto.
    rewrite <- H in H2.
    destruct H2 as [h [b ?]].
    destruct H1.
    red in H2.
    simpl in H3, H2.
    icase (nth_error bars bn).
    icase (nth_error bars' bn).
    destruct H3.
    destruct d.
    destruct H2.
    destruct dbs_wp.
    simpl in *.
    subst.
    red in H3.
    destruct d0.
    destruct dbs_wp.
    destruct H3 as [_ [_ [? _]]].
    destruct H2; disc.
    inversion H2.
    trivial.
  subst rightdirection.
  assert (rightmove <> mv).
    case (eq_nat_dec rightmove mv); auto; intro.
    subst rightmove.
    apply (lookup_move_bp_fun _ _ _ _ _ _ _ _ H2) in H10.
    destruct H10.
    subst pre0 post0.
    generalize (bp_lupmvslt_pre _ _ _ _ _ _ _ H2 H8); intro.
    destruct (lupmvbp_lupstbp _ _ _ _ _ _ H2) as [bst ?].
    generalize (bp_lupsltst _ _ _ _ _ _ H19 H8); intro.
    change thds with (cs_thds (cstate _ _ _ _ pf)) in H9.
    change (barpool bars bpst wfbp) with (cs_barpool (cstate _ _ _ _ pf)) in H19.
    generalize (lupst_joins_thd _ _ _ H19 _ _ _ _ _ H9); intro.
    unfold cs_thds, cs_barpool in *.
    destruct (lookup_move_bp_facts _ _ _ _ _ _ H2) as [cs [_ [prub [_ [? ?]]]]].
    destruct H21 as [tst ?].
    spec H22 tst (slot2state slt) (empty_stack, hp', bs').
    detach H22.
    2: apply join_sub_trans with bst; auto; eexists; apply H21.
    inv_sc H11.
    subst s s0.
    destruct H27.
    inversion H28.
    inversion H11.
    subst h0 b0 h b.
    clear H28 H11.
    detach H22.
    2: apply join_com in H21; apply join_sub_trans with (empty_stack, hp0, bs0); try (eapply join_join_sub; eauto; contradiction); exists (empty_stack, hp'', bs''); repeat (split; trivial).
    rewrite <- prub in H24.
    spec H24 empty_stack.
    spec H22 H10 H24.
    destruct slt as [? [? ?]].
    inversion H22.
    subst h b.
    clear H22.
    destruct H20 as [stx ?].
    apply join_com in H11.
    destruct (join_assoc _ _ _ _ _ H11 H21) as [f [? ?]].
    apply join_com in H20.
    assert (join (empty_stack, hp', bs') (empty_stack, hp'', bs'') (empty_stack, hp0, bs0)) by repeat (split; trivial).
    apply join_com in H27.
    destruct (join_assoc _ _ _ _ _ H27 H20) as [f' [? ?]].
    generalize (join_self _ _ H28); intro.
    subst f'.
    assert (identity (empty_stack, hp', bs')) by (rewrite identity_unit_equiv; trivial).
    generalize (identity_emptystate _ H30); intro.
    detach H31; auto.
    inversion H31.
    subst hp' bs'.
    destruct H23 as [ps ?].
    apply H23 in H10.
    clear -H10.
    inv_sc H10.
    unfold slot2state in *.
    subst s s0.
    generalize (split_identity _ _ H2 empty_bm_identity); intro.
    apply identity_emptybm in H.
    subst b.
    destruct H3.
    unfold empty_bm, barst_builder in H.
    destruct ps.
    simpl in H.
    disc.
  copy H1.
  remember (nth_error bars bn).
  remember (nth_error bars' bn).
  icase e.
  icase e0.
  destruct H20.
  destruct d0.
  destruct dbs_wp.
  destruct d1.
  destruct dbs_wp.
  inversion Heqe0.
  subst.
  inversion Heqe.
  subst.
  clear Heqe0 Heqe.
  rewrite <- Heqe1 in H21.
  rewrite <- Heqe2 in H21.
  destruct H21.
  red in H21.
  destruct H21 as [? [? [? [? [? [? [? ?]]]]]]].
  subst.
  eapply stuck_not_bp_ready.
  apply H14.
  eapply bpiwp_lupmvbp; eauto.
  eapply bpiwp_lupmvbp; eauto.
  red.
  simpl.
  rewrite <- Heqe2.
  reflexivity.
  rewrite <- (bp_inc_wp_constant _ _ _ _ _ _ H1); auto.
  apply H8.
  apply H7.
  assert (lookup_state_bp (cs_barpool (cstate _ _ _ _ pf)) bn wp_st0).
    red.
    simpl.
    rewrite <- Heqe1.
    trivial.
  change thds with (cs_thds (cstate _ _ _ _ pf)) in H9.
  change thds with (cs_thds (cstate _ _ _ _ pf)) in H0.  
  destruct (lupst_joins_thd2 _ _ _ H21 _ _ _ _ _ _ _ _ _ _ n H9 H0) as [thsts [? ?]].
  clear -H23 H25 H11 H28.
  inv_sc H11.
  subst s s0.
  destruct H3.
  inversion H4.
  inversion H.
  subst h0 b0 h b.
  clear H4 H.
  unfold slot2state in H28.
  destruct H25 as [tst ?].
  assert (join (empty_stack, hp', bs') (empty_stack, hp'', bs'') (empty_stack, hp0, bs0)) by repeat (split; trivial).
  apply join_com in H23.
  apply join_com in H.
  destruct (join_assoc _ _ _ _ _ H23 H) as [f [? ?]].
  apply join_com in H3.
  destruct (join_assoc _ _ _ _ _ H3 H4) as [f' [? ?]].
  apply (join_eq _ _ _ _ H28) in H6.
  subst f'.
  apply join_com in H5.
  destruct (join_assoc _ _ _ _ _ H7 H5) as [f'' [? ?]].
  exists f''.
  apply join_com.
  trivial.
  trivial.
  (* case 2 *)
  split.
  rewrite <- (transition_threads_constant _ _ _ H16 i).
  rewrite (replace_nth2 _ _ _ _ _ _ H13 n).
  trivial.
  change thds with (cs_thds (cstate _ _ _ _ pf)) in H9.
  generalize (cs_running_nin_bp _ _ _ _ _ _ H9); intro.
  generalize (cs_slots_unique (cstate _ _ _ _ pf)); intro.
  unfold cs_barpool in *.
  generalize (bpi_uni_nin_uni _ _ _ _ _ _ H12 H18 H17); intro.
  clear -pf H0 H15 H12 H19 n0.
  destruct pf as [? _].
  spec H bn dir mv i.
  assert (exists stk, exists hp, exists bm, exists c,
                 nth_error thds i = Some (thread stk hp bm (Waiting bn dir mv c))).
    exists stk.
    exists hp.
    exists bs.
    exists k.
    trivial.
  rewrite <- H in H1.
  destruct H1 as [h [b ?]].
  clear -n0 H1 H15 H12 H19.
  rewrite (bp_inc_wp_constant _ _ _ _ _ _ H12) in H1; auto.
  intro.
  apply n0.
  rewrite (bptrans_nio _ _ _ _ H19 H15 _ _ _ _ H1 H).
  trivial.
  split.
  eapply bptrans_slt; eauto.
  rewrite <- (bp_inc_wp_constant _ _ _ _ _ _ H12 _ _ _ _); auto.
  eapply bptrans_bstate; eauto.
  eapply bpiwp_lupbst; eauto.
Qed.

Lemma cs_lookup_move_bp: forall cs cs',
  conc_step cs cs' ->
  forall bn cst,
  lookup_bstate_bp (cs_barpool cs) bn cst ->
  lookup_bstate_bp (cs_barpool cs') bn cst ->
  forall dir mv pre post,
  lookup_move_bp (cs_barpool cs) bn dir mv pre post ->
  lookup_move_bp (cs_barpool cs') bn dir mv pre post.
Proof.
  intros.
  apply barDefs_cstep_preserved in H.
  unfold bp_bardefs in H.
  destruct cs.
  destruct cs'.
  simpl in *.
  clear - H H0 H1 H2.
  destruct cs_barpool.
  destruct cs_barpool0.
  red in H0, H1, H2.
  red.
  simpl in *.
  remember (nth_error bp_bars bn).
  icase e.
  remember (nth_error bp_bars0 bn).
  icase e.
  destruct d as [? ? ? bd ?].
  destruct d0 as [? ? ? bd' ?].
  subst dbs_cs dbs_cs0.
  replace bd' with bd; trivial.
  generalize (nth_error_map _ _ dbs_bd bp_bars bn); intro.
  generalize (nth_error_map _ _ dbs_bd bp_bars0 bn); intro.
  rewrite H in H0.
  rewrite H0 in H1.
  rewrite <- Heqe in H1.
  rewrite <- Heqe0 in H1.
  simpl in H1.
  inversion H1.
  trivial.
Qed.

Lemma cs_threads_preserved_notrunning : forall cs i stk h bs c cs',
  nth_error (cs_thds cs) i = Some (thread stk h bs (Running c)) ->
  head (cs_sched cs) <> Some i -> 
  conc_step cs cs' ->
  nth_error (cs_thds cs') i = Some (thread stk h bs (Running c)).
Proof.
  intros.
  destruct cs as [sched thds bp ap wfcs].
  inversion H1; subst.
(* case seq *)
  simpl.
  rewrite (replace_nth2 _ _ _ _ _ _ H12); trivial.
  simpl in H0.
  intro.
  apply H0.
  rewrite H2.
  trivial.
(* case switch *)
  simpl in H0, H.
  simpl.
  trivial.
(* case bar1 *)
  rewrite (replace_nth2 _ _ _ _ _ _ H11); trivial.
  simpl in H0.
  intro.
  apply H0.
  rewrite H2.
  trivial.
(* case bar2 *)
  simpl.
  generalize (replace_nth2 _ _ _ _ _ i H10).
  intro.
  simpl in H0.
  assert (i0 <> i) by (intro; apply H0; rewrite H3; trivial).
  spec H2 H3.
  simpl in H.
  rewrite H in H2.
  rewrite <- H2.
  symmetry.
  eapply transition_threads_constant; eauto.
  intro.
  apply In_nth_error in H4.
  destruct H4 as [j ?].
  rewrite nth_error_map in H4.
  remember (nth_error (dbsto_slots out) j).
  icase e.
  symmetry in Heqe.
  destruct (bp_trans_out2 _ _ _ _ H13 _ _ Heqe) as [sltx [? ?]].
  inversion H4.
  subst.
  assert ((rightmove <> j \/ rightdirection <> dbsto_dir out) \/ (rightmove = j /\ rightdirection = dbsto_dir out)).
    case (eq_nat_dec rightmove j); auto.
    case (eq_nat_dec rightdirection (dbsto_dir out)); auto.
  destruct H15.
  (* case A *)
  rewrite <- (bp_inc_wp_constant _ _ _ _ _ _ H9) in H5.
  2: destruct H15; auto.  
  rewrite H11 in H.
  clear -H5 H pf0.
  destruct pf0 as [? _].
  destruct sltx as [? [? ?]].
  spec H0 bn (dbsto_dir out) j t.
  assert (exists h, exists b, lookup_slot_bp bp bn (dbsto_dir out) j (t, (h, b))) by (exists h0; exists b; trivial).
  rewrite H0 in H1.
  destruct H1 as [? [? [? [? ?]]]].
  simpl in H.
  rewrite H1 in H.
  disc.
  (* case B *)
  destruct H15.
  subst.
  apply bpiwp_luplstbp in H9.
  apply (lookup_slot_bp_fun _ _ _ _ _ _ H5) in H9.
  subst.
  simpl in *.
  rewrite H11 in H3.
  auto.
Qed.

End ConcMachine.

